#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"


if [info exists tk_patchLevel ] {
    set level $tk_patchLevel
} else {
    switch -glob -- $tcl_patchLevel {
	{7.5} {set level 4.1}
	{7.6} {set level 4.2}
	{*} {set level $tcl_patchLevel}
    }
}


set SESHLIB [file join [pwd] tcllib]
set env(SESHLIB) $SESHLIB


proc makeentete {fic} {
    global env SESHLIB tcl_platform
    puts $fic \#!/bin/sh
    puts -nonewline $fic {# the next line restarts using wish}
    puts $fic \\
    puts $fic [concat  exec [info nameofexecutable] {"$0" "$@"}]
    if {[string equal $tcl_platform(platform) "unix"]} {
	puts $fic "set env(SESHLIB) [list $SESHLIB]"
    } else {
	puts $fic {set env(SESHLIB) [file join [pwd] tcllib]}
    }
    puts $fic {set SESHLIB $env(SESHLIB)}
    puts $fic "#for $tcl_platform(platform)"
    puts $fic {set env(ldext) so}
    puts $fic {set env(DEFAULTSELECTION) CLIPBOARD}
    puts $fic {event add <<Paste>> <Control-y>}
    puts $fic {event add <<Paste>> <Button-2>}
}

set env(ldext) [info sharedlibextension]

##############################################################################
# tkseshb

set fic [open tkseshb.tcl w]

makeentete $fic

puts $fic {source [file join $SESHLIB essaisesh.tcl]}
close $fic

##############################################################################
# Fontedit.tcl

set fic [open fontedit.tcl w]
makeentete $fic
puts $fic {set x [list $SESHLIB [file join $SESHLIB stooop] [file join $SESHLIB scwoop]]}
puts $fic {set x [concat $x $auto_path]; lappend x . ; set auto_path $x}
puts $fic {source [file join $SESHLIB FontEdit fonteed.tcl]}
close $fic

##############################################################################
# ligedit.tcl
set fic [open ligedit.tcl w]
makeentete $fic
puts $fic {set x [list $SESHLIB [file join $SESHLIB stooop] [file join $SESHLIB scwoop]]}
puts $fic {set x [concat $x $auto_path]; lappend x . ; set auto_path $x}
puts $fic {source [file join $SESHLIB ligeditor.tcl]}
close $fic

##############################################################################
# PrepareDB.tcl
set fic [open prepareDB.tcl w]
makeentete $fic
puts $fic {set x [list $SESHLIB]}
puts $fic {set x [concat $x $auto_path]; lappend x . ; set auto_path $x}
puts $fic {package require dbmaker}
puts $fic {package require stooop}
puts $fic {namespace import stooop::*}
puts $fic {wm withdraw .}
puts $fic {dbmaker::makeDB}
puts $fic {package require Tksesh}
puts $fic {source [file join $env(SESHLIB) SQLLoad sqlload.tcl]}
puts $fic {dbmaker::loadTmpDico}
puts $fic {exit}
close $fic


##############################################################################

switch -exact -- "$tcl_platform(platform)" {
    unix {
	file attributes tkseshb.tcl -permissions 0755
	file attributes fontedit.tcl -permissions 0755
	file attributes ligedit.tcl -permissions 0755
	file attributes prepareDB.tcl -permissions 0755
    }
}

set curr [pwd]

lappend auto_path $SESHLIB [file join $SESHLIB stooop] [file join $SESHLIB scwoop]

# Due to a bug in TCL803, we don't build the pkgindex file if
# we are using this version of tcl. Let's hope the rest works
if {[string compare $level "8.0.3"] != 0} {
    pkg_mkIndex -direct $SESHLIB legendes.tcl seshbase.tcl utils.tcl \
	normalize.tcl\
	Glossary.tcl\
	tag.tcl\
	yellow.tcl signesdef.tcl\
	fileentry.tcl dbfilesel.tcl\
	seshw.tcl \
	search.tcl tradedit.tcl signes.tcl \
	seshuser.tcl seshtext.tcl\
	bzrToType1.tcl\
	ligeditor.tcl

    pkg_mkIndex [file join $SESHLIB LIBS] \
	    tksesh$env(ldext) dbinter$env(ldext)\
	    tclsqlite$env(ldext)\
	    pbmtoxbm$env(ldext)\
	    pbmtoxbm$env(ldext)\
	    sbzr$env(ldext)
}
cd $curr

wm withdraw .

# Bitmap generation  is only possible on UNIX platform by now. Too bad :-(

if {[string compare "$tcl_platform(platform)" "unix"] == 0} {
package require seshw
#package require signes

#source [file join $SESHLIB fonctions.tcl]
# Should be in only one file !
}
tk_dialog .dialog "Welcome to Tksesh !" "Tksesh installation is now complete.\nYou can start tksesh by clicking on tkseshb.tcl" "" 0 "OK"

exit


