/*
 * A portable database
 * - System independant, the database files should be usable on any system
 *   (the CONTENT may not be portable, though)
 * 
 */


/*  1998, 2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: btdb.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/*
 * $Log: btdb.c,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 * Revision 1.11  2001/10/14 12:42:11  rosmord
 * *** empty log message ***
 *
 * Revision 1.10  2000/07/22 12:25:03  rosmord
 * Deletion function added. Some test dones
 *
 */


#define _POSIX_SOURCE 1
#include <ctype.h>							/* char testing functions */
#include <errno.h>							/* errors */
#include <fcntl.h>							/* fd manipulation */
#include <stddef.h>							/* NULL, size_t, and so on */
#include <string.h>

#include "general.h"
#include "chaine.h"

#include "btdb.h"
#include "btree.h"

/* #define TclAccess access*/
/* Private functions defined here : */

/* Wrappers around fread and fwrite */
PRIVATE void write_dbptr(struct MyFile *fd, dbptr_t val);
PRIVATE dbptr_t read_dbptr(struct MyFile *fd);
/* open and close plain db files */
PRIVATE int db0_open(db0* db, char* name, int flags, unsigned char blocksize, mode_t mode);
PRIVATE void db0_close(db0 *db);
/* alloc a new block */
PRIVATE dbptr_t db0_alloc(db0 *db);
/* Free a record */
PRIVATE void
db0_free_current_record(db0 *);
/* */
PRIVATE dbptr_t db0_go_to_end_of_record(db0 *db);
/* load a new block */
PRIVATE void db0_chose(db0 *db, dbptr_t where);
/* write current block to file if needed */
PRIVATE void db0_writecurrent(db0 *db);
/* init a new record */
PRIVATE dbptr_t db0_init_record(db0 *db);
/* open a record */
PRIVATE dbptr_t db0_open_record(db0 *db, dbptr_t rec);
/* read data from the current record*/
PRIVATE int db0_read(db0 * db, char * s, unsigned int size);
/* write data to the current record*/
PRIVATE int db0_write(db0 *db, char * s, unsigned int size);
/* seek in the record */
PRIVATE int db0_seek(db0 * db, int offset, int whence);
/* write data in record */
PRIVATE void db0_write_dbptr(db0 *d, dbptr_t val);
PRIVATE dbptr_t db0_read_dbptr(db0 *d);
/* pascal like string */
PRIVATE void free_pstring_data(pstring *p);
PRIVATE void duplicate_pstring(pstring *dest, pstring *src);
PRIVATE void set_pstring_from_char(pstring* p, char *s);
PRIVATE void transfer_pstring(pstring* dest, pstring *src);
PRIVATE int compare_pstring(pstring *a, pstring *b);
PRIVATE void db0_read_pstring(db0* d, pstring *ps);
PRIVATE void db0_write_pstring(db0*d, pstring *ps);
/* BTree functions */
PRIVATE void btdb_init_new_node(btdb_t *arbre, btree_node_t *n);
PRIVATE void btdb_write_node(btdb_t *arbre, btree_node_t *t);
PRIVATE void btdb_read_node(btdb_t *arbre, btree_node_t *t, dbptr_t nodepos);
PRIVATE void btdb_free_node_in_memory(btree_node_t *t);
PRIVATE void db_simple_insert_node(
																	 btree_node_t *current_node,
																	 int position,
																	 pstring * key,
																	 dbptr_t data,
																	 dbptr_t next
																	 );
PRIVATE void btdb_insert_node_aux(
																	btdb_t * arbre,	/* the tree */
																	btree_node_t * t, /* the current node */
																	pstring *key, /* the key */
																	pstring *data,	/* the data */
																	int * explode,	/* did the tree explode ? */
																	pstring *new_key, /* the new node element, key part */
																	dbptr_t *new_data, /* the new node element, data part */
																	dbptr_t *new_node_ptr,	/* the new node */
																	int (*compare)(pstring*, pstring*)	/* comparison function */
																	);
#ifdef DEBUG
PRIVATE void btdb_dump_aux(btdb_t * arbre, dbptr_t ptr, int prof);
#endif
PRIVATE void btdb_apply_aux(btdb_t * arbre,
														dbptr_t ptr,
														void (*f)(pstring *, pstring *, void *),
														void *external_data,
														int keyonly);
PRIVATE btdb_index_t btdb_index_empty_index(btdb_t * arbre);
PRIVATE int btdb_index_is_empty(btdb_index_t idx);
PRIVATE void btdb_index_pop(btdb_index_t idx);
PRIVATE struct btdb_index_elt * btdb_index_read(btdb_index_t idx);
PRIVATE void btdb_index_push(btdb_index_t idx, struct btdb_index_elt *elt);
PRIVATE void btdb_index_get_first(btdb_t * arbre,
																	dbptr_t node, 
																	btdb_index_t res);
PRIVATE void db_simple_delete_node(btree_node_t *current_node,
																	 int position
																	 );
PRIVATE void btdb_merge_nodes(btdb_t * arbre,
															btree_node_t * t,
															int pos,
															btree_node_t * left,
															btree_node_t * right);


/* USER'S OVERVIEW :
	 A number of level of commands are defined here :

	 level 0 :
	   This correspond to a database that is simply a set of records.
		 A record has a number, and data. The numbers are given 
		 by the system. They are basically like addresses.

		 **************************************************************
		 These two open and close a database

		 PRIVATE int
		 db0_open(db0* db, char* name, int flags, unsigned char blocksize, mode_t mode)

		 PRIVATE void
		 db0_close(db0 *db)

		 **************************************************************
		 
		 New record :  
		 a= db0_init_record(&db);

		 Old record :
		 db0_open_record(&d, b);
		 
		 Writing into a record :
		 
		 db0_write(&d, aa, 4);

		 Reading a record :

		 db0_read(&d, &c, 1)
		 
		 Seeking a record :

		 PRIVATE int
		 db0_seek(db0 * db, int offset, int whence)
		 
		 */


PRIVATE void
write_dbptr(struct MyFile *fd, dbptr_t val)
{
	dbptr s;
	s[0]= (val >> 24) % 256;
	s[1]= (val >> 16) % 256;
	s[2]= (val >> 8) % 256;
	s[3]= val % 256;
	
	MyWrite(fd, s, sizeof(dbptr));
}

PRIVATE dbptr_t
read_dbptr(struct MyFile *fd)
{
	dbptr s;
	dbptr_t val;
	MyRead(fd, s, sizeof(dbptr));
	val= 	 ((unsigned int)s[0] << 24) + ((unsigned int)s[1] << 16) + 
		((unsigned int)s[2] << 8) +  (unsigned int)s[3];
	return val;
}



/* The database is made of two parts :
	 - a linked list of blocks, which contain data and key values
	 - a B-tree file, which contains pointers to the data file

	 System and data independance : everything is read like unsigned chars ;
	 only aftewards, addresses and pointers are converted.
	 High bits go first: 
	 e.g.
	 (unsigned int)s[0] << 24 + (unsigned int)s[1] << 16 + 
	 (unsigned int)s[2] << 8 +  (unsigned int)s[3]

	 Btree part :
	 a node :	 
	 Number of ITEM : octet
	 ADDRESS of father
	 table of N items
	 
	 item :: KEY_ADRESS
	 */



PRIVATE void
db0_writecurrent(db0 *db);

PRIVATE void
db0_chose(db0 *db, dbptr_t where);



/**
	like open, BUT : 
	flags is either O_RDONLY for reading a database
	or O_RDWR for reading/writing.
	in this case, the database is created if it doesn't already
	exist

	blocksize is only used when creating the base.
	*/
	
PRIVATE int
db0_open(db0* db, char* name, int flags, unsigned char blocksize, mode_t mode)
{
	enum MyOpenFlags opts;
	
	/* only these flags are supported */
	switch(flags)
		{
		case O_RDONLY:
			opts= MYREAD;
			break;
		case O_RDWR:
			opts= MYWRITE;			
			break;
		default:
			return -1;
		}
	
	db->has_been_modified= FALSE;
	db->flags= flags;
	db->has_current= FALSE;

	/* if the flag is O_RDWR, 
		 test if the file exist and create it if necessary */
	
	if (MyAccess(name, F_OK)== -1 && flags == O_RDWR)
		{
			struct MyFile fd;
			

			if (MyOpen(&fd, name, MYCREATE) != -1)
				{
					/* size of block */
					MyWrite(&fd, &blocksize, 1);
					write_dbptr(&fd,0);
					MyClose(&fd);
				}
			else
				return -1;
		}

	if(MyOpen(&(db->fd), name, opts) != -1)
		{
			MyRead(&(db->fd), &(db->blocksize), sizeof(db->blocksize));
			db->firstfree= read_dbptr(&(db->fd));
			return 0;
		}
	else
		return -1;
}

PRIVATE void
db0_close(db0 *db)
{
	if (db->flags== O_RDWR)
		{
			db0_writecurrent(db);
			MySeek(&(db->fd), 0, SEEK_SET);
			if (db->has_been_modified)
				{
					MyWrite(&(db->fd), &(db->blocksize), sizeof(char));
					write_dbptr(&(db->fd), db->firstfree);
				}
		}
	MyClose(&(db->fd));
}

/* returns a free dbptr */
PRIVATE dbptr_t
db0_alloc(db0 *db)
{
	dbptr_t result;
	unsigned char used= 0;
	struct MyFile * fd= &(db->fd);

	if (db->firstfree != 0)
		{
			result= db->firstfree;
			MySeek(fd, result, SEEK_SET);
			db->firstfree= read_dbptr(fd);
			db->has_been_modified= TRUE;
			MySeek(fd, result, SEEK_SET);
		}
	else													/* no free dbptr : increase size */
		{
			char c= 0;
			int i;
			/* go to the end */
			MySeek(fd, 0, SEEK_END);
			/* write db->blocksize blocs */
			for (i=0; i< db->blocksize; i++)
				MyWrite(&(db->fd), &c, sizeof(char));
			MySeek(fd, -db->blocksize, SEEK_CUR);	/* beginning of the block */
			result= MyTell(fd);
		}
	
	write_dbptr(&(db->fd),0);
	write_dbptr(&(db->fd),0);
	MyWrite(&(db->fd), &used, sizeof(char));
	return result;
}

/* goes to the end of the current record 
	 a further read whould produce EOF, whereas a write would append a new
	 character to the file.
	 returns a dbptr_t to the current block.
 */

PRIVATE dbptr_t
db0_go_to_end_of_record(db0 *db)
{
	while (db->next != 0)
		db0_chose(db, db->next);
	db->position= db->used;
	return db->current;
}

/* frees current dbptr */
PRIVATE void
db0_free_current_record(db0 *db)
{
	struct MyFile *fd= &(db->fd);
	dbptr_t where;
	dbptr_t ending;

	where= db->current;
	ending= db0_go_to_end_of_record(db);
	
	MySeek(fd, ending, SEEK_SET);
	write_dbptr(&(db->fd), db->firstfree);
	db->firstfree= where;
	db->has_been_modified= TRUE;
	db->has_current= FALSE;
}

/**
	chooses a position in a binary db
	returns 0 or -1
	*/

PRIVATE void
db0_chose(db0 *db, dbptr_t where)
{
	struct MyFile *fd= &(db->fd);
	db0_writecurrent(db);
	db->position= 0;
	db->current= where;
	MySeek(fd, where, SEEK_SET) ;
	db->next= read_dbptr(fd);
	db->previous= read_dbptr(fd);
	MyRead(fd, &(db->used), sizeof(char));
	MyRead(fd, db->current_block, sizeof(char)*(db->blocksize - HEADSIZE));
}
	
/**
	save the current block to the disk
 */
PRIVATE void
db0_writecurrent(db0 *db)
{
	struct MyFile* fd= &(db->fd);
	if (db->has_current && db->current_must_be_saved)
		{
			MySeek(fd, db->current, SEEK_SET);
			write_dbptr(fd, db->next);
			write_dbptr(fd, db->previous);
			MyWrite(fd, &(db->used), sizeof(char));
			MyWrite(fd, db->current_block, (db->blocksize - HEADSIZE));
		}
}

/** 
	initialise a new record
	*/

PRIVATE dbptr_t
db0_init_record(db0 *db)
{
	dbptr_t new;
	db->first= new= db0_alloc(db);
	db0_chose(db, new);
	db->has_current= TRUE;
	db->current_must_be_saved= TRUE;
	return new;
}

/** 
	open a new record
	(already existing)
	*/

PRIVATE dbptr_t
db0_open_record(db0 *db, dbptr_t rec)
{
	db0_chose(db, rec);
	db->first= rec;
	db->has_current= TRUE;
	db->current_must_be_saved= FALSE;
	return rec;
}

/**
	reads size char or -1 in a db
	*/

PRIVATE int
db0_read(db0 * db, char * s, unsigned int size)
{
	int lu= 0;
	int end= FALSE;
	while (lu < size && !end)
		{
			int qte= db->used - db->position;
			int alire= size - lu;
			if ( alire <= (qte))
				{
					memcpy(s, db->current_block + db->position,alire);
					db->position+= alire;
					lu= size;
				}
			else
				{
					memcpy(s, db->current_block + db->position, 
								 qte);
					lu+= qte;
					s+= qte;
					if (db->next)
						{
							db0_chose(db, db->next);
						}
					else
						end= TRUE;
				}
		}
	return lu;
}

/**
	writes a char in a db returns 0 or -1
	*/

PRIVATE int
db0_write(db0 *db, char * s, unsigned int size)
{
	int written= size;
	while (size > 0)
		{
			if (size <= ((db->blocksize - HEADSIZE) - db->position))
				{
					memcpy(db->current_block+ db->position, s, size);
					db->current_must_be_saved= TRUE;
					db->position+= size;
					size= 0;
					if (db->used < db->position)
						db->used= db->position;
				}
			else
				{
					dbptr_t old_one;
					
					memcpy(db->current_block+ db->position, s, 
								 (db->blocksize - HEADSIZE) - db->position);
					db->current_must_be_saved= TRUE;
					db->used= (db->blocksize - HEADSIZE);
					size-= (db->blocksize - HEADSIZE) - db->position;
					s+= (db->blocksize - HEADSIZE) - db->position;
					
					/* now, get a new block : 
						 either next one, or a new one
						 */
					old_one= db->current;
					if (!db->next)
						{
							dbptr_t newdb;
							newdb= db0_alloc(db);
							db->next= newdb;
						}
					
					db0_chose(db, db->next);
					db->previous= old_one;
				}
		}
	return written;
}

/**
	moves in a record.
	if record size is too small,
	then it will grow.
 */

PRIVATE int
db0_seek(db0 * db, int offset, int whence)
{
	switch (whence)
		{
		case SEEK_SET:
			{
				int total= 0;

				db0_chose(db, db->first);
				while(total + (db->blocksize - HEADSIZE) < offset && db->next)
					{
						total+= (db->blocksize - HEADSIZE);
						db0_chose(db, db->next);
					}
				/* get to the last char of current record : */
				total+= db->used;
				db->position= db->used;
				/* If we weren't successful, let's increase the record's size */
				while (total < offset)
					{
						char c= 0;
						total++;
						db0_write(db, &c, 1);
					}
				/* now we are ready : */
				db->position= offset % (db->blocksize - HEADSIZE);
			}
			break;
		default:
			return -1;
		}
	return 0;
}

/**
	 dbptr and other stuff for db0 
*/

PRIVATE void
db0_write_dbptr(db0 *d, dbptr_t val)
{
	dbptr s;
	s[0]= (val >> 24) % 256;
	s[1]= (val >> 16) % 256;
	s[2]= (val >> 8) % 256;
	s[3]= val % 256;
	
	db0_write(d, (char *)s, sizeof(dbptr));
}

PRIVATE dbptr_t
db0_read_dbptr(db0 *d)
{
	dbptr s;
	dbptr_t val;
	db0_read(d, (char *)s, sizeof(dbptr));
	val= 	 ((unsigned int)s[0] << 24) + ((unsigned int)s[1] << 16) + 
		((unsigned int)s[2] << 8) +  (unsigned int)s[3];
	return val;
}


/* 
	 void
	 write_pstring(db0*d, pstring *s)
	 writes a string s of l characters to the current
	 position in d.

	 void
	 read_pstring(db0* d, pstring *s)

	 Reads a string from a db0. The string must have been written
	 with write_file_string. 
	 d : the database to read from 
	 *ps: pointer to the string in memory. Sufficient space will be allocated
	    which means a free is needed afterwards on the string part
*/

PRIVATE void 
free_pstring_data(pstring *p)
{
	if (p->data != NULL)
		{
			LIBERE(p->data);
			p->data= NULL;
		}
}

PRIVATE void
duplicate_pstring(pstring *dest, pstring *src)
{
	dest->length= src->length;
	dest->data= (char*) ALLOUER(sizeof(char)*src->length);
	memcpy(dest->data, src->data, src->length);
}

PRIVATE void
set_pstring_from_char(pstring* p, char *s)
{
	p->data= s;
	p->length= strlen(s);
}

/*
	void
	transfer_pstring(pstring* dest, pstring *src)
	
	-- transfers data between 2 pstrings. The source
	   is set to the empty string, and the data 
		 is transfered to the new string
		 No memory  allocation takes place
*/

PRIVATE void
transfer_pstring(pstring* dest, pstring *src)
{
	dest->length= src->length;
	dest->data= src->data;
	src->length=0;
	src->data=0;
}

/* compare_pstring returns -1 if a < b, 0 if a == b, and else 1 */
PRIVATE int 
compare_pstring(pstring *a, pstring *b)
{
	int i_max; 
	int i;

	if (a->length > b->length) 
		i_max= b->length;
	else 
		i_max= a->length;

	i=0;

	while(i< i_max && a->data[i] == b->data[i])
		i++;
	
	if (i == i_max)
		{
			/* the shortest one is the good one */
			if (a->length == b->length) 
				return 0;
			else if (a->length > b->length) 
				return 1;
			else 
				return -1;
		}
	else
		{
			if (a->data[i] > b->data[i])
				return 1;
			else
				return -1;
		}
}

PRIVATE void
db0_read_pstring(db0* d, pstring *ps)
{
	ps->length= db0_read_dbptr(d);
	ps->data= (char*)ALLOUER(sizeof(char)* ps->length);
	db0_read(d, ps->data, ps->length);
}

PRIVATE void
db0_write_pstring(db0*d, pstring *ps)
{
	db0_write_dbptr(d, ps->length);
	db0_write(d, ps->data, ps->length);
}


/**
	level 2 :
	
	A B-tree is an INDEX for a data file.
	As such, it contains the number of the field used as index.
	Note that nothing prevents the user to have empty fields in the
	data part.


	A B-Tree is an index for a key-value structure file like the previous one.
	A B-Tree Node is A RECORD of fixed length

	A btree file has the following form :
	
	Btree-file ::= HEADER {NODES}
	NODE ::= ()| NEXT_FREE_NODE(dbptr) filling
	HEADER::= NODE_SIZE(dbptr)  FIRST_FREE_NODE(dbptr)
	
	The first node is ALWAYS node 0. 
	
	E.G.
	
  PTR0 V0 PTR1 V1 PTR1 V2 ...
	is a tree node that represent 

	
 
	** Insertion in a B-tree :
	 
	 1) if the node is NOT a leaf :
	 find the next node
	 2) else :
	 if the node is not full, insert the
	 data in the right place
	 3) if the node is full
	    a) create a new node
      b) redistribute the values, except the middle value
			c) 
 */



/* Basic Node processing functions */

/* Init a new node in memory */

PRIVATE void
btdb_init_new_node(btdb_t *arbre, btree_node_t *n)
{
	n->is_on_disk= FALSE;
	/* n->save_needed= TRUE;*/
	n->address_on_disk= 0;
	n->n=0;
	/* There are n+1 defined pointers,
		 so we'd better put the first one at 0 */
	n->ptrs[0]= 0;
}

/* 
	 saves a node on disk. 
	 If address_on_disk is NULL, then saves the node at a free place.
	 else saves the node at address_on_disk.
*/

PRIVATE void
btdb_write_node(btdb_t *arbre, btree_node_t *t)
{
	int i;

	if (!t->is_on_disk)
		{
			/* create a disk record */
			t->is_on_disk= TRUE;
			t->address_on_disk= db0_init_record(&(arbre->tree_nodes));
		}
	else
		db0_open_record(&(arbre->tree_nodes), t->address_on_disk);

	/* writes node data */	
	db0_write_dbptr(&(arbre->tree_nodes), t->n);
	for (i= 0; i<t->n; i++)
		{
			db0_write_pstring(&(arbre->tree_nodes), &(t->elts[i]));
			db0_write_dbptr(&(arbre->tree_nodes), t->data[i]);
		}
	for (i= 0; i< t->n + 1; i++)
		{
			db0_write_dbptr(&(arbre->tree_nodes), t->ptrs[i]);
		}
}

/*
	Reads a node from the disk to memory
*/

PRIVATE void
btdb_read_node(btdb_t *arbre, btree_node_t *t, dbptr_t nodepos)
{
	int i;
	
	t->is_on_disk= TRUE;
	t->address_on_disk= nodepos;
	

	db0_open_record(&(arbre->tree_nodes), nodepos);

	t->n= db0_read_dbptr(&(arbre->tree_nodes));
	
	for (i= 0; i< t->n; i++)
		{
			db0_read_pstring(&(arbre->tree_nodes), &(t->elts[i]));
			t->data[i]= db0_read_dbptr(&(arbre->tree_nodes));
		}

	for (i= 0; i< t->n+1; i++)
		{
			t->ptrs[i]= db0_read_dbptr(&(arbre->tree_nodes));
		}
}



/*
	btdb_free_node_in_memory --
	frees the space used by the node
	substructures.
	Doesn't free the node itself and can
	thus be used on static variables
	*/

PRIVATE void
btdb_free_node_in_memory(btree_node_t *t)
{
	int i;
	for (i= 0; i< t->n; i++)
		{
			free_pstring_data(&(t->elts[i]));
		}
}


/*
 * Frees the space associated with a node on the disk
 * ONLY FREES THE TREE PART. DATA IS ANOTHER STORY
 */

PRIVATE void
btdb_free_node_on_disk(btdb_t *arbre, btree_node_t *t)
{
	if (t->is_on_disk)
		{
			db0_open_record(&(arbre->tree_nodes), t->address_on_disk);
			db0_free_current_record(&arbre->tree_nodes);
		}
}

/*
	Btree creation function.
  int
  btdb_open(btdb_t * tree,
					char *path,
					int flags,
					unsigned char baseblocksize,
					mode_t mode)
  -- result :
     tree is set for a database called after path.

	-- returns :
     0 if success, -1 if failure	 
*/

PUBLIC int
btdb_open(btdb_t * tree,
					char *path,
					int flags,
					unsigned char baseblocksize,
					mode_t mode)
{
	/* 
		 Should give a value to 
		 tree_nodes, 
		 base,
		 and root_node.
		 The database is made of two files :
		 one with the tree (.tr), the other with the data (.db)
	*/
	char * dbname, *treename;
	unsigned int taille_nom;
	int CREATION= FALSE;
	
	/* file names */
	taille_nom= strlen(path);
	dbname= (char *)ALLOUER(sizeof(char)*(taille_nom+4));
	treename= (char *)ALLOUER(sizeof(char)*(taille_nom+4));
	strcpy(dbname, path);
	strcpy(treename, path);
	strcat(dbname, ".db");
	strcat(treename, ".tr");

	/* normalise mode */
	flags &= (O_RDWR | O_RDONLY);

	/* If the files don't exit and O_RDWR is set,
		 we'll need to create them */

	CREATION= (flags  & O_RDWR);
	CREATION= CREATION && ( MyAccess(treename, F_OK)== -1);

	/* Open everyone */

	if (db0_open(&(tree->base), dbname, flags, baseblocksize, mode) == -1)
		return -1;

	if (db0_open(&(tree->tree_nodes), treename, flags, (unsigned char)TREEBLOCK, mode) == -1)
		return -1;

	tree->root_node= ROOTSIZE; /* not a very good thing. It would be better
																to store this somewhere */
	
	if (CREATION) {
			btree_node_t root_node;
			/* The empty tree has an empty node */
			btdb_init_new_node(tree, &root_node) ;
			/* root_node.n= 0; ALREADY DONE BY 
			 *              	 btdb_init_new_node which creates an empty node*/
			btdb_write_node(tree, &root_node);
	}
	LIBERE(treename);
	LIBERE(dbname);
 	return 0;
}

PUBLIC void
btdb_close(btdb_t * tree)
{
	db0_close(&(tree->base));
	db0_close(&(tree->tree_nodes));
}


/* Btree insertion function */

/*
	insert key and data at position  in current_node,
	moving the rest of the nodes if needed
*/


PRIVATE void
db_simple_insert_node(
											btree_node_t *current_node,
											int position,
											pstring * key,
											dbptr_t data,
											dbptr_t next
											)
{
	int j;
	for(j=current_node->n; j > position ; j--)
		{
			current_node->elts[j]= current_node->elts[j-1];
			current_node->data[j]= current_node->data[j-1];
			current_node->ptrs[j+1]= current_node->ptrs[j];			
		}
	transfer_pstring(&(current_node->elts[position]), key);
	current_node->data[position]= data;
	current_node->ptrs[position+1]= next;
	(current_node->n)++;
}

/* 
	 insert DATA into the btree_node t,
	 
	 explode is TRUE iff a new node must be inserted in the 
	 upper level node

	 In this case, 
	 new_node is the new node label
	 and
	 new is the new block.
	 */

/* PRE :
 *	   arbre is a valid tree
 *		 t points to a valid representation of an actual tree node
 *     data  is a valid pstring
 *		 key  is  a valid *DISCARDABLE* pstring
 *     explode points to a valid integer variable,
 *		 new_key etc... point to existing variables,
 *		 compare is a ordering function. The user-key is the FIRST argument 
 *	 
 * POST :
 *     key's content is garbage
 *     the parts of the tree including t are all safe and saved on disk.
 *     t still points to a valid representation of itself
 *     explode is 0 if t is not full and 1 if t was full
 *     if explode is true :
 *           t contains the first NNODES/2 keys it would have contained
 *           new_key is a key to be inserted in the node above t
 *           new_data is the corresponding data
 *           new_node_ptr is a pointer to the new tree just after new_key.
 */

PRIVATE void
btdb_insert_node_aux(
									 btdb_t * arbre,	/* the tree */
									 btree_node_t * t, /* the current node */
									 pstring *key, /* the key */
									 pstring *data,	/* the data */
									 int * explode,	/* did the tree explode ? */
									 pstring *new_key, /* the new node element, key part */
									 dbptr_t *new_data, /* the new node element, data part */
									 dbptr_t *new_node_ptr,	/* the new node */
									 int (*compare)(pstring*, pstring*)	/* comparison function */
									 )
{
	int i= 0;
	int comp_result;
	int T_SHOULD_BE_SAVED= FALSE;

	*explode= FALSE;

	/* 
		 we set comp_result to -1. 
		 It means that if t is empty, we will act as if 
		 the new element should be inserted.
		 actually, any non null value would do.
	*/

	comp_result= -1; 
	
	while (i< t->n && (comp_result= compare(key, &(t->elts[i]))) > 0)
		i++;

	/*
		Now, two cases :
		either we found our man (c==0)
		and then, all we need is saving the new data's value,
		or we didn't, in which case insertion is needed
	*/

	if (comp_result == 0)
		{
			/* t->data[i] points to the right data : */
			db0_open_record(&(arbre->base), t->data[i]);
			db0_write_pstring(&(arbre->base), data);
		}
	else
		{
			if (t->ptrs[i] == 0)
				{
					/* We are in a leaf */
					/* So : insert key,data, at position i in t,
						 with no subnode pointer
					*/
					dbptr_t data_on_disk;

					data_on_disk= db0_init_record(&(arbre->base));
					db0_write_pstring(&(arbre->base), data);
					db_simple_insert_node(t, i, key, data_on_disk, 0);
					T_SHOULD_BE_SAVED= TRUE;
				}
			else
				{
					btree_node_t next_explored_node;
					int sub_explo;
					pstring sub_key;
					dbptr_t sub_data;
					dbptr_t sub_node;
			
					btdb_read_node(arbre, &next_explored_node, t->ptrs[i]);
			
					btdb_insert_node_aux(arbre,
														 &next_explored_node,
														 key,
														 data,
														 &sub_explo,
														 &sub_key,
														 &sub_data,
														 &sub_node,
														 compare);
					if (sub_explo)
						{
							db_simple_insert_node(t, i, 
																		&sub_key, 
																		sub_data,
																		sub_node);
							T_SHOULD_BE_SAVED= TRUE;
						};

					btdb_free_node_in_memory(&next_explored_node);
				}
	
			/* Now, let's explode the node if needed.
				 If the node is full, it will part into three.
				 let's say 
				 t= [ P0 KEY0 P1 KEY1 P2 KEY2 P3]
				 Now, we have :
				 t1 = [ P0 KEY0 P1 ] (address P00)
				 t2 = [ P2 KEY2 P3 ] (address P01)
				 where KEY1 , P00, and P01 are values of
				 interest for our caller.

				 Now, if we are an internal node,
				 we can keep t = t1,
				 and it will be up to the caller to insert :
				 [......... P00 KEY1 P01 .....]
				 just BEFORE the key used in the caller

				 If t is the root node, 
				 t should be [P00 KEY1 P01],
				 and the two others should be new nodes

			*/
			if (t->n == NNODES + 1)
				{
					int j, k;
					btree_node_t new_node;

					*explode= TRUE;

					/* The central element will 
						 climb to the upper levels.
						 Note that since numbering starts with 0,
						 the central element's index  is NNODES/2
					*/
					*new_key= t->elts[NNODES/2];
					*new_data= t->data[NNODES/2];

					/* create a new block with the 
						 second half of the nodes.
						 This node is P01
					*/
					btdb_init_new_node(arbre, &new_node) ;

					new_node.n= NNODES/2;
			
					for(k=0, j= NNODES/2+1; j < NNODES+1; j++, k++)
						{
							new_node.elts[k]= t->elts[j];
							new_node.data[k]= t->data[j];
							new_node.ptrs[k]= t->ptrs[j];
						}
					new_node.ptrs[k]= t->ptrs[j];			
			
					btdb_write_node(arbre,&new_node);
					*new_node_ptr= new_node.address_on_disk;
					
					/* finally, set the current node's size */
					t->n= NNODES/2;

					btdb_free_node_in_memory(&new_node);

				}
	
			if (T_SHOULD_BE_SAVED)
				btdb_write_node(arbre,t);
		}
}



/* Entry point for the insertion function */ 

/*
 *	PRE : 
 *		a valid tree arbre,
 *		an allocated key,
 *  	an allocated data
 *	  a comparison function
 *
 *
 * POST :
 *
 *    key and data are STILL ALLOCATED
 *    arbre is valid
 */

PUBLIC void
btdb_insert_node(
							 btdb_t * arbre,
							 pstring * key,
							 pstring * data,
							 int (*compare)(pstring*, pstring*)
							 )
{
	int explode;
	btree_node_t t;
	pstring new_key; /* key for a possible new element from below*/
	dbptr_t new_data;	/* data for a possible new element */
	dbptr_t new_node_ptr;
	pstring key_copy;
	pstring data_copy;

	duplicate_pstring(&key_copy, key);
	duplicate_pstring(&data_copy, data);

	btdb_read_node(arbre, &t, arbre->root_node);

	btdb_insert_node_aux(arbre,
										 &t, 
										 &key_copy,
										 &data_copy,
										 &explode,
										 &new_key,
										 &new_data,
										 &new_node_ptr,
										 compare);

	if (explode)
		{
			int i;
			btree_node_t nouv;

			/* The situation was :
			 *
			 * [ P0 KEY0 P1 KEY1 P2 KEY2 P3]
			 * with everything full, and
			 * we got a new key.
			 * So what we have now is like :
			 * t = tree root = [P0 KEY0 P1]
			 * new_key = KEY1
			 * new_node= [P2 KEY2 P3]
			 *
			 * So we build :
			 *
			 *       [P00 KEY1 P01]
			 *        /          \  
			 *  [P0 KEY0 P1]   [P2 KEY2 P3]   
			 *
			 * the new root having the same address as
			 * the old one.
			 *
			 */

			/* Create a new node */
			btdb_init_new_node(arbre, &nouv) ;

			/* Copy the (exploded) old head in it */
			nouv.n= t.n;
			for (i= 0; i< t.n; i++)
				{
					nouv.elts[i]= t.elts[i];
					nouv.data[i]= t.data[i];
					nouv.ptrs[i]= t.ptrs[i];
				}
			nouv.ptrs[t.n]= t.ptrs[t.n];

			btdb_write_node(arbre, &nouv);
			/* The new head has only got one entry*/

			t.n= 1;		
			t.elts[0]= new_key;
			t.data[0]= new_data;
			t.ptrs[0]= nouv.address_on_disk;
			t.ptrs[1]= new_node_ptr;
			btdb_write_node(arbre, &t);
			btdb_free_node_in_memory(&nouv);
		}
	btdb_free_node_in_memory(&t);
}



PUBLIC void
btdb_insert_strings(
							 btdb_t * arbre,
							 char * key,
							 char * data
							 )
{
	pstring k, d;
	set_pstring_from_char(&k, key);
	set_pstring_from_char(&d, data);
	btdb_insert_node(arbre, &k, &d, compare_pstring);
}


/* Database dump with internal information */

#include <stdio.h>

PRIVATE void
btdb_dump_aux(btdb_t * arbre, dbptr_t ptr, int prof)
{
	if (ptr != 0) 
		{
			btree_node_t t;
			int i,j ;
	
			btdb_read_node(arbre,&t, ptr);
			for (j= 0; j< prof; j++)
				write(1,"*",1);
			write(1,"\n",1);

			for(i= 0; i<t.n; i++)
				{
					int j;
					btdb_dump_aux(arbre, t.ptrs[i], prof + 1);
					for (j= 0; j< prof; j++)
						write(1,">",1);
					write(1, t.elts[i].data, t.elts[i].length);
					write(1,"\n",1);
				}
			btdb_dump_aux(arbre, t.ptrs[i], prof + 1);
			for (j= 0; j< prof; j++)
				write(1,"*",1);
			write(1,"\n",1);

			btdb_free_node_in_memory(&t);
		}
}

/*
 * list free blocs in a plain database.
 */
PRIVATE void
db0_dump_struct(db0 *db, char * name)
{
	dbptr_t mybloc;
	mybloc= db->firstfree;
	printf("bloc table for %s\n", name);
	while ( mybloc!= 0)
		{
			db0_chose(db, mybloc);
			printf(" %d", (int)mybloc);
			mybloc= db->next;
		}
	printf("\n");
}

PUBLIC void
btdb_dump(btdb_t * arbre)
{
	btdb_dump_aux(arbre, arbre->root_node, 0);
	/* Information on .tr structure : */
	db0_dump_struct(&(arbre->tree_nodes), "tree");
	db0_dump_struct(&(arbre->base), "data");
}


/*
 * Apply a given function f to all nodes.
 *   arbre is the btree
 *	
 *   f is a function that will be called for all nodes, 
 *   	with the following arguments :
 *      key : a pstring, the key for the current record
 *      data : a pstring, the data for the current record
 *      external data : a pointer to the external data.
 *
 * 	external_data points to some arbitrary data, which f can use
 *    instead of using some global variable. It can of course be NULL
 *
 *  keyonly is a boolean. If it is TRUE, Then the data for the records
 *      won't be retrieved, and only the key will be valid.
 *      This is interesting if you are only interested in listing the 
 *      keys, for example.
 */


	
PRIVATE void
btdb_apply_aux(btdb_t * arbre,
							dbptr_t ptr,
							void (*f)(pstring *, pstring *, void *),
							void *external_data,
							int keyonly)
{
	if (ptr != 0) 
		{
			btree_node_t t;
			int i;
			
			btdb_read_node(arbre,&t, ptr);

			for(i= 0; i<t.n; i++)
				{
					pstring *dataptr;
					pstring data;
					
					btdb_apply_aux(arbre, t.ptrs[i], f, external_data, keyonly);
					if (!keyonly)
						{
							db0_open_record(&(arbre->base), t.data[i]);
							db0_read_pstring(&(arbre->base), &data);
							dataptr= &data;
						}
					else
						dataptr= 0;
					
					f(&(t.elts[i]), dataptr, external_data);
					
					if(!keyonly)
						free_pstring_data(&data);
				}
			btdb_apply_aux(arbre, t.ptrs[i], f, external_data, keyonly);
			btdb_free_node_in_memory(&t);
		}
}

PUBLIC void
btdb_apply(btdb_t * arbre,
					void (*f)(pstring *, pstring *, void *),
					void *external_data, 						
					int keyonly)

{
	btdb_apply_aux(arbre, arbre->root_node,
								f, external_data, keyonly);
}

/* Search function */
PUBLIC int
btdb_extended_find_node(
											btdb_t *arbre,
											pstring *key,
											pstring *data,
											int (*compare)(pstring*, pstring*)						 
											)
{
	int NOT_TROUVE= TRUE;
	int SOMETHING_LEFT= TRUE;
	btree_node_t current_node;

	btdb_read_node(arbre, &current_node, arbre->root_node);

	while (NOT_TROUVE && SOMETHING_LEFT)
		{
			int i;
			int comp=-1;

			i=0;
			while (i< current_node.n && (comp= compare(key, &(current_node.elts[i]))) > 0)
				i++;

			if (comp == 0)
				{
					NOT_TROUVE= FALSE;
					db0_open_record(&(arbre->base), current_node.data[i]);
					db0_read_pstring(&(arbre->base), data);
				}
			else {
				if (current_node.ptrs[i] == 0)
					SOMETHING_LEFT= FALSE;
				else
					{
						dbptr_t next_node;

						next_node= current_node.ptrs[i];
						btdb_free_node_in_memory(&current_node);
						btdb_read_node(arbre, &current_node, next_node);
					}
			}
		}
	btdb_free_node_in_memory(&current_node);
	return (!NOT_TROUVE);
}

PUBLIC int
btdb_find(btdb_t *arbre,
				pstring *key,
				pstring *data)
{
	return 
		btdb_extended_find_node(arbre, key, data, compare_pstring);
}


/* creates an empty index. Auxiliary function in any case */
PRIVATE btdb_index_t
btdb_index_empty_index(btdb_t * arbre)
{
	btdb_index_t res;
	res= (btdb_index_t)ALLOUER(sizeof(struct btdb_index));
	res->n= -1;										/* index of last elt ! */
	init_Chaine(&(res->s), sizeof(struct btdb_index_elt));
	res->arbre= arbre;
	return res;
}


PRIVATE int 
btdb_index_is_empty(btdb_index_t idx)
{
	return (idx->n == -1);
}


PRIVATE int 
btdb_index_is_last(btdb_index_t idx)
{
	int result;
	struct btdb_index_elt *elt;
	int i;

	elt = NULL;
	result= TRUE;
	
	/* All indexes part must point to elt->nd.n */
	for (i=0; result && i < size_of_Chaine(idx->s); i++) 
		{
			elt= (struct btdb_index_elt *) get_from_Chaine(idx->s, i);
			if (elt->i != elt->nd.n)
				result= FALSE;
		}
	
	/* The last part, if any, must be without subtree */
	if (result && elt != NULL && elt->nd.ptrs[elt->i] != 0)
		result= FALSE;
	
	return result;
}


PRIVATE int 
btdb_index_is_first(btdb_index_t idx)
{
	int result;
	struct btdb_index_elt *elt;
	int i;

	elt = NULL;
	result= TRUE;
	
	/* All indexes part must point to 0 */
	for (i=0; result && i < size_of_Chaine(idx->s); i++) 
		{
			elt= (struct btdb_index_elt *) get_from_Chaine(idx->s, i);
			if (elt->i != 0)
				result= FALSE;
		}
	
	/* The last part, if any, must be without subtree */
	if (result && elt != NULL && elt->nd.ptrs[0] != 0)
		result= FALSE;
	
	return result;
}

/* 
	 pops up an index level
*/

PRIVATE void
btdb_index_pop(btdb_index_t idx)
{
	struct btdb_index_elt *elt;
	elt= (struct btdb_index_elt *) get_from_Chaine(idx->s, idx->n);
	btdb_free_node_in_memory(&(elt->nd));
	(idx->n)--;
	restrict_Chaine_size_to(idx->s, idx->n + 1);
}

/* reads the informations about the last node */
PRIVATE struct btdb_index_elt *
btdb_index_read(btdb_index_t idx)
{
	return (struct btdb_index_elt *) get_from_Chaine(idx->s, idx->n);
}

/* pushes node. elt will be COPIED */

PRIVATE void
btdb_index_push(btdb_index_t idx, struct btdb_index_elt *elt)
{
	add_to_Chaine(idx->s, elt);
	(idx->n)++;
}

/* 
 * extends an index, pointing to the first element in the node
 * subtree 
 */
PRIVATE void
btdb_index_get_first(btdb_t * arbre,
										 dbptr_t node, 
										 btdb_index_t res)
{
	struct btdb_index_elt elt;
	do {
		btdb_read_node(arbre, &elt.nd, node);
		elt.i= 0;
		btdb_index_push(res, &elt);
		node= elt.nd.ptrs[0];
	} while (node != 0);
}

/* 
 * extends an index, pointing to the last element in the node
 * subtree 
 */
PRIVATE void
btdb_index_get_last(btdb_t * arbre,
										dbptr_t node, 
										btdb_index_t res)
{
	struct btdb_index_elt elt;
	do {
		btdb_read_node(arbre, &elt.nd, node);
		/* formerly : elt.i= elt.nd.n - 1;*/
		/* node= elt.nd.ptrs[elt.i + 1];*/
		
		/*
			NOTE The invariant on indexes :
			let pa be an index part, and pb the next one.
			Then pb.nd = pa.nd.ptrs[pa.i]
		*/ 

		/* 
		 * For the last node, i points to the last ELEMENT.
		 * For other nodes, i points to the last subtree.
		 * The index of the last ELEMENT is elt.nd.n - 1.
		 * the index of the last subtree is elt.nb.n
		 */

		node= elt.nd.ptrs[elt.nd.n];

		if (node == 0)
			elt.i= elt.nd.n - 1;
		else
			elt.i= elt.nd.n;

		btdb_index_push(res, &elt);
	} while (node != 0);
}


/*
 * btdb_first_index --
 * sends back an index to a tree's first element.
 */

PUBLIC btdb_index_t
btdb_first_index(btdb_t * arbre)
{
	btdb_index_t res;
	res= btdb_index_empty_index(arbre);
	btdb_index_get_first(arbre, arbre->root_node, res);
	return res;
}

/*
 * btdb_last_index --
 * sends back an index to a tree's last element.
 */

PUBLIC btdb_index_t
btdb_last_index(btdb_t * arbre)
{
	btdb_index_t res;
	res= btdb_index_empty_index(arbre);
	btdb_index_get_last(arbre, arbre->root_node, res);
	return res;
}

/*
 * btdb_free_index --
 * frees an index data structure.
 */
PUBLIC void
btdb_free_index(btdb_index_t idx)
{
	unsigned int i;
	for (i= 0; i < size_of_Chaine(idx->s); i++)
		{
			struct btdb_index_elt *elt;
			elt= (struct btdb_index_elt *) get_from_Chaine(idx->s, i);
			btdb_free_node_in_memory(&(elt->nd));
		}
	free_Chaine(idx->s);
}

/*
 * btdb_index_to_value --
 * given a tree and an element index,
 * sends back the element's value.
 * Both key and data are freshly allocated memory, which should
 * be freed by the user.
 * if data is NULL, no data will be read.
 *
 */

PUBLIC void
btdb_index_to_value(btdb_index_t idx,
										pstring *key,
										pstring *data)
{
	struct btdb_index_elt *elt;
	elt= (struct btdb_index_elt *) get_from_Chaine(idx->s, idx->n);
	duplicate_pstring(key, &(elt->nd.elts[elt->i]));
	db0_open_record(&(idx->arbre->base), elt->nd.data[elt->i]);
	if (data != NULL)
		db0_read_pstring(&(idx->arbre->base), data);
}

/*
 * btdb_next_index --
 * given a tree and an element index,
 * sets the index to point to the next element.
 * returns 1 if success, 0 if there is no next element.
 * If the operation is unsuccessful, the index is still usable. It simply doesn't move
 */

PUBLIC int
btdb_next_index(btdb_index_t idx)
{
	struct btdb_index_elt *p;
	dbptr_t next_tree;
	int result= TRUE;
	
	if (btdb_index_is_last(idx))
		return FALSE;

	p= btdb_index_read(idx); 
	p->i++;												/* horizontal move */
	/* first elt in next subtree, if there is a next subtree */
	next_tree= p->nd.ptrs[p->i];
	if ( next_tree != 0)
		{
			btdb_index_get_first(idx->arbre, next_tree, idx);
		}
	else													/* no next subtree */
		{
			int is_empty= FALSE;
			/* go up until  we get something */
			while (! is_empty					/* we can go up */
						 && p->i == p->nd.n) /* and the whole node has been explored */
				{
					btdb_index_pop(idx);
					is_empty= btdb_index_is_empty(idx);
					if (! is_empty) 
						p= btdb_index_read(idx);
				}
			if (is_empty)
				result= FALSE;
		}
	return result;
}


/*
 * btdb_previous_index --
 * given a tree and an element index,
 * sets the index to point to the previous element.
 * returns 1 if success, 0 if there is no previous element.
 *
 * AT THE TIME BEING : CALLS next_index !
 */

PUBLIC int
btdb_previous_index(btdb_index_t idx)
{
	struct btdb_index_elt *p;
	dbptr_t previous_tree;
	int result= TRUE;

	if (btdb_index_is_first(idx))
		return FALSE;

	p= btdb_index_read(idx);
	/* last elts in previous subtree */
	previous_tree= p->nd.ptrs[p->i];
	if ( previous_tree != 0)
		{
			btdb_index_get_last(idx->arbre, previous_tree, idx);
		}
	else													/* no previous subtree */
		{
			int is_empty= FALSE;
			/* go up until  we get something */
			while (! is_empty					/* we can go up */
						 && p->i == 0) /* and the whole node has been explored */
				{
					btdb_index_pop(idx);
					is_empty= btdb_index_is_empty(idx);
					if (! is_empty) 
						p= btdb_index_read(idx);
				}
			if (is_empty)
				result= FALSE;
			else
				p->i --;
		}
	return result;
}

/*
 * 
 * btdb_get_index_to --
 * gets an index to the first element greater or equal 
 * to the key.
 *
 * returns 1 if such an element exists, else 0
 */

PUBLIC int
btdb_get_index_to(btdb_t *arbre,
									btdb_index_t *idxp,
									pstring *key
									)
{
	int NOT_TROUVE= TRUE;
	int SOMETHING_LEFT= TRUE;
	int result= FALSE;
	struct btdb_index_elt elt;
	
	*idxp= btdb_index_empty_index(arbre);

	
	/* First, let's start looking for key ---
		 we'll get an index. 
		 Then, we shall see where we have gone :
		 if the current element has the key, so far so good.
		 Else, the good element is the next one.
	*/

	btdb_read_node(arbre, &(elt.nd), arbre->root_node);
	elt.i= 0;

	while (NOT_TROUVE && SOMETHING_LEFT)
		{
			int comp=-1;

			/* look for the place for our element */
			while (elt.i< elt.nd.n 
						 && (comp= compare_pstring(key, &(elt.nd.elts[elt.i]))) > 0)
				elt.i++;

			/* saves it on the index stack */
			btdb_index_push(*idxp, &elt);

			if (comp == 0)
				NOT_TROUVE= FALSE;
			else if (elt.nd.ptrs[elt.i] == 0)
				SOMETHING_LEFT= FALSE;
			else
				{
					dbptr_t next_node;
					
					next_node= elt.nd.ptrs[elt.i];
					btdb_read_node(arbre, &elt.nd, next_node);
					elt.i= 0;
				}
		}

	/* The element we point to (if any) is normally the biggest element
		 smaller than our key */
	
	if (NOT_TROUVE == FALSE)
		result= TRUE;
	else if (btdb_index_is_empty(*idxp))
		result= FALSE;
	else {
		struct btdb_index_elt *p;
		int EMPTY= FALSE;
		/* In the case where the element was not found,
			 we have to get sure we are at a good point.
			 It means that elt.i must point to a valid place,
			 i.e. not be equal to elt.nd.n 
		*/
		p= btdb_index_read(*idxp);
		while(! EMPTY && p->i == p->nd.n)
			{
				btdb_index_pop(*idxp);
				EMPTY= btdb_index_is_empty(*idxp);
				if (!EMPTY)
					p= btdb_index_read(*idxp);
			}
		if (EMPTY)
			result= FALSE;
		else
			result= TRUE;
	}
	if (! result)
			btdb_free_index(*idxp);
	return result;
}



/*
 * Now, a rather tricky part :
 * elements deletion
 */

/*
 * db_simple_delete_node --
 *
 *	actual deletion of a pointer in a node
 *
 * arguments --
 *	current_node : 
 *  position :
 *  key
 *  data
 *  next
 */
	
PRIVATE void
db_simple_delete_node(
											btree_node_t *current_node,
											int position
											)
{
	int j;
	for(j=position; j < current_node->n; j++)
		{
			current_node->elts[j]= current_node->elts[j+1];
			current_node->data[j]= current_node->data[j+1];
			/* In theory, we are in a leaf. So no need to modify  */
			/* current_node->ptrs[] */
		}
	(current_node->n)--;
}



/***
 * btdb_find_lesser --
 *
 * finds element just lesser (or greater) than the numth element in a given node
 * This is an auxiliary function in the deletion process.
 * We should probably try to unify the various research functions
 * used in the code.
 *
 * PRE : father should not be a leaf
 * 
 * arbre : the btree
 * father : the element from which to start the search
 * num : the index for the element in father
 * resultNode : a btree_node_t element which will be filled if something is found
 * res: will contain the disk address for the element.
 * pos : will contain the position of the element in the node
 * compare : comparison function
 * WHAT : find lesser or greater element
 *
 */

PRIVATE void
btdb_find_lesser(
								 btdb_t *arbre,
								 btree_node_t *father,											
								 int num,
								 btree_node_t *resultNode,											
								 int * pos
								 )
{
	dbptr_t nextNode;

	/* 
		 father is the current node. We need 
		 to go down the tree in order to find the next element.
	*/
	nextNode=  father->ptrs[num];

	btdb_read_node(arbre, resultNode,nextNode);
	while (resultNode->ptrs[resultNode->n] != 0)
		{
			nextNode= resultNode->ptrs[resultNode->n];
			btdb_free_node_in_memory(resultNode);
			btdb_read_node(arbre, resultNode,nextNode);
		}
	*pos= resultNode->n - 1;
}



/* 
	 Merges the two nodes around position t
 */
PRIVATE void
btdb_merge_nodes(btdb_t * arbre,
								 btree_node_t * t,
								 int pos,
								 btree_node_t * left,
								 btree_node_t * right)
{
	int i, j;
	/* Everything goes in left : */
	/* first, the central element */
	i= left->n;
	left->elts[i]= t->elts[pos];
	left->data[i]= t->data[pos];
	left->n++;
	/* The pointer at left->ptrs[i] is already here*/

	for (i=left->n, j=0; j < right->n; i++, j++, left->n++)
		{
			left->elts[i]= right->elts[j];
			left->data[i]= right->data[j];
			left->ptrs[i]= right->ptrs[j];
		}

	left->ptrs[i]= right->ptrs[j];
	/* Now, let's update t : */
	for(i= pos; i< t->n-1; i++)
		{
			t->elts[i]= t->elts[i+1];
			t->data[i]= t->data[i+1];
			t->ptrs[i+1]= t->ptrs[i+2];
		}	
	t->n--;
	/* should we suppress right ? */
	right->n= 0;
	btdb_write_node(arbre,t);
	btdb_write_node(arbre,left);
}



/*
 * btdb_delete_node_aux --
 * 
 * Auxiliary function for entry deletion in a btree
 * Returns 0 or the address for the data part of the entry.
 *
 */


PRIVATE dbptr_t 
btdb_delete_node_aux(
									 btdb_t * arbre,	/* the tree */
									 btree_node_t * father, /* the current node's father */
									 int pos_in_father, /* the current node position in its father */
									 /* btree_node_t * t,  the current node */
									 dbptr_t addr,
									 pstring *key, /* the key */
									 int (*compare)(pstring*, pstring*)	/* comparison function */
									 )
{
	int i= 0;
	int comp_result;
	int T_SHOULD_BE_SAVED= FALSE;
	btree_node_t t;
	dbptr_t result= 0;
	
	/* search element to delete */
	btdb_read_node(arbre, &t, addr);
	
	comp_result= -1; 
	
	while (i< t.n && (comp_result= compare(key, &(t.elts[i]))) > 0)
		i++;

	/*
		Now, two cases :
		either we found our man (c==0)

		and thus we need to destroy it,
		or we didn't, in which case deletion will be done on other nodes
	*/
	
	if (comp_result == 0)
		{
			T_SHOULD_BE_SAVED= TRUE;
			result= t.data[i];
			if (t.ptrs[i] == 0)
				{
					/* CASE A) We are in a leaf.
					*  We suppress the entry
					*  Maybe we'll need to balance the tree later*/
					db_simple_delete_node(&t, i);
				}
			else
				{
					/* CASE B) We are not in a leaf */
					/* get the element immediately lesser than key */
					/*
					 *	Note : this is always possible. If there is no element immediately lesser 
					 *  than the current one, then we are in a leaf (hence, case A)
					 */
					btree_node_t leafnode;
					int pos;
					btdb_find_lesser(arbre, &t, i, &leafnode, &pos);
					/* we want to destroy the current element. */
					/* So we replace it element by the immediately smaller one */
					free_pstring_data(&(t.elts[i]));
					duplicate_pstring(&(t.elts[i]), &(leafnode.elts[pos]));
					t.data[i]= leafnode.data[pos];
					/* we don't need leafnode anymore in memory. */
					/* let's free the memory it uses */
					btdb_free_node_in_memory(&leafnode);
					/* Now, the smaller element replaces the destroyed element.
					 * But this smaller element also appears at its previous position
					 * in the tree. So, we destroy it there 
					 * IMPORTANT NOTICE :
					 *
					 *   This recursive use of btdb_delete_node_aux *FORBIDS US*
					 *   to delete the data part for the node in the
					 *   btdb_delete_node_aux function.
					 * 
					 */
					btdb_delete_node_aux(arbre, &t, i, t.ptrs[i], &(t.elts[i]), compare);
				}
		}
	else
		{
			/* We haven't found the element. */
 			if (t.ptrs[0] != 0)
				{
					/* this is not a leaf */
					/* i contains the index for the first element greater than key */
					/* hence, we should explore the t.ptrs[i] subtree */
					result= btdb_delete_node_aux(arbre, &t, i, t.ptrs[i], key, compare);
				}
			else
				{
					/* this is a leaf, suppression is a failure */
					result= 0;
				}
		}

	/* Now, if this node is not the root node and 
		 doen't contains enough elements,
		 redistribute or merges its elements
	*/
	
	if (result != 0)
		{
			if ( father != 0 && t.n < NNODES / 2)
				{
					int LEFT_BROTHER_EXISTS= FALSE;
					int RIGHT_BROTHER_EXISTS= FALSE;
					btree_node_t left_brother, right_brother;
			
					LEFT_BROTHER_EXISTS= (pos_in_father > 0);
					if (LEFT_BROTHER_EXISTS)
						{
							btdb_read_node(arbre, &left_brother, father->ptrs[pos_in_father-1]);
						}
					RIGHT_BROTHER_EXISTS= (pos_in_father< father->n);
					if (RIGHT_BROTHER_EXISTS)
						{
							btdb_read_node(arbre, &right_brother, father->ptrs[pos_in_father+1]);
						}
					/* if we have enough elements in the left brother, 
						 let's take one of them
					*/
					if (LEFT_BROTHER_EXISTS && left_brother.n > NNODES / 2)
						{
							int j;
							/* we take an element in father,
								 which goes then first in current node.*/
							/* Note : we can simply exchange pointers, 
								 no need for copying here */
							for(j=t.n; j>0; j--)
								{
									t.elts[j]= t.elts[j-1];
									t.data[j]= t.data[j-1];
									t.ptrs[j+1]= t.ptrs[j];
								}
							t.ptrs[1]= t.ptrs[0];
							/* translation done */
							t.elts[0]= father->elts[pos_in_father-1];
							t.data[0]= father->data[pos_in_father-1];
							t.ptrs[0]= left_brother.ptrs[left_brother.n];
							t.n++;
							/* Now, t is good */
							/*
								Then we replace this element in father
								by the biggest element in left_brother */
							father->elts[pos_in_father-1]= left_brother.elts[left_brother.n-1];
							father->data[pos_in_father-1]= left_brother.data[left_brother.n-1];
							/* Finally, we modify left_brother : */
							left_brother.n--;
							/* Save left brother now, and free the memory it uses */
							btdb_write_node(arbre,&left_brother);
							/* father should also be saved */
							btdb_write_node(arbre,father);
							T_SHOULD_BE_SAVED= TRUE;
						}
					else if (RIGHT_BROTHER_EXISTS && right_brother.n > NNODES / 2)
						{
							int j;
							/* else if we have enough elements in the right brother,
								 let's take one of them : */
							/* we take an element in father, which goes last in t */
							t.elts[t.n]= father->elts[pos_in_father];
							t.data[t.n]= father->data[pos_in_father];
							t.ptrs[t.n+1]= right_brother.ptrs[0];
							t.n++;
							/* Now, t is good */
							father->elts[pos_in_father]= right_brother.elts[0];
							father->data[pos_in_father]= right_brother.data[0];
							/* father is fine */
							/*Now, we suppress the FIRST element of right_brother : */
							for(j=0; j<right_brother.n-1; j++)
								{
									right_brother.elts[j]= right_brother.elts[j+1];
									right_brother.data[j]= right_brother.data[j+1];
									right_brother.ptrs[j]= right_brother.ptrs[j+1];							
								}
							right_brother.ptrs[right_brother.n-1]= 
								right_brother.ptrs[right_brother.n];
							right_brother.n--;
							/* OK. Everyone is fine now */
							/* Save right brother now, and free the memory it uses */
							btdb_write_node(arbre,&right_brother);
							/* father should also be saved */
							btdb_write_node(arbre,father);
							T_SHOULD_BE_SAVED= TRUE;					
						}
					else if (LEFT_BROTHER_EXISTS)
						{
							/* else if there is a left brother, merge with it */
							btdb_merge_nodes(arbre, father, pos_in_father - 1, &left_brother, &t);
						}
					else if (RIGHT_BROTHER_EXISTS)
						{
							/* else if there is a right brother, merge */
							btdb_merge_nodes(arbre, father, pos_in_father, &t, &right_brother);
						}
					else
						{
							/* else ? impossible */
							exit (1);
						}
			
					if (LEFT_BROTHER_EXISTS)
						btdb_free_node_in_memory(&left_brother);
					if (RIGHT_BROTHER_EXISTS)
						btdb_free_node_in_memory(&right_brother);
				}
			else if (father == 0 && t.n == 0)
				{
					/* All nodes should have at least NNODES / 2 nodes,
						 except the root node (obviously, if there is only one
						 entry in the tree, there is only the root node, and it contains less than
						 NNODES / 2 nodes)

						 Now, the rule node should not be empty, except if the whole tree is empty.
						 Hence, if 
						 a) the rule node is empty. 
						 b) there is a node N below it (at t.ptrs[0])
						 this node N should become the root node.

						 IMPORTANT NOTE : for this to function, t.ptrs[0] must be accurate.
						 That is, for an empty tree, t.ptrs[0] == 0.
					*/
					if (t.ptrs[0] != 0)
						{
							btree_node_t new_root;
							dbptr_t old_addr;
							old_addr= t.address_on_disk;
							/* get the child node */
							btdb_read_node(arbre, &new_root, t.ptrs[0]);
							/* free the space used for child node on disk */
							btdb_free_node_on_disk(arbre, &new_root);
							/* copy it at address 0 */
							t= new_root;
							t.address_on_disk= old_addr;
							T_SHOULD_BE_SAVED= TRUE;
						}
				}
		}

	if (T_SHOULD_BE_SAVED)
		btdb_write_node(arbre,&t);
	btdb_free_node_in_memory(&t);
	return result;
}


/* Entry point for the deletion function */ 

/*
 *	PRE : 
 *		a valid tree arbre,
 *		an allocated key,
 *	  a comparison function
 *
 *
 * POST :
 *
 *    key is STILL ALLOCATED
 *    arbre is valid
 *    if key was found, its entry is suppressed in the tree.
 *    the data associated with key is also suppressed.
 */

PUBLIC void
btdb_delete_node(btdb_t * arbre,
								 pstring * key,
								 int (*compare)(pstring*, pstring*))
{
	dbptr_t data= btdb_delete_node_aux(arbre,
																		 0, 
																		 0,
																		 arbre->root_node,
																		 key,
																		 compare);
	if (data != 0)
		{
			db0_open_record(&(arbre->base), data);
			db0_free_current_record(&arbre->base);
		}
}

PUBLIC void
btdb_delete_string(btdb_t * arbre,
									 char * key)

{
	pstring k;
	set_pstring_from_char(&k, key);
	btdb_delete_node(arbre, &k, compare_pstring);
}

/*
 * Local Variables:
 * tab-width: 2
 * End:
 */

