#ifndef _LIGATUR_H
#define _LIGATUR_H

#include "general.h"
/*>>>>ligatur.c */

/**
  Ligatures now work this way :
  
  A ligature is defined with a number of parameters :
  *****
  size : dx dy,
  and a list of signs and sign position : SIGN x y size
  *****
  
  The definitions is done this way :
  l= ligature_create(dx,dy,nbrsigns);
  and a number of times :
  ligature_subsign_init(l, num, code, x, y, size)
  finish with :
  ligature_make_ident(l);
	and 
  ligature_complete(l); if you are interested in storing the ligature.

  NOTA : code is a char *, 
  but there's no need for it to be allocated to a permanent storage.
  
  ****
  ligature_delete(l)
  will delete a ligature
  ****
  
  Ligature usage :
  To get a ligature data, one must have built a ligature vector.
  
  The commands for this are :
  c= ligcode_create();
  a number of :
  ligcode_add2(c, code);
  ...
  
  A ligcode can be destroyed by ligcode_delete(c);

  Now, one can compare a ligature structure and 
  a ligcode one by getting their codes :
  
  X= ligcode_to_code(c)
  X= ligature_to_code(l)
  */

struct Ligature {
  char *code;		/* computed identifier for ligature */
  int codesize;
  double w, h;		/* ligature dimensions */
  int nbr;			/* number of signs in Ligature */
  struct sign_ligpos {
    Lettre *l;			/* sign letter*/
    double x,y;
    double size;
  } *signs;			/* table of signs characteristics */
};

/*
  The definitions is done this way :
  l= ligature_create(dx,dy,nbrsigns);
	
	defines a new ligature of dim dx*dy containing nbrsigns signs
*/

PUBLIC struct Ligature *
ligature_create(int nbrsigns);

/*
	ligature_subsign_init(l, num, code, x, y, size)
	
	the num^th sign for l is code, at position x,y, size % of the
	original sign.

		returns TRUE if initialization was successful, else FALSE
*/

PUBLIC int
ligature_subsign_init(struct Ligature *l, int num, 
		      char *code, double x, double y, double size);

/*
	  ligature_make_ident(l);
		make the ligature unique identifier valid
*/
PUBLIC void
ligature_make_ident(struct Ligature *l);

/*
	  ligature_complete(l);
		makes the ligature ready to use.
		should be called for real ligatures,
		not for temporary ligatures created for searching.
*/

PUBLIC void
ligature_complete(struct Ligature *l);

/*
	ligature_delete(l)
*/
PUBLIC void
ligature_delete(struct Ligature *l);

/*
  X= ligature_to_code(l)
*/

PUBLIC char *
ligature_to_code(struct Ligature *l);

/*<<<<*/

#endif
/* Let's preserve the present presentation
 * Local Variables:
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
