

/***************************************************************************\
* 	      Definition du texte au niveau le plus haut                       *
*
* Note on regions :
*
* Some commands have an effect only on one textual item. For these,
* the meaning of the GeneralIndex is straightforward.
*
* Others affect regions of text. They use both starting and ending marks.
* The meaning of these marks is the following :
* the affected area begins AFTER the starting marks, and spans text up to,
* and INCLUDING the ending mark.
*
\***************************************************************************/

/*
  Un texte est une suite de quadras.
*/

#ifndef _TEXTE_H
#define _TEXTE_H

#include "quadra.h"
#include "mdctypes.h"

/**
 *
 * Types declared :
 * 
 * text_coords : textual description of coordinates
 * Marque : marks
 * Liste_marques : list of marks
 *
 * Texte A textual item : A cadrat in a text.
 * 
 * SignePix : used to manipulate the list of signs in a text
 */


/*
 * For Text Manipulation
 */

#define EST_GRIS(x) ((x) && ((x)->q) && ((x)->q->est_grise))
#define SET_GRIS(x) ((x) && ((x)->q) && ((x)->q->est_grise = TRUE))
#define UNSET_GRIS(x) ((x) && ((x)->q) && ((x)->q->est_grise = FALSE))
#define EST_ROUGE(t) ((t)->couleur == red)

/*
 * For GeneralIndex manipulation :
 *
 * GETTEXTOF(id) => Texte * associated to id
 * GETSIGNOF(id) => Signe * associated with id
 *
 */

#define GETTEXTOF(id) ((id).t)
#define GETSIGNOF(id) ((id).signe)
#define GETSIGNPOSOF(id) ((id).s)

/*
 * Fonctions publiques 
 */

/*>>>>MDCtexte.c */

/**
 *
 * PUBLIC Texte * 
 * NewTexte()
 * 
 */

PUBLIC Texte * 
NewTexte();

/**
 */

PUBLIC void
DeleteTexte(Texte *tt);

/**
   PUBLIC void
   FreeTexte(Texte *tt)
   
   completely frees a text. Do not use to delete a text.
   Only to free the text when deleting a sesh widget.
*/

PUBLIC void
FreeTexte(Texte *tt);

/**
 */
PUBLIC Texte *
MDC_FirstTexte();

/**
 */
PUBLIC void
add_page_after(Texte *txt);

/**
 * PUBLIC void 
 * SESH_ordonner_Texte(Texte **a, Texte **b)
 * 
 * ensure that 'a' is before 'b' in the text, swapping them
 * if needed.
 */

PUBLIC void 
SESH_ordonner_Texte(Texte **a, Texte **b);

/**
 * PUBLIC Texte *
 * detruire_Texte(Texte *debut, Texte *fin)
 * 
 * destroys the text between two Texte;
 * it returns a  safe spot in front of the destroyed zone.
 * The first text (be it debut or fin) is NOT destroyed.
 * graphically :
 *
 * debut                  fin
 *     |..................|
 *      *******************
 *      ^^^^^^^^^^^^^^^^^^^
 *        this is deleted 
 * The rationale is that a zone between Texte * 
 * designates text AFTER the first one up to (and including
 * the last one.
 */

PUBLIC Texte *
detruire_Texte(GeneralIndex gdebut, GeneralIndex gfin);

/**
 * PUBLIC Texte *
 * ajoute_texte_apres(Texte *montexte)
 **	ajoute du texte aprs un autre texte. Si le second argument est la
 ** ligne courante (et non NULL) alors, si le nouveau texte rajout
 ** devient la nouvelle fin de ligne, cela est gr. 
 *********************************************************************/

PUBLIC Texte *
ajoute_texte_apres(Texte *montexte);

/**
 */
PUBLIC void
SESH_init_textes();


/**
	Inserts text after gori.
 */

PUBLIC Texte *
SESH_insere_texte(GeneralIndex gori, Texte *insere);

/**
	Sends back a string representing g where the uppermost
	construct has been broken. Former separator are replaced by '\n'
	(newline).
	e.g. : 
     pt:p*t => p\np*t
	   p*t => p\nt
		 <-ra-ms-s-s-> => ra*ms*s*s
 */

PUBLIC char *
SESH_breaktext(GeneralIndex g);

/**
	sends back a string representing the text between debut and fin
	with Text separators ('-') replaced by newlines.
	*/

PUBLIC char *
SESH_breakspan(GeneralIndex debut, GeneralIndex fin);

/**
	returns true if g is a pile of quadras,
	and else, false.
	BEWARE :
	  mn:n*w is a pile,
		mn*(n:t) is not !
 */
	
PUBLIC int
SESH_is_vertical(GeneralIndex g);

/**
 */
PUBLIC char *
SESH_texte_to_string(GeneralIndex debut, GeneralIndex fin);

/**
	PUBLIC char *
	SESH_listesignes(GeneralIndex gdebut, GeneralIndex gfin, int mode)
	
	List of signs between gdebut and gfin
	mode allows to get more information. Information for each
	sign is given as a list, the first element of whom is the 
	sign's code.
	
	mode :
	  SIGNS_ONLY => list of the signs
		SIGNS_EOW  => for each sign which is a word ending, indicates it by 
		              appending the tag "WORD" after the sign.
		SIGNS_GRAM => append the tag "GRAM" if sign belongs to a grammatical ending
		
		the order of the tags is always 
		    SIGN ["GRAM"] ["WORD"]

	note: if SIGNS_IGNORE is not there, the sign will simply be ignored !
 */

#define SIGNS_ONLY 0
#define SIGNS_EOW (1)
#define SIGNS_GRAM (1<<1)
#define SIGNS_IGNORE (1<<2)

PUBLIC char *
SESH_listesignes(GeneralIndex gdebut, GeneralIndex gfin, int options);

/**
 */
PUBLIC int
SESH_Texte_est_hieroglyphe(Texte *t);

/**
 */
PUBLIC void 
SESH_set_Texte_texte(Texte *t, char * texte);

/**
 *
 * PUBLIC int
 * SESH_nombres_de_signes_dans_Texte(Texte *t);
 *
 * returns the number of signs in a quadra
 */

PUBLIC int
SESH_nombres_de_signes_dans_Texte(Texte *t);

/**
 * Sends back a SignePix handler to the first  sign
 * RETURNS TRUE if success, else FALSE 
 */

PUBLIC int
SESH_premier_signe(SignePix *s, Texte *t);

/**
 * Sends back a SignePix handler to the nth sign
 * RETURNS TRUE if success, else FALSE 
 */

PUBLIC int
SESH_nieme_signe(SignePix *s, Texte *t, int n);

/**
 * Have p point to the next sign
 * RETURNS TRUE if success, else FALSE 
 */

PUBLIC int
SESH_signe_suivant(SignePix *p);

/**
 *
 * BEWARE : the value returned should NOT be written to
 */

PUBLIC char *
SESH_get_signe_rep_from_pix(SignePix p);

/*********************************************************************
 *   information about individual signs 
 *
 */

/**
 * SESH_set_sign_property --
 * set or toggle the value of a sign's property.
 *
 * GeneralIndex g : the sign
 * Signproperty prop : the property to change
 * what : the property new value (an integer). If the property is boolean,
 *        -1 means "toggle"
 */
PUBLIC void
SESH_set_sign_property(GeneralIndex g, Signproperty prop, int what);

/**
	 get a sign's property value
 */

PUBLIC int
SESH_sign_get_p(GeneralIndex g, Signproperty prop);


/**
 */
PUBLIC void
SESH_set_selected(GeneralIndex gdebut, GeneralIndex gfin, int selectedp);

/**
 * public GeneralIndex 
 * SESH_make_general_index_to(Texte *t, int signe)
 * 
 * Sends back an GeneralIndex to a position.
 * t is the cadrat
 * signe is the number of the sign in t, 
 *       0 if t as a whole is meant.
 */

PUBLIC GeneralIndex 
SESH_make_general_index_to(Texte *t, int signe);

/**
 */
PUBLIC int
SESH_not_last_index(GeneralIndex g);

/**
	moving in the Text
 */
PUBLIC GeneralIndex 
SESH_next_index(GeneralIndex g);

/**
	moving in the Text
 */
PUBLIC GeneralIndex 
SESH_previous_index(GeneralIndex g);

/*<<<<*/

/*
  Un mot:
  entree(texte)
  prononciation
  
  les TEXTES dirigent les mots. i.e.
  chaque texte contient un pointeur vers :
  * le premier mot du quadra,
  * le dernier mot du quadra.
  le texte lui-meme est compose' par la se'rie de caracteres tapes
  pour obtenir ledit texte, plus une ``prononciation'', laquelle pourra 
  etre changee en cas de selection.

  est-ce la peine pour l'instant de les taper?
*/


#endif

