/*
	 Interface graphique de TKSESH.
	 Version $Id: tgraph.h,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $	
	 
	 Il suffira en thorie de la r-crire pour obtenir
	 une sortie diffrente (exemple intressant : 
	 criture dans un fichier.)
	 
	 
*/

#ifndef TKSESHGRAPHE_H
#define TKSESHGRAPHE_H

#include <tk.h>
#include "chaine.h"


#define TKSOC(x) ((TKSESH_output*)(x))

typedef enum 
{ 
	white,												/* erase */
	black,												/* normal drawing color */
	gram,													/* color for grammar */
	blue,													/* word ending */
	gramend,											/* grammar + end */
	aux,														/* for auxiliary drawings */
	red,													/* red */
	ignore,													/* colour of ignored  signs */
	grayfill,											/* gray filling */
	invert												/* inverting */
} TKSO_state;

/*
 * Available width for line drawing.
 */

typedef enum {
	nullWidth,
	thinWidth,
	wideWidth,
	extraWideWidth
} TKSO_lineWidth;

typedef struct 
{
	double x; 
	double y;
} TKSESH_point ;


/*
	Only init rectangles through 
	TKSO_init_rectangle(double llx, double lly, double hrx, double hry)
*/

typedef struct
{
	TKSESH_point bottom_left, top_right;
	/* double llx, lly, urx, ury; */
} TKSESH_rectangle;

typedef struct TKSESH_output_struct {
	struct TKSO_device_struct * driver;
	/* Geometry */
	TKSESH_rectangle window;
	double scale;
	int w, h;
	/* Colour */
	int reverse;
	TKSO_lineWidth lineWidth;
	TKSO_state state;
	/* Various capability indicators : */
	int isLowRes;									/* true for low-res devices */
	int hasSplines;								/* true if the system has a native spline drawing command */
	int useFonts;									/* true if fonts are available for texts */
	/* Here should come the real data */
} TKSESH_output;


/*
 * Definition of types for the implementation functions
 */

typedef void (*TKSO_cleanP)(TKSESH_output *);
typedef void (*TKSO_DrawLineP)(TKSESH_output *, TKSESH_point, TKSESH_point);
typedef void (*TKSO_DrawLinesP)(TKSESH_output *, int, TKSESH_point *);
typedef void (*TKSO_hachureP)(TKSESH_output *, TKSESH_rectangle rect);
typedef void (*TKSO_fillPolygonP)(TKSESH_output *, int , TKSESH_point * , int);
typedef void (*TKSO_DrawArcP)(TKSESH_output *, TKSESH_point, double, double, int, int);
typedef double (*TKSO_String_widthP)(TKSESH_output *, char *);
typedef void (*TKSO_DrawStringP)(TKSESH_output *, TKSESH_point, char *);
typedef void (*TKSO_fill_rectangleP)(TKSESH_output *, TKSESH_rectangle);
typedef void (*TKSO_set_stateP)(TKSESH_output *o, TKSO_state s);
typedef void (*TKSO_set_lineWidthP)(TKSESH_output *o, TKSO_lineWidth s);


typedef struct TKSO_device_struct {
	char * name;
	int TKSOSize;									/* Size of the particular TKSO device */
	TKSO_cleanP cleanP;
	TKSO_DrawLineP DrawLineP;
	TKSO_DrawLinesP DrawLinesP;
	TKSO_hachureP hachureP;
	TKSO_fillPolygonP fillPolygonP;
	TKSO_DrawArcP DrawArcP;
	TKSO_String_widthP String_widthP;
	TKSO_DrawStringP DrawStringP;
	TKSO_fill_rectangleP fill_rectangleP;
	TKSO_set_stateP set_stateP;
	TKSO_set_lineWidthP set_lineWidthP;
} TKSO_device;

TKSESH_rectangle
TKSO_get_window(TKSESH_output *o);

void
TKSO_set_window(TKSESH_output *o, TKSESH_rectangle r, int w, int h);

/* gets width of the drawing window, in virtual units */
double
TKSO_get_real_width(TKSESH_output *o);

/* sets the scale. The scale is in Pixel/(virtual units) */
void 
TKSO_set_scale(TKSESH_output *o, double scale);

void
TKSO_toggle_reverse(TKSESH_output *o);

void 
TKSO_set_state(TKSESH_output *o, 
							 TKSO_state s);

TKSO_state 
TKSO_get_state(TKSESH_output *o);

void 
TKSO_set_lineWidth(TKSESH_output *o, 
									 TKSO_lineWidth wd);

TKSO_lineWidth
TKSO_get_lineWidth(TKSESH_output *o);

TKSESH_point
XPoint_to_TKSESH_point(TKSESH_output *o, XPoint m);

XPoint
TKSESH_point_to_XPoint(TKSESH_output *o, TKSESH_point p);

/* Drawing FUNCTIONS */

void
TKSO_clean(TKSESH_output *o);

void
TKSO_DrawLine(TKSESH_output *destination,
								TKSESH_point a,
								TKSESH_point b);

void
TKSO_DrawLines(TKSESH_output *destination,
								 int npoints, 
								 TKSESH_point * lespoints);

void
TKSO_hachure(TKSESH_output *destination,
							 TKSESH_rectangle rect);

void
TKSO_fillPolygon(TKSESH_output *destination,
								 int npoints, 
								 TKSESH_point * lespoints,
								 int draw);

/**
 *
 *  upperLeft	: upper left point of the bounding rectangle.
 *	angle1, angle2: expressed in 1/64 of degree. angle1 is relative to 
 *  the 3 o'clock position. angle2 is RELATIVE to angle1
 *
 */

void
TKSO_DrawArc(TKSESH_output *destination,
						 TKSESH_point upperLeft,
						 double width, double height,
						 int angle1, int angle2);


double
TKSO_String_width(TKSESH_output *destination,
										char *texte);

void
TKSO_DrawString(TKSESH_output *destination,
								TKSESH_point ou,
								char *texte);


void
TKSO_fill_rectangle(TKSESH_output *o, TKSESH_rectangle r);


/* Device independant functions */
TKSESH_point
TKSO_init_point(double x, double y);

TKSESH_rectangle
TKSO_init_rectangle(double llx, double lly, double hrx, double hry);

int
TKSO_isLowRes(TKSESH_output *o);

/* The following macros help calling the functions directly on real devices */

#define TKSO_GET_WINDOW(o) TKSO_get_window(TKSOC(o))
#define TKSO_SET_WINDOW(o,a,b,c) TKSO_set_window(TKSOC(o),a,b,c)
#define TKSO_GET_REAL_WIDTH(o) TKSO_get_real_width(TKSOC(o))
#define TKSO_SET_SCALE(o,a) TKSO_set_scale(TKSOC(o),a)
#define TKSO_TOGGLE_REVERSE(o) TKSO_toggle_reverse(TKSOC(o))
#define TKSO_SET_STATE(o,a) TKSO_set_state(TKSOC(o),a)
#define TKSO_GET_STATE(o) TKSO_get_state(TKSOC(o))
#define XPOINT_TO_TKSESH_POINT(o,a) XPoint_to_TKSESH_point(TKSOC(o),a)
#define TKSESH_POINT_TO_XPOINT(o,a) TKSESH_point_to_XPoint(TKSOC(o),a)
#define TKSO_CLEAN(o) TKSO_clean(TKSOC(o))
#define TKSO_DRAWLINE(o, a, b) TKSO_DrawLine(TKSOC(o), a, b)
#define TKSO_DRAWLINES(o, a, b) TKSO_DrawLines(TKSOC(o), a, b)
#define TKSO_HACHURE(o, a) TKSO_hachure(TKSOC(o), a)
#define TKSO_FILLPOLYGON(o, a, b, c) TKSO_fillPolygon(TKSOC(o), a, b, c)
#define TKSO_DRAWARC(o, a, b, c, d) TKSO_DrawArc(TKSOC(o), a, b, c, d)
#define TKSO_STRING_WIDTH(o, a) TKSO_String_width(TKSOC(o), a)
#define TKSO_DRAWSTRING(o, a, b) TKSO_DrawString(TKSOC(o), a, b)
#define TKSO_FILL_RECTANGLE(o, r) TKSO_fill_rectangle(TKSOC(o), r)


#define TKSOWIN(x) (TKSOC(x)->window)
#define TKSOSCALE(x) (TKSOC(x)->scale)
#define TKSODRIVER(x) (TKSOC(x)->driver)
#define TKSOW(x) (TKSOC(x)->w)
#define TKSOH(x) (TKSOC(x)->h)
#define TKSOREVERSE(x) (TKSOC(x)->reverse)
#define TKSOSTATE(x) (TKSOC(x)->state)

/*
 * The following functions can be used to build new drivers, if the
 * corresponding functionnality is not inherent to the target device.
 */

void 
TKSO_genericHachureP(TKSESH_output * o, TKSESH_rectangle rect);

#endif


/* Let's preserve the present presentation
 * Local Variables:
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
