/*
	Interface for prolog interaction in TCL
	should work with SWI prolog first, then
	with a home grown prolog
*/


#define _POSIX_SOURCE 1

/*
	 interface between Tcl and btdb
*/

/* #include <fcntl.h>							 fd manipulation */
/* #include <string.h>*/
/* #include <stdlib.h> */

/* #include "general.h"*/
/* #include "btdb.h"*/


/* Beginning of Windows compatibility part */

#if defined(__WIN32__)
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   undef WIN32_LEAN_AND_MEAN

/*
 * VC++ has an alternate entry point called DllMain, so we need to rename
 * our entry point.
 */

#   if defined(_MSC_VER)
#	define EXPORT(a,b) __declspec(dllexport) a b
#	define DllEntryPoint DllMain
#   else
#	if defined(__BORLANDC__)
#	    define EXPORT(a,b) a _export b
#	else
#	    define EXPORT(a,b) a b
#	endif
#   endif
#else
#   define EXPORT(a,b) a b
#endif


/*
 * Declarations for functions defined in this file.
 */

#include <SWI-Prolog.h>
#include <tcl.h>

EXTERN EXPORT(int,Tclpl_Init) _ANSI_ARGS_((Tcl_Interp *interp));

#ifdef __WIN32__
BOOL APIENTRY
DllEntryPoint(hInst, reason, reserved)
    HINSTANCE hInst;		/* Library instance handle. */
    DWORD reason;		/* Reason this function is being called. */
    LPVOID reserved;		/* Not used. */
{
    return TRUE;
}
#endif

PL_extension PL_extensions [] = {
  { NULL,	0, 	NULL,		0 }	/* terminating line */
};


int
tclpl_call(ClientData clientData,
					 Tcl_Interp *interp,
					 int argc,
					 char **argv)
{
	int i;
	predicate_t pred;
	term_t t;
	char *rep;
	fid_t fid ;
	
	/* argv[1] : predicate
		 argv[2] : arg1 for predicate
		 predicate has argc - 2 args
	*/
	if (argc < 2)
		{
			Tcl_AppendResult(interp, argv[0], " takes at least two arguments.", (char *) NULL);
			return TCL_ERROR;
		}

	fid = PL_open_foreign_frame();	
	pred = PL_predicate(argv[1], argc - 1, "user");
	t = PL_new_term_refs(argc - 1);
	PL_put_variable(t);

	for (i=0; i < argc-2; i++)
		PL_put_atom_chars(t + i + 1, argv[i+2]);

	if (PL_call_predicate(NULL, TRUE, pred, t))
		{
			PL_get_atom_chars(t, &rep);	
			Tcl_AppendResult(interp, rep, (char *) NULL);
		}
	
	PL_discard_foreign_frame(fid);
	
	return TCL_OK;
}

int tclpl_load(ClientData clientData,
							 Tcl_Interp *interp,
							 int argc,
							 char **argv)
{
	predicate_t pred;
	term_t t;
	fid_t fid ;

	
	/* argv[1] : predicate
		 argv[2] : arg1 for predicate
		 predicate has argc - 2 args
	*/
	if (argc != 2)
		{
			Tcl_AppendResult(interp, argv[0], " file to load.", (char *) NULL);
			return TCL_ERROR;
		}

	fid = PL_open_foreign_frame();
	pred = PL_predicate("consult", argc - 1, "user");
	t = PL_new_term_refs(1);

	PL_put_atom_chars(t, argv[1]);

	PL_call_predicate(NULL, FALSE, pred, t);

	PL_discard_foreign_frame(fid);	
	return TCL_OK;
}

EXPORT(int,Tclpl_Init)(interp)
     Tcl_Interp *interp;
{

	char * pargs[4];

	pargs[0]= "prolog";
	pargs[1]= "-g";
	pargs[2]= "true";
	pargs[3]= NULL;

	/* The current line is for version 2.9.6. For older versions,
		 consider using : 	if ( !PL_initialise(3, pargs,0)) ... */
		 
	if ( !PL_initialise(3, pargs))
		PL_halt(1);

  Tcl_CreateCommand(interp, "plcall", tclpl_call,
										(ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "plload", tclpl_load,
										(ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);

	return Tcl_PkgProvide(interp, "Tclpl", "1.0");
  return TCL_OK;
}

/*
 * Local Variables:
 * tab-width: 2
 * End:
 */
