/*
 * bitmap drawing device for tksesh
 */

/*  1998, 2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: TKSO_bitmap.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/*
 * $Log: TKSO_bitmap.c,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 */

#define _POSIX_SOURCE 1

#define TKSO_PI 3.1415926535897932384626433832795029L
// #define ANGLE_RATIO (64.0*(TKSO_PI/180.0))
#define ANGLE_RATIO (TKSO_PI/(180.0*64.0))
#include <math.h>
#include <stddef.h>

#include "chaine.h"
#include "tgraphImpl.h"
#include "TKSO_bitmap.h"
#include "SimpleBitmapTcl.h"
#include "general.h"

#define TKBM(x) ((TKSO_bitmap*)(x))

static void cleanP(TKSESH_output *);
static void DrawLineP(TKSESH_output *, TKSESH_point, TKSESH_point);
static void DrawLinesP(TKSESH_output *, int, TKSESH_point *);
static void fillPolygonP(TKSESH_output *, int , TKSESH_point * , int);
static void DrawArcP(TKSESH_output *, TKSESH_point, double, double, int, int);
static double String_widthP(TKSESH_output *, char *);
static void DrawStringP(TKSESH_output *, TKSESH_point, char *);
static void fill_rectangleP(TKSESH_output *, TKSESH_rectangle);

static TKSO_device
TKSO_bitmap_driver= {
	"bitmap",
	sizeof(TKSO_bitmap),
	cleanP,
	DrawLineP,
	DrawLinesP,
	TKSO_genericHachureP,
	fillPolygonP,
	DrawArcP,
	String_widthP,
	DrawStringP,
	fill_rectangleP,
	0,
	0
};

TKSO_bitmap *
TKSO_bitmap_init(double x1, double y1,
		 double x2, double y2,
		 int w, int h)
{
  TKSO_bitmap * o;
  o= NOUV(TKSO_bitmap);
  TKSO_init_plain((TKSESH_output *)o,x1,y1,x2,y2,w,h);
	TKSOC(o)->isLowRes= TRUE;
  TKSODRIVER(o)= &TKSO_bitmap_driver;
  o->bitmap= SimpleBitmap_init(w, h);
  return o;
}



void
TKSO_bitmap_free(TKSO_bitmap *o)
{
  SimpleBitmap_free(o->bitmap);
  LIBERE(o);
}



static void
cleanP(TKSESH_output *o)
{
  SimpleBitmap_clean(TKBM(o)->bitmap);
}

static int
getColor(TKSESH_output *o)
{
  int result;
  switch (o->state)
    {
    case white:
      result= 0;
      break;
    case ignore:
    case black:
    case gram:
    case blue:
    case gramend:
    case aux:
    case red:
    case grayfill:
    case invert:
      result= 1;
    }
  return result;
}


static void
fill_rectangleP(TKSESH_output *o, TKSESH_rectangle r)
{
  int w, h;
  XPoint aa, bb;
  aa= TKSESH_point_to_XPoint(o, r.bottom_left);
  bb= TKSESH_point_to_XPoint(o, r.top_right);
  w= (aa.x - bb.x);
  w = (w>0)?w:(-w);
  h= (aa.y - bb.y);
  h = (h>0)?h:(-h);
}


static void
DrawLineP(TKSESH_output *o,
	  TKSESH_point a,
	  TKSESH_point b)
{
  XPoint aa, bb;
  aa= TKSESH_point_to_XPoint(o, a);
  bb= TKSESH_point_to_XPoint(o, b);
  SimpleBitmap_iline(TKBM(o)->bitmap,aa.x, aa.y, bb.x, bb.y, getColor(o));
}


static void
hachureP(TKSESH_output *destination,
	 TKSESH_rectangle r)
{
  
}


static void
fillPolygonP(TKSESH_output *destination,
	     int npoints, 
	     TKSESH_point * lespoints,
	     int sens)
{
  XPoint m;
  int i;
  static Chaine chp= NULL;
  
  if (chp == empty_chaine) 
    init_Chaine(&chp, sizeof(XPoint));
	
  for (i=0; i< npoints; i++)
    {
      m= TKSESH_point_to_XPoint(destination, lespoints[i]);
      add_to_Chaine(chp, &m);
    }
	
  SimpleBitmap_fillPolygon(TKBM(destination)->bitmap, 
			   npoints,
			   Chaine_to_data(chp), 
			   sens);
  
  reset_Chaine(chp);
}




static void
DrawArcP(TKSESH_output *destination,
				 TKSESH_point ul,
				 double width,
				 double height,
				 int angle1, int angle2)
{
	int alpha;
	double angle;
	double width2, height2;
	TKSESH_point old, current, center;
	
	width2= width/2.0;
	height2= height/2.0;
	center.x= ul.x + width2;
	center.y= ul.y - height2;
	
	angle= angle1 * ANGLE_RATIO;
	old.x= cos(angle)*width2+center.x;
	old.y= sin(angle)*height2+center.y;
	
	alpha= angle1 + 1;

	while (alpha < angle1 + angle2)
		{
			angle= alpha * ANGLE_RATIO;
			current.x= cos(angle)*width2+center.x;
			current.y= sin(angle)*height2+center.y;
			TKSO_DrawLine(destination, old, current);
			old= current;
			alpha++;
		}
}


static void
DrawLinesP(TKSESH_output *destination,
	   int npoints, 
	   TKSESH_point * lespoints)
{
  SimpleBitmap *o= TKBM(destination)->bitmap;
  static Chaine chp= NULL;
  XPoint m, *xpoints;
  int i;


  if (npoints == 0)
    return;

  if (chp == empty_chaine) 
    init_Chaine(&chp, sizeof(XPoint));
	
  /* Convert the TKSESH_points into XPoints (real to integer) */
  
  for (i=0; i< npoints; i++)
    {
      m= TKSESH_point_to_XPoint(destination, lespoints[i]);
      add_to_Chaine(chp, &m);
    }
  
  xpoints= Chaine_to_data(chp);

  /* Now, draw the lines */
  for(i=1; i < npoints; i++)
    {
      SimpleBitmap_iline(o, 
			 xpoints[i-1].x, xpoints[i-1].y, 
			 xpoints[i].x, xpoints[i].y, getColor(destination));
    }
  
  /*
		SimpleBitmap_iline(o, xpoints[npoints-1].x, xpoints[npoints-1].y, 
		xpoints[0].x, xpoints[0].y, getColor(destination));
	*/
  
  reset_Chaine(chp);
}


static void
DrawStringP(TKSESH_output *destination,
	    TKSESH_point ou,
	    char *texte)
{
}



static double
String_widthP(TKSESH_output * destination,
	      char *texte)
{
  return 0.0;
}

Tcl_Obj *
TKSO_bitmap_getPBM(TKSO_bitmap *o)
{
  return SimpleBitmap_toPBM(o->bitmap);
}

Tcl_Obj *
TKSO_bitmap_getXBM(TKSO_bitmap *o)
{
  return SimpleBitmap_toXBM(o->bitmap, "toto");
}

/*
 * Local Variables:
 * tab-width: 2
 * End:
 */
