
#define _POSIX_SOURCE 1

#include "general.h"
#include <ctype.h>							/* char testing functions */
#include <tk.h>
#include <strings.h>
#include <stdlib.h>

#include "duplique.h"
#include "mdctexte.h"
#include "mdccoord.h"
/* #include <ctype.h> */
#include "bigstr.h"

/* The following value is used if the coordinate system passed is NULL */

static
Coordinate_System default_coordinate_system = {
	NULL,
 "%s",                          /*output*/
  "([0-9]+)",                   /*input*/
{ 
  0,1,0,0,0,0,0,0,0,0
},
{
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0},
{ numeric_coordinates, 0,0}
}
};

/*
 * protected_add(BigString *out, char *in, int len) --
 * Utilitary function. Adds in to out. If in is not an alphanumeric character,
 * protect it with "\"
 */

static void
protected_add(BigString *out, char *in, int len);

static void
plain_add(BigString *out, char *in, int len);


/*****************************************************************************
	coordonnes textuelles.
	
	****************************************************************************/
void
set_type_coordinate(Coordinate_System * sys, int number, int argc, char **argv)
{
	int i;
	
	if (sys->coordinates_description[number].type == symbolic_coordinates)
	{
		for (i=0; i < sys->coordinates_description[number].max_val; i++)
			LIBERE(sys->coordinates_description[number].coords[i]);
		LIBERE(sys->coordinates_description[number].coords);
	}
	
	if (strcmp(argv[0], "NUM") == 0)
	{
		sys->coordinates_description[number].type= numeric_coordinates;
	}
	else
	{
		sys->coordinates_description[number].type= symbolic_coordinates;
		sys->coordinates_description[number].coords= (char **)ALLOUER(sizeof(char*) *
																																 argc);
		sys->coordinates_description[number].max_val= argc;
		for (i=0; i<argc; i++)
		{
			sys->coordinates_description[number].coords[i]= dupliquechaine(argv[i]);
		}
	}
}



void
set_format_coordinate(Coordinate_System *sys, char *descr)
{
	BigString *input, *output;
	char * s;
	int j=0;
	int ordre= 1;
	
	input= create_BigString();
	output= create_BigString();

	for (j=1; j< 10; j++)
		sys->coordinates_order[j]= 0;

	/* Loop. Remember that descr is a UTF-8 string */
	s= descr;
	while (*s)
		{
			int len;
			Tcl_UniChar current;
			/* copy the string until next "$" */
			while (*s && *s != '$')
				{
					len = Tcl_UtfToUniChar(s, &current);			
					protected_add(input, s, len);
					plain_add(output, s, len);
					s+= len;
				}
			if (*s == '$')
				if (isdigit((unsigned)*(s+1)))
					{
						int varnum;
						varnum= *(s+1) - '0';
						sys->coordinates_order[varnum]= ordre;
						ordre++;
						append_to_BigString(output, "%s");
						switch (sys->coordinates_description[varnum].type)
							{
							case symbolic_coordinates:
								{
									int k;
									add_to_BigString(input, '(');
									for (k= 0; k < sys->coordinates_description[varnum].max_val; k++)
										{
											char *l;
											for (l= sys->coordinates_description[varnum].coords[k]; *l; l+= len)
												{		
													len = Tcl_UtfToUniChar(l, &current);
													protected_add(input, l, len);
												}
											add_to_BigString(input, '|');
										}
									delete_last_from_BigString(input); /* last '|' */
									add_to_BigString(input, ')');
									break;
								}
							case numeric_coordinates:
								{
									append_to_BigString(input, "([0-9]+)");
									break;
								}
							}
						s+= 2;
					}
				else if (*(s+1) == '$')
					{
						add_to_BigString(input, '\\');
						add_to_BigString(input, '$');
						add_to_BigString(output, '$');
						s+= 2;
					}
				else
					s++;										/* should be an error */
		}
	add_to_BigString(input, '$');
	dissolve_BigString_in_string(&input, &(sys->read_coordinates));
	dissolve_BigString_in_string(&output, &(sys->show_coordinates));

}



void
set_text_coordinate(Coordinate_System *sys, char* descr, coordsvect coords)
{
	char *tcoords[10];
	char *debut;
	char *fin;
	int i;
	Tcl_RegExp re;
	
	for (i=0; i<10; i++)
	{
		tcoords[i]= NULL;
		coords[i]= 0;
	}
	coords[1]= 1;									/* default : line 1 */
	coords[2]= 1;									/* default : page 1 */
	
	
	re= Tcl_RegExpCompile(sys->interp, sys->read_coordinates);
	if (Tcl_RegExpExec(sys->interp, re, descr, descr))
	{
		i= 1;
		Tcl_RegExpRange(re, i, &debut, &fin);
		while (debut)
		{
			tcoords[i]= (char*)ALLOUER(sizeof(char)*(fin-debut+1));
			strncpy(tcoords[i], debut, fin-debut);
			tcoords[i][fin - debut]= 0;
			i++;
			Tcl_RegExpRange(re, i, &debut, &fin);
		}
		
		for (i=1; i<10; i++)
		{
			if (sys->coordinates_order[i])
			{
				if (sys->coordinates_description[i].type == numeric_coordinates)
				{
					coords[i]= atoi(tcoords[sys->coordinates_order[i]]);
				}
				else
				{
					int val= -1;
					int j;
					char *s1= tcoords[sys->coordinates_order[i]];
					for(j=0; val == -1 && j < sys->coordinates_description[i].max_val; j++)
						if (s1 && strcmp(s1, sys->coordinates_description[i].coords[j])==0)
							val= j;
					coords[i]= val+1;
				}
			}
		}
	}
	for (i=1; i<10; i++)
		if (tcoords[i])
			LIBERE(tcoords[i]);
}




char *
get_text_coordinates(Coordinate_System *sys, coordsvect coords)
{
	int i;
	int longueur_max= 0;
	char *tcoords[10];						/* texte associ  chaque coordonne */
	char *retour;

	if (!sys)
		sys= &(default_coordinate_system);

	for (i=1; i<10; i++)
	{
		int numvar;
		tcoords[i]= NULL;

		for(numvar=0; 
				numvar < 10 && sys->coordinates_order[numvar] != i;
				numvar++);
		
		if (numvar != 10)						/* i correspond to coord numvar*/
		{
			if (sys->coordinates_description[numvar].type == numeric_coordinates)
			{
				char s[20];
				sprintf(s, "%d", coords[numvar]);
				longueur_max+= strlen(s);
				tcoords[i]= dupliquechaine(s);
			}
			else
			{
				/* coords start at 1, arrays at 0 : */
				if (coords[numvar] > 0 && coords[numvar] < 
						sys->coordinates_description[numvar].max_val + 1)
					{
						tcoords[i]= 
							dupliquechaine(sys->coordinates_description[numvar].coords[coords[numvar] - 1]);
						longueur_max+= strlen(tcoords[i]);
					}
				else
					{
						char s[20];
						sprintf(s, "%d", coords[numvar]);
						longueur_max+= strlen(s);
						tcoords[i]= dupliquechaine(s);
					}
			}
		}
	}
	longueur_max+= strlen(sys->show_coordinates);
	retour= (char*)ALLOUER(sizeof(char)*(longueur_max+1));
	sprintf(retour, sys->show_coordinates,
					tcoords[1], tcoords[2], tcoords[3], tcoords[4], tcoords[5],
					tcoords[6], tcoords[7], tcoords[8], tcoords[9]);
	for (i=1; i<10; i++)
		if (tcoords[i] != 0)
			LIBERE(tcoords[i]);
	return retour;
}



void
init_text_coordinates(Coordinate_System *sys, Tcl_Interp *interp)
{
	*sys= default_coordinate_system;
	sys->interp= interp;
}



static void
protected_add(BigString *out, char *in, int len)
{
	if (len == 1 && ! isalnum(*in))
		{
			add_to_BigString(out, '\\');
			add_to_BigString(out, *in);
		}
	else
		plain_add(out, in, len);
}

static void
plain_add(BigString *out, char *in, int len)
{
	while (len-- != 0)
		{
			add_to_BigString(out, *(in++));
		}
}

/* Let's preserve the present presentation
 * Local Variables:
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
