
#define _POSIX_SOURCE 1


/* 
	 Module suprieur de gestion de la mmoire.
*/
#include "quadra.h"
#include "mynew.h"
#include "general.h"
#include <string.h>

DECLARE_STACK(Quadra)
IMPLEMENTE_STACK(Quadra)
DECLARE_STACK(Vertlist)
IMPLEMENTE_STACK(Vertlist)
DECLARE_STACK(Listesignes)
IMPLEMENTE_STACK(Listesignes)

/**
 * PUBLIC void 
 * initialise_quadras()
 * 
 * Should be called once.
 * initialise memory for quadras and such
 * The call is mandatory
 */

PUBLIC void 
initialise_quadras()
{
	INIT_STACK(Quadra);
	INIT_STACK(Vertlist);
	INIT_STACK(Listesignes);
}

/**
 * PUBLIC void 
 * detruit_quadras()
 * 
 * Definitively frees memory from quadras & co.
 * 
 * to be called once in the end
 */

PUBLIC void 
detruit_quadras()
{
	DESTROY(Quadra);
	DESTROY(Vertlist);
}
		 
/**
 * Declares all space used for quadras and the like
 * as free space for new quadras
 * 
 * Useful if you want 
 *   1) to destroy current quadras
 *   2) to create some new one later
 *
 * can be called any number of times.
 */

PUBLIC void 
libere_quadras()
{
	CLEANUP(Quadra);
	CLEANUP(Vertlist);
}

/**
 * PUBLIC Quadra *
 * new_quadra(Typegroupe val)
 * 
 * Creates a new quadra of type val
 */

PUBLIC Quadra *
new_quadra(Typegroupe val)
{
	Quadra * q;
	q= NEW(Quadra);
	q->est_grise= FALSE;
	q->type= val;
	q->last= q;
	q->scale= 1.0;
	q->h=0.0;
	q->wcompact=0.0;
	q->w=0.0;
	q->n=0;
	q->kerning= 0;
	q->hachures= NOHACHURE;
	q->separateur= S_sansseparateur;
	q->next= NULL;
	q->previous= NULL;
	return q;
}

/**
 *
 * PUBLIC Quadra* 
 * new_sign(Typesigne s)
 *
 * Creates a new SIGN of type s
 *
 */

PUBLIC struct _Signe * 
new_sign(Typesigne ts)
{
	struct _Signe * s;
	s= NOUV(struct _Signe);
	s->type=ts;
	s->islong= 2;
	s->istall= 2;
	s->ignore= 0;
	s->estfinmot= 0;
	s->estgram= 0;
	s->selected= 0;
	s->rotation= 0;
	s->revert= 0;
	s->size= 100;
	s->typehachure= S_aucune;
	return s;
}

/**
 * PUBLIC void 
 * free_sign()
 *
 */
PUBLIC void
free_sign(struct _Signe *s)
{
	switch (s->type)
		{
		case 	S_romain:
		case 	S_inconnu:
		case S_textesuper:
			if (s->donnees.texte != NULL)
				LIBERE(s->donnees.texte);
			break;
		default:
			break;
		}
	LIBERE(s);
}

/**
 * PUBLIC Vertlist * 
 * new_vert()
 *
 * Creates and initialise a new Vertlist
 */

PUBLIC Vertlist * 
new_vert()
{
	Vertlist * h;
	h= NEW(Vertlist);
	h->un_quadra=0;
	h->scale=1.0;
	h->h=0.0;
	h->w=0.0;
	h->hcompact=0.0;
	h->n=0;
	h->kerning=0;
	h->last= h;
	h->next= NULL;
	h->previous= NULL;
	return h;
}

/**
 * PUBLIC void 
 * new_liggroup()
 *
 */

PUBLIC struct Liggroup *
new_liggroup()
{
	struct Liggroup *res;
	res= NOUV(struct Liggroup);
	res->l= NULL;
	res->nbr= 0;
	res->signs= NULL;
	return res;
}

PUBLIC struct _Signinlig *
new_signinlig()
{
	struct _Signinlig *res;
	res= NOUV(struct _Signinlig);
	res->nbr= 0;
	res->next= NULL;
	return res;
}

/**
 * PUBLIC void 
 * detruit_liggroup(struct Liggroup *l)
 *
 */
PRIVATE void
detruit_liggroup(struct Liggroup *l)
{
	struct _Signinlig *s, *next;
	s= l->signs;
	while (s)
		{
			next= s->next;
			free_sign((s->sign));
			LIBERE(s);
			s= next;
		}
	LIBERE(l);
}


/**
 * PUBLIC void 
 * detruit_quadra(Quadra *q)
 *
 * Frees a quadra
 */

PUBLIC void 
detruit_quadra(Quadra *q)
{
	while (q) {
		Quadra *next;
		next= S_NEXT_OF(*q);
		/* nettoyage du quadra */
		switch (S_TYPE_OF(*q))
		{
		case S_quadra:
			detruit_vertlist(QD_VERTLIST_OF(*q));
			break;
		case S_ligature:
			detruit_liggroup(QD_LIG_OF(*q));
			break;
		case S_carto :
		case S_debcarto:
		case S_milcarto :
		case S_fincarto :
		case S_serekh :
		case S_enceinte :
		case S_chateau :
		case S_debserekh :
		case S_milserekh :
		case S_finserekh :
		case S_debenceinte :
		case S_milenceinte :
		case S_finenceinte :
		case S_debchateau :
		case S_milchateau :
		case S_finchateau :
		case S_rightleft:
		case S_superfetatoire:
			/* case S_overwrite:*/
		case S_efface:
		case S_disparu:
		case S_ajoutscribe:
		case S_ajoutauteur:
			detruit_quadra(QD_QUADRA_OF(*q));
			break;
		case S_signe:
			free_sign((QD_SIGNE_OF(*q)));
			break;
		}
		DELETE(Quadra,q);
		q= next;
	}
}

/**
 * PUBLIC void 
 * detruit_vertlist(Vertlist *hl)
 * Frees an Vertlist
 *
 */

PUBLIC void 
detruit_vertlist(Vertlist *hl)
{
	while (hl) {
		Vertlist *next;
		next= S_NEXT_OF(*hl);
		detruit_quadra(VL_QUADRA_OF(*hl));
		DELETE(Vertlist,hl);
		hl= next;
	}
}

/**
 *
 * Auxiliary function
 */

PRIVATE Listesignes **
SESH_creer_listesignes_aux(Vertlist *h,
													 Listesignes **l)
{
	Listesignes **last;
	last = l;
	while(h)
		{
			Listesignes **suivant;
			suivant= SESH_creer_listesignes(VL_QUADRA_OF(*h), last);
			if ((*last) != NULL) {
				(*l)->last= (*last)->last;
			}
			last= suivant;
			h= h->next;
		}
	return last;
}

/**
 * PUBLIC Listesignes **
 * SESH_creer_listesignes(Quadra *q,
 *                        Listesignes **l)
 *
 * writes in *l a pointer to a signlist
 * linking the signs found in q.
 * 
 * returns a pointer to the last Listesignes*
 * which is normaly the last ->next pointer of the list
 * and whose value is NULL
 *
 * Return NULL if the list is empty
 *
 */

PUBLIC Listesignes **
SESH_creer_listesignes(Quadra *q,
											 Listesignes **l)
{
	Listesignes ** last;
	last= l;
	switch(q->type)
		{
		case S_signe:
			{
				Typesigne letype;
				letype= S_TYPE_OF(*QD_SIGNE_OF(*q));				
				if (letype == S_hieroglyphe || letype == S_inconnu) 
					{
						*l= NEW(Listesignes);
						(*l)->next= NULL;
						(*l)->last= *l;
						(*l)->signe= (QD_SIGNE_OF(*q));
						last= &((*l)->next);
					}
				else if (letype == S_overwrite)
					{
						int i;
						struct _Signe *s= QD_SIGNE_OF(*q);

						for(i=0; i<2; i++)
							{
								if (s->donnees.over[i]->type == S_hieroglyphe || s->donnees.over[i]->type == S_inconnu)
									{
										*last= NEW(Listesignes);
										(*last)->next= NULL;
										(*last)->last= *l;
										(*last)->signe= s->donnees.over[i];
										last= &((*last)->next);
									}
							}
						
					}
				break;
			}
		case S_quadra:
			last= SESH_creer_listesignes_aux(QD_VERTLIST_OF(*q),l);
			break;
		case S_ligature:
			{
				struct _Signinlig *s;
				s= QD_LIG_OF(*q)->signs;
				while (s)
					{
						if (s->sign->type == S_hieroglyphe || s->sign->type == S_inconnu)
							{
								*last= NEW(Listesignes);
								(*last)->next= NULL;
								(*last)->last= *l;
								(*last)->signe= s->sign;
								last= &((*last)->next);
							}
						s= s->next;
					}
			}
			break;
		case S_carto:
		case S_debcarto:
		case S_rightleft:
		case S_milcarto:
		case S_fincarto:
		case S_serekh:
		case S_enceinte:
		case S_chateau:
		case S_debserekh:
		case S_milserekh:
		case S_finserekh:
		case S_debenceinte:
		case S_milenceinte:
		case S_finenceinte:
		case S_debchateau:
		case S_milchateau:
		case S_finchateau:
		case S_superfetatoire:
		case S_efface:
		case S_disparu:
		case S_ajoutscribe:
		case S_ajoutauteur:
			last= SESH_creer_listesignes(QD_QUADRA_OF(*q), l);
			break;
		}
	if (q->next)
		{
			Listesignes **suivant;
			suivant= SESH_creer_listesignes(q->next,last);
			if (*last)
				(*l)->last= (*last)->last;
			last= suivant;
		}
	return last;
}

/**
 *PUBLIC void
 *SESH_detruire_listesignes(Listesignes *l)
 *
 * Destroys a signlist. 
 * it DOESN'T set l to NULL.
 *
 */

PUBLIC void
SESH_detruire_listesignes(Listesignes *l)
{
	Listesignes *next;
	while (l)
		{
			next= l->next;
			DELETE(Listesignes,l);
			l= next;
		}
}

/**
 *PUBLIC int
 *SESH_length_listesignes(Listesignes *l)
 *
 * returns the number of signs in a signlist.
 */

PUBLIC int
SESH_length_listesignes(Listesignes *l)
{
	int ret=0;
	while (l)
		{
			l= l->next;
			ret++;
		}	
	return ret;
}

/**
 *	PUBLIC Listesignes *
 *	SESH_listesignes_nth(Listesignes *l, int n)
 *  
 *  returns the nth sign of a signlist, or NULL if
 *  there is no such sign.
 *
 *  the 1st signe is numbered 1
 *
 *  The pointer points to the actual sign in the quadra,
 *  which allows direct manipulation 
 *  (e.g. changing some flags)
 */

PUBLIC Listesignes *
SESH_listesignes_nth(Listesignes *l, int n)
{
	if ( n == 0 && l)
		{
			return l->last;
		}
	else
		{
			int i;
			
			for (i=0; l && i<(n-1); i++, l= l->next)
				/* empty */;
		}
	return l;
}

/*
 * Local Variables:
 * header: "quadra.h"
 * tab-width: 2
 * End:
 */

