
/*
  Interface from hieroglyphs to Tcl :
	definition of signs and their retrieval.
*/

#define _POSIX_SOURCE 1
#include "general.h"
#include <tcl.h>
#include "bfont.h"
#include "bfaux.h"
#include "signs.h"
#include "tclsigns.h"
#include "tksinter.h"
#include "ligatur.h"

extern Tcl_CmdProc BZR_LoadFont;

/**********************************************************************
 *
 * 	1) Loading of a font
 */



PRIVATE void
BFontDeleteProc(ClientData clientData,
								Tcl_Interp *interp)
{
	BFont * f= (BFont*) clientData;
	BZR_delete(f);
}



int
BZR_LoadFont(ClientData clientData,
							Tcl_Interp *interp,
							int argc,
						  CONST char **argv)
{
	BFont *f;
	if (argc < 2)
		{
			Tcl_AppendResult(interp, argv[0], " FILENAME", (char *) NULL);
			return TCL_ERROR;
		}
	if ((f= BZR_read(argv[1])))
		{
			char s[20];
			int maxf;
			int isnew;
			struct tksesh_global *g;
			Tcl_HashEntry * entree;

			g= tksesh_get_global();
			maxf= g->nfonts;
			g->nfonts++;
			sprintf(s,"bzrf%d", maxf);
		
			entree= Tcl_CreateHashEntry(g->fonts, s, &isnew);
			Tcl_SetHashValue(entree, (ClientData)f);
			Tcl_SetResult(interp, s, TCL_VOLATILE);
			return TCL_OK;
		}
	else
		{
			Tcl_AppendResult(interp, argv[1], " does not contain a proper bzr font", (char *) NULL);
			return TCL_ERROR;
		}
}



/**********************************************************************
 *
 * 	2) giving a code to a character from a loaded font
 */


PUBLIC int
BZR_SetCharFromFont(ClientData clientData,
						Tcl_Interp *interp,
						int objc,
						Tcl_Obj *CONST objv[])
{
	struct tksesh_global *global;
	char *code;
	char *fontname;
	Tcl_HashEntry * entree;
	int num;

	if (objc != 4)
		{
			Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), " CODE FONTID NUM", (char *) NULL);
			return TCL_ERROR;
		}

	global= tksesh_get_global();

	code=  Tcl_GetStringFromObj(objv[1],NULL); /* code */
	fontname=  Tcl_GetStringFromObj(objv[2],NULL); /* font */
	entree= Tcl_FindHashEntry(global->fonts, fontname);
	if (entree ==  NULL)
		{
			Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), " error : ",
											 fontname, " is not a proper font loaded with tksesh_font", (char *) NULL);
			return TCL_ERROR;
		}

	Tcl_GetIntFromObj(interp,objv[3],&num); /* number in font */

	if (num < 256 )
	{
		BFont* font;
		Lettre *l;
		font= (BFont *)Tcl_GetHashValue(entree);
		l= &(font->lettres[num]);
		if (l->defined)
			{
				definit_lettre(code,l);
				return TCL_OK;
			}
		else
			{
				Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), 
												 " : char number ", Tcl_GetStringFromObj(objv[3],NULL), 
												 " is not defined in ", fontname, 
												 (char *) NULL);
				return TCL_ERROR;
			}
	}
	else
		{
			Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), 
											 " char number  : ", Tcl_GetStringFromObj(objv[3],NULL), 
											 " should be less than 256",
											 (char *) NULL);
			return TCL_ERROR;
		}
}



/**********************************************************************
 *
 * 	3) On the fly definition of a sign :
 * 
 *  BZR_SetChar 
 *
 */

/* Lettre setting command */

PRIVATE int
set_trait(Tcl_Interp *interp, TKSESH_point *lastpoint, Tcl_Obj *rep, Chaine path_to_build)
{
	enum {ligne, spline} type;
	TKSESH_point dernier, controle1, controle2;
	int objc;
	Tcl_Obj      **objv;
	
					
	if (Tcl_ListObjGetElements(interp, rep, &objc, &objv) == TCL_OK && objc > 0)
		{
			/* first element => code */
			char *code;

			if ((code=Tcl_GetStringFromObj(objv[0], NULL)) == NULL)
				{
					return FALSE;
				}

			switch (*code) {
			case 'l':
				{
					double px, py;
					if (objc != 3)
						return FALSE;
					if (Tcl_GetDoubleFromObj(interp, objv[1], &px) != TCL_OK)
						{
							return FALSE;
						}
					if (Tcl_GetDoubleFromObj(interp, objv[2], &py) != TCL_OK)
						{
							return FALSE;
						}
					
					type= ligne;
					dernier.x= px;
					dernier.y= py;
				}
				break;
			case 's':
				{
					double px,py,c1x,c1y, c2x, c2y;
					if (objc != 7)
						return FALSE;
					if (Tcl_GetDoubleFromObj(interp, objv[1], &c1x) != TCL_OK)
						{
							return FALSE;
						}
					if (Tcl_GetDoubleFromObj(interp, objv[2], &c1y) != TCL_OK)
						{
							return FALSE;
						}
					if (Tcl_GetDoubleFromObj(interp, objv[3], &c2x) != TCL_OK)
						{
							return FALSE;
						}
					if (Tcl_GetDoubleFromObj(interp, objv[4], &c2y) != TCL_OK)
						{
							return FALSE;
						}
					if (Tcl_GetDoubleFromObj(interp, objv[5], &px) != TCL_OK)
						{
							return FALSE;
						}
					if (Tcl_GetDoubleFromObj(interp, objv[6], &py) != TCL_OK)
						{
							return FALSE;
						}
					type= spline;
					controle1.x= c1x;
					controle1.y= c1y;
					controle2.x= c2x;
					controle2.y= c2y;
					dernier.x= px;
					dernier.y= py;
				}
				break;
			default :
				return FALSE;
			}
			switch (type)
				{
				case ligne:
					add_to_Chaine(path_to_build, &dernier);
					break;
				case spline : 
					BZR_fixe_trace(*lastpoint, controle1,
										 controle2, dernier, path_to_build);
					break;
				}
				*lastpoint= dernier;
				return TRUE;
		}
	else 	
		return FALSE;
}



PRIVATE Contour *
set_contour(Tcl_Interp *interp, Tcl_Obj *rep)
{
	int objc;
	Tcl_Obj      **objv;
	Contour *res= NULL;
	static Chaine path_to_build= NULL;
	TKSESH_point lastpoint, start;

	if (!path_to_build)
		init_Chaine(&path_to_build, sizeof(TKSESH_point));

	if (Tcl_ListObjGetElements(interp, rep, &objc, &objv) == TCL_OK && objc > 4)
		{
			int i;
			char *type;
			double x, y;
			enum senscontour s;

			/* First element is contour type. Cuttently, only CMP. One day, perhaps 
			 * circles and the like. 
			*/

			/* element 1 => positive, negative, or nofill (was objv[1]) */
			if ((type= Tcl_GetStringFromObj(objv[1], NULL)) == NULL)
				{
					return NULL;
				}
			
			if (*type == 'P')
				s = positif;
			else if (*type == 'N')
				s= negatif;
			else if (*type == 'T')
				s= nofill;


			/* second element => first point x */
			if (Tcl_GetDoubleFromObj(interp, objv[2], &x) != TCL_OK)
				{
					return NULL;
				}
			
			/* third element => first point y */
			if (Tcl_GetDoubleFromObj(interp, objv[3], &y) != TCL_OK)
				{
					return NULL;
				}
			
			/* time to allocate */

			res= NOUV(Contour);
			res->next= NULL;
			res->type= s;
			res->virtuel_trace= NULL;
			res->scaled_trace= NULL;

			start.x= x;
			start.y= y;
			lastpoint= start;
			add_to_Chaine(path_to_build, &start);

			/* Segments */
			for (i= 4; i< objc; i++)
				{

					if (! set_trait(interp, &lastpoint, objv[i], path_to_build))
						{
							BZR_delete_contour(res);
							reset_Chaine(path_to_build);
							return NULL;
						}
				}
			res->nbr_pts=  size_of_Chaine(path_to_build);
			res->virtuel_trace= (TKSESH_point *)Chaine_to_cstring(path_to_build);
			res->scaled_trace= (TKSESH_point *)Chaine_to_cstring(path_to_build);
			if (res->nbr_pts > 2)
				{
					/* if (res->type != nofill)
						 fixe_sens_contour(res);*/
				}
			else
				{
					/* LIBERE(res);					 pas de contour sans trait chez moi ! */
					/* res= NULL; */
				}
			reset_Chaine(path_to_build);
		}
	return res;
}



PRIVATE int
buildSignFromTclList(Tcl_Interp *interp, 
					 Lettre *l, 
					 Tcl_Obj *rep)
{
	int objc;
	Tcl_Obj      **objv;
	Contour *previous;

	if (l->defined)
		BZR_delete_lettre(l);
	l->defined= TRUE;
	l->loaded= TRUE;

	if (Tcl_ListObjGetElements(interp, rep, &objc, &objv) == TCL_OK && objc > 6)
		{
			int code;
			int i;
			/* first element => letter code (not really used !) */
			if (Tcl_GetIntFromObj(interp, objv[0], &(code)) != TCL_OK)
				{
					l->defined= FALSE;
					return FALSE;
				}

			l->code= code;
			/* second element => width */
			if (Tcl_GetDoubleFromObj(interp, objv[1], &(l->width)) != TCL_OK)
				{
					l->defined= FALSE;
					return FALSE;
				}
			
			/* bounding box */
			if (Tcl_GetDoubleFromObj(interp, objv[2], &(l->llx)) != TCL_OK)
				{
					l->defined= FALSE;
					return FALSE;
				}
			
			if (Tcl_GetDoubleFromObj(interp, objv[3], &(l->lly)) != TCL_OK)
				{
					l->defined= FALSE;
					return FALSE;
				}
			
			if (Tcl_GetDoubleFromObj(interp, objv[4], &(l->urx)) != TCL_OK)
				{
					l->defined= FALSE;
					return FALSE;
				}
			
			if (Tcl_GetDoubleFromObj(interp, objv[5], &(l->ury)) != TCL_OK)
				{
					l->defined= FALSE;
					return FALSE;
				}

			l->contours= NULL;
			previous= NULL;
			/* contours */
			for (i= 6; i< objc; i++)
				{					
					Contour *c;
					if ((c= set_contour(interp, objv[i])) != NULL)
						{
							if (previous == NULL)
								{
									l->contours= c;
									previous= l->contours;
								}
							else
								{
									previous->next= c;
									previous= c;
								}
						}
					else
						{
							BZR_delete_lettre(l);
							return FALSE;
						}
				}
		} else {
			l->defined= FALSE;
			return FALSE;
		}
	return TRUE;
}

/***
 *  BZR_SetChar --
 *
 */

PRIVATE int
buildSignFromTclList(Tcl_Interp *interp, 
					 Lettre *l, 
					 Tcl_Obj *rep);


PUBLIC int
BZR_SetChar(ClientData clientData,
						Tcl_Interp *interp,
						int objc,
						Tcl_Obj *CONST objv[])
{
	char *code;
	Lettre *l;

	if (objc != 3)
		{
			Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), " CODE VALUE", (char *) NULL);
			return TCL_ERROR;
		}
	code=  Tcl_GetString(objv[1]);
	l= NOUV(Lettre);
	l->defined= FALSE;
	l->loaded= FALSE;
	l->font= NULL;
	if (buildSignFromTclList(interp, l, objv[2]))
		{
			definit_lettre(code,l);
			return TCL_OK;
		}
	else
		{
			Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), " bad definition : ", Tcl_GetStringFromObj(objv[2],NULL), (char *) NULL);
			return TCL_ERROR;
		}
}


/**********************************************************************
 *
 * 	3bis) Sign Equivalence
 * 
 *  BZR_defineSignEquivalence
 *
 *  takes two sign codes as arguments. The first one is said to be equivalent
 *  (has the same drawing) as the second one.
 *
 */

PRIVATE int
BZR_defineSignEquivalence(ClientData clientData,
													Tcl_Interp *interp,
													int objc,
													Tcl_Obj *CONST objv[])
{
	char * code1, * code2;
	union _Donnees signData;
	if (objc != 3)
		{
			Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), " CODE1 CODE2", (char *) NULL);
			return TCL_ERROR;
		}
	code1=  Tcl_GetString(objv[1]);
	code2=  Tcl_GetString(objv[2]);
	/* try to get code2's char */
	signData= seshtrouvesigne(code2);
	/* if code2 was found, attach the result to code1 */
	if (signe_trouve)
		{
			definit_lettre(code1, signData.hieroglyphe.l);
			return TCL_OK;
		}
	else
		{
			Tcl_AppendResult(interp,  Tcl_GetString(objv[0]), " could not find : ",
											 Tcl_GetString(objv[2]), (char *) NULL);
			return TCL_ERROR;
		}
}



/**********************************************************************
 *
 * 	4) definition of a ligature.
 */

PUBLIC int
TKSESH_Define_Ligature(ClientData clientData,
											 Tcl_Interp *interp,
											 int objc,
											 Tcl_Obj *CONST objv[])
{
	struct tksesh_global *global;
	int i, j, isnew;
	struct Ligature * lig;
	Tcl_HashEntry * entree;
	int nbr;

	if (objc < 6)
		{
			Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), 
											 " NBRSIGNS ?SIGN1 X1 Y1 SIZE1? SIGN2 ...", (char *) NULL);
			return TCL_ERROR;
		}
	if (Tcl_GetIntFromObj(interp, objv[1], &(nbr)) != TCL_OK)
		{
			Tcl_AppendResult(interp, Tcl_GetStringFromObj(objv[0],NULL), 
											 " first argument should be an integer.: ",
											 Tcl_GetStringFromObj(objv[1],NULL),
											 (char *) NULL);
			return TCL_ERROR;
		}

	if (objc != 2 + nbr * 4)
		{
			Tcl_AppendResult(interp, Tcl_GetStringFromObj(objv[0],NULL), 
											 " bad number of arguments ",
											 (char *) NULL);
			return TCL_ERROR;			
		}

	lig= ligature_create(nbr);
	for(i=0, j=2; i< nbr; i++)
		{
			char * code;
			double x, y, size;
			
			code= Tcl_GetStringFromObj(objv[j],NULL);
			j++;
			
			if (Tcl_GetDoubleFromObj(interp, objv[j], &(x)) != TCL_OK)
				{
					char s[10];
					sprintf(s,"%d", j);
					Tcl_AppendResult(interp, Tcl_GetStringFromObj(objv[0],NULL), 
													 " argument ", s, " should be a double.: ",
													 Tcl_GetStringFromObj(objv[j],NULL),
													 (char *) NULL);
					return TCL_ERROR;
				}
			j++;
			if (Tcl_GetDoubleFromObj(interp, objv[j], &(y)) != TCL_OK)
				{
					char s[10];
					sprintf(s,"%d", j);
					Tcl_AppendResult(interp, Tcl_GetStringFromObj(objv[0],NULL), 
													 " argument ", s, "should be a double.: ",
													 Tcl_GetStringFromObj(objv[j],NULL),
													 (char *) NULL);
					return TCL_ERROR;
				}
			j++;
			if (Tcl_GetDoubleFromObj(interp, objv[j], &(size)) != TCL_OK)
				{
					char s[10];
					sprintf(s,"%d", j);
					Tcl_AppendResult(interp, Tcl_GetStringFromObj(objv[0],NULL), 
													 " argument ", s, "should be a double.: ",
													 Tcl_GetStringFromObj(objv[j],NULL),
													 (char *) NULL);
					return TCL_ERROR;
				}
			j++;
			if (ligature_subsign_init(lig,i,code,x,y,size) == FALSE)
				{
					Tcl_AppendResult(interp, 
													 " code ", code, " is unknown",
													 (char *) NULL);
					return TCL_ERROR;
				}
		}
	ligature_make_ident(lig);
	ligature_complete(lig);
	/* OK. Now, our ligature is ready to use. */
	/* let's register it : */
	global= tksesh_get_global();
	entree= Tcl_CreateHashEntry(global->ligatures, ligature_to_code(lig), &isnew);
	Tcl_SetHashValue(entree, (ClientData)lig);
	return TCL_OK;
}


/**********************************************************************
 *
 * 	5) CallBack function for dealing with unknown signs.
 */

/*
 * sets the callBack function.
 */

PUBLIC int
TKSESH_Signs_setsCallBack(ClientData clientData,
													Tcl_Interp *interp,
													int objc,
													Tcl_Obj *CONST objv[])
{
	struct tksesh_global *g;

	if (objc > 2)
		{
			Tcl_AppendResult(interp,  Tcl_GetStringFromObj(objv[0],NULL), 
											 " bad number of arguments !", (char *) NULL);
			return TCL_ERROR;
		}

	g= tksesh_get_global();
	/*
	 * suppress the previous callBack
	 */
	
	if (g->f.callBack != NULL)
		{
			Tcl_DecrRefCount(g->f.callBack);
			g->f.callBack= NULL;
			g->f.interp= NULL;
		}

	if (objc == 2) {
		/* We shall use objv[1] as callback function */
		/* to do this, let's ensure its value will be kept ! */
		g->f.callBack= objv[1];
		g->f.interp= interp;
		Tcl_IncrRefCount(objv[1]);
	}
	return TCL_OK;
}

/*
 * the actual call :
 */

void
TKSESH_Signs_searchCode(char * code)
{
	struct tksesh_global *g= tksesh_get_global();
	if (g->f.callBack != NULL)
		{
			Tcl_Obj * comms[2];
			comms[0]= g->f.callBack;
			comms[1]= Tcl_NewStringObj(code, strlen(code));
			Tcl_EvalObjv(g->f.interp, 2, comms, 0);
		}
}

/**********************************************************************
 *
 * 	6) initialisation in Tcl
 */

PUBLIC void
TKSESH_Signs_Init(Tcl_Interp *interp)
{
	struct tksesh_global *g= tksesh_get_global();
	g->fonts= NOUV(Tcl_HashTable);
	g->ligatures= NOUV(Tcl_HashTable);

	SESH_setLoadSignFunction(TKSESH_Signs_searchCode);
	
	Tcl_InitHashTable(g->fonts, TCL_STRING_KEYS);
	Tcl_InitHashTable(g->ligatures, TCL_STRING_KEYS);
  
	Tcl_CreateObjCommand(interp, "tksesh_defchar",BZR_SetChar,
											 (ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);
	/*
	 * A code equivalence command :
	 * tksesh_sameSign A B => A will use the *same* drawing as B.
	 * Note that B will be created if it doesn't exist.
	 * The command will fail if B can't be created.
	 *
	 */
	Tcl_CreateObjCommand(interp, "tksesh_samesign",BZR_defineSignEquivalence,
										(ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "tksesh_font",BZR_LoadFont,
											 (ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateObjCommand(interp, "tksesh_setchar",BZR_SetCharFromFont,
											 (ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateObjCommand(interp, "tksesh_setlig",TKSESH_Define_Ligature,
											 (ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateObjCommand(interp, "tksesh_unknownSignCommand",TKSESH_Signs_setsCallBack,
											 (ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);

}


/* final cleanup */
PUBLIC void
TKSESH_Signs_Free(Tcl_Interp *interp)
{
	Tcl_HashEntry * e;
	Tcl_HashSearch   pos;
	struct tksesh_global *g= tksesh_get_global();

	/* first delete the font table */
	e= Tcl_FirstHashEntry(g->fonts, &pos);
	while (e) {
		/* From what I understand of the present Hash table system,
		 * the memory space for the key is under control of the Hash table
		 */
		BZR_delete((BFont *)Tcl_GetHashValue(e));
		e= Tcl_NextHashEntry(&pos);
	}
	Tcl_DeleteHashTable(g->fonts);
	LIBERE(g->fonts);

	e= Tcl_FirstHashEntry(g->ligatures, &pos);
	while (e) {
		/* From what I understand of the present Hash table system,
		 * the memory space for the key is under control of the Hash table
		 */
		ligature_delete((struct Ligature *)Tcl_GetHashValue(e));
		e= Tcl_NextHashEntry(&pos);
	}
	Tcl_DeleteHashTable(g->ligatures);
	LIBERE(g->ligatures);

}


/* Let's preserve the present presentation
 * Local Variables:
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
