
#define _POSIX_SOURCE 1

#define POSX0 0.1

/*
 * procedures for displaying the quadras
 */
#include <limits.h>
#include <tk.h>

#include "mdctexte.h"
#include "tksaff.h"
#include "TKSO_tkwin.h"
#include "tksaux.h"
#include "affsigne.h"
#include "general.h"


/*****************************************************************************
											 Procdures spcialises
*****************************************************************************/

/* le quadra est-il ne serait-ce que partiellement visible */

static enum _etat
calcule_etat(TKSESH_output *out, TKSTexteGeometry *geom, TKSESH_rectangle fen, Texte *t)
{
	double xb, yb, xt, yt;
	if (!t->real_dims)
		return E_CACHE;
	xb= t->x+0.1e-6;
	yb= t->y+0.1e-6;
	xt= t->x+ SESH_texte_width(out, geom, t)-0.1e-6;
	yt= t->y + SESH_texte_height(out, geom, t) - 0.1e-6;
	if (( fen.bottom_left.x <= xb &&  fen.bottom_left.y <= yb) &&
			( fen.top_right.x >= xt &&  fen.top_right.y >= yt))
		return E_AFFICHE;
	else if ((( fen.bottom_left.x <= xb &&  fen.bottom_left.y <= yb) 
						&& ( fen.top_right.x >= xb &&  fen.top_right.y >= yb)) 
					 ||
					 (( fen.bottom_left.x <= xt &&  fen.bottom_left.y <= yt) &&
						( fen.top_right.x >= xt &&  fen.top_right.y >= yt)))
		return E_PARTIEL;
	else
		return E_CACHE;
}



static void
initPositions(Sesh * hiero, double *posx, double *posy)
{
	/***  to ensure posy is really <= 0, we used to substract 1e-6 from it */
	double dy= 1.0; /* Should be a constant ????  */
	*posy= hiero->virtuallinesep -0.05;

	if (TKSGeom_isInColumnsP(&hiero->geom))
		{
			*posx= -dy*COLUMNRATIO;
		}
	else
		{
			*posx= POSX0;
		}
}

/*
 * Now, there is a big problem here.
 * x and y were just kept to improve the rendering speed.
 * But on the other hand, we have some use for x,
 * especially in TKSESH_isCadratHorizontallyVisible.
 *
 * We'd like to compute x in exactly ONE place. 
 * However, it's difficult here. Our architecture doesn't scale well
 * with multiple rendering algorithms.
 */

static double
computeHorizontalCoord(Sesh *hiero, Texte *t)
{
	double res= 0;
	double posx, posy;
	Texte *t1;

	initPositions(hiero, &posx, &posy);
	
	if (TKSGeom_isInColumnsP(&hiero->geom))
		{
			/* if the display is column oriented :
			 * find the first column displayed. compute ????
			 */
			t1= t;
			
			/* find beginning of line. Should always work. Should be a function !!!*/

			while (t1->type != T_NL && t1->type != T_NP)
				t1= t1->previous;
			
			/* Now, find t.*/

			while (t1 != t)
				{
					t1= t1->next;
				}
			
		}
	else
		{
			/* if the display is line oriented :
			 * find the beginning,
			 * compute the length.
			 */
			
		}
	
	return res;
}


/**
 * TKSESH_isCadratHorizontallyVisible --
 * 		Sends back TRUE if the horizontal position of the cadrat is such that the cadrat should 
 *    be visible.
 * 
 *    If it isn't, *pos will contain a proposition for a left window start coordinate which
 *    would make the cadrat visible.
 *
 *    works correctly only if TKSESH_redrawWidget has tried to draw the text.
 */

int 
TKSESH_isCadratHorizontallyVisible(Sesh * hiero, GeneralIndex cadrat, double *pos)
{
	Texte *t;
	TKSESH_rectangle window;
	int pasvux= TRUE;

	t= GETTEXTOF(cadrat);
	window= TKSO_GET_WINDOW(hiero->affichage);

	/* RMQ : in displayCadrat, compteur is up to date if its value is the same 
		 as precedent_compteur. Now, here, the good value is simply hiero->compteur
		 */
	if (!hiero->wrap)
		{
			double x;
			SESH_fix_dims(TKSOC(hiero->affichage), &hiero->geom, t);
			/* x= computeHorizontalCoord(hiero, t);*/
			x= t->x;
			if (window.bottom_left.x <= x 
					&& 
					window.top_right.x >= x + SESH_texte_width(TKSOC(hiero->affichage), &hiero->geom, t))
				pasvux= FALSE;							/* the X window is correct */
	
			if (pasvux) {
				double width= window.top_right.x- window.bottom_left.x;			
				*pos= x - width/2.0;
				if (*pos<0.0) *pos= 0.0;
			} 
	
			return ! pasvux;
		}
	else
		return TRUE;
}

int 
TKSESH_isCadratVerticallyVisible(Sesh * hiero, GeneralIndex cadrat, GeneralIndex *debut)
{
	Texte *t ;
	TKSESH_rectangle window;
	int pasvuy= TRUE;
	int diffy;
	GeneralIndex ecran;

	t= GETTEXTOF(cadrat);
	window= TKSO_get_window(TKSOC(hiero->affichage));

	/* RMQ : in displayCadrat, compteur is up to date if its value is the same 
		 as precedent_compteur. Now, here, the good value is simply hiero->compteur
		 */

	/* t1=t;
		 SESH_compute_line_length(hiero,&t1);*/
	
	if (!hiero->wrap)
		{
			ecran= SESH_get_marque(hiero, "ecran");
			diffy= SESH_get_screen_pos(hiero,ecran, cadrat);
			if (diffy >= 0 && diffy < hiero->nombre_de_lignes_affichees)
				pasvuy= FALSE;
			
			if (pasvuy)
				{
					int saut;
					
			/* calcul de debut */
					saut= (hiero->nombre_de_lignes_affichees-1) / 2;
					*debut= SESH_moveby_lines(hiero, cadrat, -saut);
				}
			
			return !  pasvuy;
		}
	else
		{
			/* RMQ : in displayCadrat, compteur is up to date if its value
			 *	 is the same as precedent_compteur. Now, here, the good value
			 *	 is simply hiero->compteur 
			 */
			pasvuy= (t->etat != E_AFFICHE) || (t->compteur != hiero->compteur);
			if (pasvuy) {
				/* should it be screenlines ? */
				*debut= SESH_moveby_lines(hiero, cadrat, 1);
				*debut= SESH_moveby_lines(hiero, *debut, -1);
			}
			return !pasvuy;
		}
}



#define VERYSMALLSKIP 0.025

static void
displayCadrat(Sesh *hiero, TKSO_tkwin *old, TKSO_tkwin *new, 
							 Texte *montexte, 
							 double posx, double posy)
{
	// double resol;
	TKSESH_rectangle window;
	enum _etat ancien_etat;
	
	if (!montexte->q)
		return;

	/* resol= hiero->resolution; */
	window= TKSO_get_window(TKSOC(hiero->affichage));

	SESH_fix_dims(TKSOC(hiero->affichage), &hiero->geom, montexte);

	montexte->x= posx;
	montexte->y= posy;

	if (montexte->compteur != hiero->precedent_compteur) /* etat == cache */
		ancien_etat= E_CACHE;
	else
		ancien_etat= montexte->etat;

	
	montexte->etat= calcule_etat(TKSOC(hiero->affichage), &hiero->geom, window, montexte);

	if (montexte->etat != E_CACHE)
		{
			if (ancien_etat == E_AFFICHE)
				{
					TKSESH_rectangle source;
					TKSESH_point destination;
					double wd= SESH_texte_width(TKSOC(hiero->affichage), &hiero->geom, montexte);
					double ht= SESH_texte_height(TKSOC(hiero->affichage), &hiero->geom, montexte);
					source= TKSO_init_rectangle(montexte->oldx -VERYSMALLSKIP, 
																			montexte->oldy -VERYSMALLSKIP,
																			montexte->oldx+ wd + VERYSMALLSKIP,
																			montexte->oldy+ ht + VERYSMALLSKIP);
					destination= TKSO_init_point(posx -VERYSMALLSKIP, posy -VERYSMALLSKIP);
					TKSO_tkwin_trans_copie_fenetre(old, new, source, destination);
					/* For demonstration purpose of what signs are actually redrawn, 
						 change the 0 below in 1 */
					if (0) {
						TKSESH_point a,b;
						a.x= (montexte->x);
						a.y= (montexte->y);
						b.x= (montexte->x+ wd);
						b.y= (montexte->y+ ht);
						TKSO_DRAWLINE(new, a, b);
					}
				}
			else
				{
					TKSO_SET_STATE(new, montexte->couleur);
					SESH_aff_texte(TKSOC(new), &hiero->geom, &posx, &posy, montexte);
				}
			montexte->oldx= montexte->x;
			montexte->oldy= montexte->y;
		}
	montexte->compteur= hiero->compteur;
}



/* Tksesh rely heavily on Pixmap copying and reservation. It works
	well and fast under LINUX, but Tk under other OSes seem to have
	difficulties. Hence we limit the burden of the system here by
	recycling old resources ourselves */


int
get_new_affichage(Sesh *hiero, TKSO_tkwin* newout)
{
	int res;
	if (hiero->keep_old && hiero->past_affichage_is_usable)
		{
			/* so far, so good. Now, does it have the good characteristics ??? */
			res= TKSO_tkwin_recycle(newout, 
															hiero->affichage,
															hiero->past_affichage);
			hiero->past_affichage_is_usable= FALSE;
		}
	else
		{
			res= TKSO_tkwin_get_working_copy(newout, hiero->affichage);
		}
	return res;
}

static void 
keep_old(Sesh *hiero)
{
	if (hiero->keep_old) {
		TKSO_tkwin_steal_copy(hiero->past_affichage, hiero->old_affichage);
		hiero->past_affichage_is_usable= TRUE;
	} else {
		TKSO_tkwin_free_draw(hiero->old_affichage);
	}
}


/**
 * TKSESH_fixPosition --
 * 	 
 *   This function's job is to compute the coordinates (x,y) of
 *   a given cadrat. These depend on the cadrat AND on the way it is displayed.
 */

void
TKSESH_fixPosition(Sesh *hiero, GeneralIndex cadrat)
{
}



/******************************************************************************
													 Screen Painting
******************************************************************************/

/* 
 * NOTE ON POSITION : as of before Thu Apr 26 10:33:20 CEST 2001, a
 * text item's x and y position where updated by displayCadrat().  The
 * problem is that these coordinates are also used by other functions,
 * notation TKSESH_isCadratHorizontallyVisible, BEFORE displayCadrat
 * is called.
 */

void
TKSESH_redrawWidget(Sesh *hiero)
{
	int numligne, maxlignes;
	double posx, posy, dy;
	TKSESH_point bascursor, hautcursor;
	Texte * texte;
	static TKSO_tkwin *newout=0;
	int tout_refaire;

	if(newout==0)
		newout= TKSO_tkwin_init(0,0,0,0,0,0,0,0, 0);

	tout_refaire= hiero->display_needed;
	hiero->display_needed= FALSE;
	
	TKSO_tkwin_init_draw(hiero->affichage);
	

	if (! TKSO_tkwin_is_ready(hiero->old_affichage))
		TKSO_tkwin_get_working_copy(hiero->old_affichage, 
																hiero->affichage);
	else if (!tout_refaire)
		{
			/* effaage du curseur par affichage de l'ancienne version
				 s'il n'y a pas de nouveau dessin*/
			
			TKSO_tkwin_copy(hiero->old_affichage, hiero->affichage);
		}
	
	
	if (tout_refaire)
		{
			double maxx= 0.0;
			// double maxw= 0.0;

			maxlignes= hiero->nombre_de_lignes_affichees;
			/* dy= hiero->htline * hiero->echelle; */
			/* why not : */
			dy= 1;
			/* ??? */
			numligne=-1;								/* On commence TOUJOURS 
																	 * par un saut de ligne */
			if (get_new_affichage(hiero, newout))
				{
					TKSO_SET_STATE(newout, black); 

					texte= SESH_start_of_screen(hiero);

					initPositions(hiero, &posx, &posy);
			
					SESH_increment_compteur(hiero);
			
					while (texte && numligne <= maxlignes)
						{
							SESH_fix_dims(TKSOC(newout), &hiero->geom, texte);
							switch (texte->type) 
								{
								case T_TEXTE:
									{
										double xtexte;
									
										if (texte->etat == E_NOUVEAU)
											{
												texte->etat= E_CACHE;
											}

										if (!TKSGeom_isInColumnsP(&hiero->geom) && hiero->wrap 
												&& (posx + SESH_texte_width(TKSOC(hiero->affichage), &hiero->geom, texte)
														> TKSOC(hiero->affichage)->window.top_right.x))
											{
												if (posx > maxx)
													maxx= posx;
												posy -= dy;
												posx = POSX0;
												numligne++;
											}
										xtexte= posx;

										if (TKSGeom_isInColumnsP(&hiero->geom))
											{
												// xtexte+= ((dy*1.2) - SESH_texte_width(texte))/2.0;
												posy-= (SESH_texte_height(TKSOC(hiero->affichage), &hiero->geom, texte) + 0.1);
											}
										displayCadrat(hiero, 
																	hiero->old_affichage,
																	newout,
																	texte, xtexte, posy);
										if (! TKSGeom_isInColumnsP(&hiero->geom))
											{
												posx+= SESH_texte_width(TKSOC(hiero->affichage), &hiero->geom, texte);
											}
									}
									break;
								case T_NL:
									if (posx > maxx)
										maxx= posx;
									if (TKSGeom_isInColumnsP(&hiero->geom))
										{
											posx+= dy*COLUMNRATIO;
											posy= hiero->virtuallinesep - dy;
										}
									else
										{
											posy -= dy;
											/* posy -= */
											posx= POSX0;
										}
									numligne++;
									break;
								case T_NP:
									{
										static char *dist= "- - - 8<- - - - - - - - - - - - - - - - - - - ";
										TKSESH_point m;
										if (posx > maxx)
											maxx= posx;
										posy -= dy;
										numligne++;
										if (!TKSGeom_isInColumnsP(&hiero->geom)) {
											m.x = (POSX0);
											m.y = (posy + 0.25*dy);
											TKSO_DRAWSTRING(newout, m, dist);
											posy -= dy;
											posx= POSX0;
											numligne++;
										}
										break;
									}
								case T_LIGNE_COMMENTAIRE:
									if (hiero->display_comments)
										{
											if (!TKSGeom_isInColumnsP(&hiero->geom)) {
												TKSESH_point m;
												if (posx > maxx)
													maxx= posx;
												m.x = (POSX0);
												m.y = (posy + 0.25*dy);
												TKSO_DRAWSTRING(newout, m, texte->texte);
												posy -= dy;
												posx= POSX0;
												numligne++;
											}
										}
									break;
								case T_FIXE_COORDONNEES:
									{
										if (!TKSGeom_isInColumnsP(&hiero->geom)) {
											TKSESH_point m;
											if (posx > maxx)
												maxx= posx;
											m.x = (POSX0);
											m.y = (posy + 0.25*dy);
											TKSO_DRAWSTRING(newout, m, texte->coordonnees.descr);
											posy -= dy;
											posx= POSX0;
											numligne++;
										}
										break;
									}
								}
							if (texte->type != T_TEXTE) 
								{
									TKSESH_rectangle window;
									window= TKSO_GET_WINDOW(hiero->affichage);
									texte->x= posx;
									texte->y= posy;
									texte->etat= calcule_etat(TKSOC(hiero->affichage), &hiero->geom, window, texte);
									texte->compteur= hiero->compteur;
								}
							texte= texte->next;
						}
					if (posx > maxx)
						maxx= posx;
				
					/* copy of the result in old_affichage; 
						 copy of old_affichage on screen
						 general rejoicing and much drinking
					*/
					/* 1) put it on screen */
					/* 2) delete old and Copy new in old*/
					keep_old(hiero);
					TKSO_tkwin_steal_copy(hiero->old_affichage, newout);
					/* copy the content of old to screen */
					TKSO_tkwin_copy(hiero->old_affichage, hiero->affichage);
			
					if (maxx != hiero->maxx)
						{
							hiero->old_maxx= hiero->maxx;
							hiero->maxx= maxx;
						}
				}
		}

	/* 4) draw cursor */
	if (hiero->montexte->etat != E_CACHE
			&& hiero->montexte->compteur == hiero->compteur
			&& hiero->draw_cursor
			)
		{
			SESH_fix_dims(TKSOC(hiero->affichage), &hiero->geom, hiero->montexte);		
			/* newout.gc= hiero->auxgc;*/
			if (TKSGeom_isInColumnsP(&hiero->geom))
				{
					bascursor= TKSO_init_point(hiero->montexte->x,
																		 hiero->montexte->y -0.1);
					hautcursor= TKSO_init_point(hiero->montexte->x+
																			SESH_texte_width(TKSOC(hiero->affichage),
																											 &hiero->geom, hiero->montexte),
																			hiero->montexte->y -0.1);
				}
			else
				{
					bascursor= TKSO_init_point(hiero->montexte->x+ SESH_texte_width(TKSOC(hiero->affichage),
																																					&hiero->geom, hiero->montexte),
																		 hiero->montexte->y);
					hautcursor= TKSO_init_point(hiero->montexte->x+ SESH_texte_width(TKSOC(hiero->affichage),
																																					 &hiero->geom, hiero->montexte),
																			hiero->montexte->y+1.0);
				}
			TKSO_SET_STATE(hiero->affichage, aux);
			TKSO_DRAWLINE(hiero->affichage, bascursor, hautcursor);
			TKSO_SET_STATE(hiero->affichage, black);
		}
}




/* Let's preserve the present presentation
 * Local Variables:
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
