
#define _POSIX_SOURCE 1


/*
	 Various routines used by the system, mainly
	 to implement the actions
*/

#include "duplique.h"
#include "tksesh.h"
#include "tksaux.h"
#include "tksaff.h"
#include "affsigne.h"
#include "general.h"
#include <strings.h>
#include <stdlib.h>
#include <limits.h>

PRIVATE Texte * 
get_text_from_coordinates(Coordinate_System * sys, char *descr, Texte *depart);

/*****************************************************************************
	Les marques
 *
 * Marks are names for places in texts.
 *
 * They can be set and retrieved and are used
 * as symbolic names to manipulate text
 *
 * A number of names are reserved for special marks :
 * 
 *     "insert" is the name of the "current" point in the text.
 *     "ecran"  is the name of the beginning of the text that 
 *              appears on the screen.
 *
 ****************************************************************************/
PRIVATE void
oter_marque_texte(Marque *m, Texte *t)
{
	Liste_marques *l, *lprev;
	int trouve;

	trouve= FALSE;
	lprev= NULL;
	l= t->liste_des_marques;
	while (l && !trouve)
	{
		lprev= l;
		trouve= (l->m == m);
		l= l->next;
	}
	if (lprev && trouve)
	{
		if (lprev->next)
			lprev->next->previous= lprev->previous;
		if (lprev->previous)
			lprev->previous->next= lprev->next;
		else
			t->liste_des_marques= lprev->next;
		LIBERE(lprev);
	}
	/* m->texte= NULL;*/
}



PRIVATE void
ajouter_marque_texte(Marque *m, Texte *t)
{
	Liste_marques *l;
	l= (Liste_marques*)ALLOUER(sizeof(Liste_marques));
	l->next= l->previous= NULL;
	l->m= m;
	if (t->liste_des_marques)
		t->liste_des_marques->previous= l;
	l->next= t->liste_des_marques;
	t->liste_des_marques= l;
}



PRIVATE Marque *
get_marque(Sesh *hiero, char *m)
{
	Marque* l, *lprev;
	int trouve;

	lprev= l= hiero->les_marques;
	trouve= FALSE;
	
	while (!trouve && l)
	{
		lprev= l;
		trouve= (strcmp(l->nom, m)==0);
		l= l->next;
	}
	if (trouve)
		return lprev;
	else
		return NULL;
}


/**
 * 
 * Sends back a pointer to the Texte 
 * item marked with m
 *
 */

PUBLIC GeneralIndex
SESH_get_marque(Sesh *hiero, char *m)
{
	Marque *l;
	if (strcmp(m,"insert") == 0)
		return SESH_make_general_index_to(hiero->montexte,0);
	l= get_marque(hiero,m);
	if (l)
		return l->index;
	else
		return SESH_make_general_index_to(hiero->montexte,0);
}


/**
 * PUBLIC void
 * SESH_detruire_marques(Sesh *hiero)
 * 
 * Suppresses all marks
 */

PRIVATE void
SESH_detruire_marque_aux(Sesh *hiero, char *m);

PUBLIC void
SESH_detruire_marques(Sesh *hiero)
{
	while (hiero->les_marques) {
		SESH_detruire_marque_aux(hiero, hiero->les_marques->nom);
	}
}


/**
 * PUBLIC void
 * SESH_detruire_marque(Sesh *hiero, char *m)
 * 
 * Suppresses the mark m
 */

PUBLIC void
SESH_detruire_marque(Sesh *hiero, char *m)
{
	if (strcmp(m, "insert")==0 || strcmp(m, "ecran")==0 )
		return;
	else
		SESH_detruire_marque_aux(hiero,m);
}


/**
 * PRIVATE void
 * SESH_detruire_marque_aux(Sesh *hiero, char *m)
 * 
 * Suppresses the mark m
 */

PRIVATE void
SESH_detruire_marque_aux(Sesh *hiero, char *m)
{
	Marque *l;
	l= get_marque(hiero,m);
	if (l)
	{
		/* destruction dans le Texte */
		if (GETTEXTOF(l->index))
			oter_marque_texte(l, GETTEXTOF(l->index));
		LIBERE(l->nom);
		/* destruction globale */
		if (l->previous)
			l->previous->next= l->next;
		if (l->next)
			l->next->previous= l->previous;
		if (hiero->les_marques == l)
			hiero->les_marques= l->next;
		LIBERE(l);
	}
}

/**
 *
 * PUBLIC void
 * SESH_set_marque(Sesh *hiero, char *m, GeneralIndex ou)
 *
 * Sets a mark to a given Text item.
 * 
 */

PUBLIC void
SESH_set_marque(Sesh *hiero, char *m, GeneralIndex ou)
{
	Marque *l;
	Texte *t;
	t= GETTEXTOF(ou);
	if (t == NULL)
		return;

	if (strcmp(m, "insert")== 0)
	{
		hiero->montexte= t;
		return;
	} 
	else if (strcmp(m, "ecran")== 0)
	{
		hiero->old_debut_ecran= 
			GETTEXTOF(SESH_get_marque(hiero, "ecran"));
	}

	l= get_marque(hiero,m);
	if (l)
	{
		oter_marque_texte(l, GETTEXTOF(l->index));
		l->index= ou;
		ajouter_marque_texte(l, t);
	}
	else
	{
		l= (Marque*)ALLOUER(sizeof(Marque));
		l->nom= dupliquechaine(m);
		l->index= ou;
		l->previous= l->next= NULL;
		ajouter_marque_texte(l,t);
		if (hiero->les_marques)
		{
			l->next= hiero->les_marques;
			hiero->les_marques->previous= l;
		}
		hiero->les_marques= l;
	}
}


/*****************************************************************************
	Les indices
	An index is a plain reference to a point in the text.
	Normaly, an index designate a text item (a cadrat)
	but it can designate a SIGN in a cadrat, if needed.
	This last meaning is not completely defined and not generaly used
 ****************************************************************************/

PRIVATE GeneralIndex
SESH_index_aux(Sesh *hiero, char *index)
{
	char nomtag[1024];
	char c;
	int x; int y;

	if (index == NULL || index[0] == 0)
		return SESH_make_general_index_to(hiero->montexte,0);
	
	if (strcmp(index, "insert") == 0)
	{
		return SESH_make_general_index_to(hiero->montexte,0);
	}
	else if (index[0] == '/')
	{
		Texte *t;
		t= get_text_from_coordinates(&(hiero->coordinates), 
																		 index + 1, hiero->debut_texte);
		return SESH_make_general_index_to(t,0);
	}
	else if (sscanf(index, "%d.%d", &x, &y) == 2)
	{
		Texte *result;
		int i;
		if (x < 1)
			x= 1;
		if (y < 0)
			y= 0;
		result= GETTEXTOF(SESH_moveby_lines(hiero, 
																			SESH_make_general_index_to(hiero->debut_texte,0), 
																			x - 1));
		for(i=0; i<y; i++)
		{
			if (result->next)
				result= result->next;
		}
		return SESH_make_general_index_to(result,0);
	}
	else if (sscanf(index, "@%d,%d", &x, &y) == 2)
	{
		XPoint p;
		Texte *debut_ecran, *t;
		debut_ecran= 
			GETTEXTOF(SESH_get_marque(hiero, "ecran"));
		p.x=x;
		p.y=y;
		t= SESH_find_screen_position(hiero, debut_ecran, 
																 XPOINT_TO_TKSESH_POINT(hiero->affichage, p));
		return 
			SESH_make_general_index_to(t, 0);
	}
	else if (strcmp(index, "end") == 0)
	{
		Texte *t;
		t= hiero->montexte;
		while (t->next)
			t= t->next;
		return SESH_make_general_index_to(t, 0);
	}
	else if (sscanf(index, "%s.firs%c", nomtag, &c) == 2)
	{
		return SESH_make_general_index_to(hiero->debut_texte,0);
	}
	else if (sscanf(index, "%s.las%c", nomtag, &c) == 2)
	{
		return SESH_make_general_index_to(hiero->debut_texte,0);
	}
	else													/* mark */
	{
		return SESH_get_marque(hiero, index);
	}
}



PRIVATE GeneralIndex
SESH_index_modifier(Sesh *hiero,
										GeneralIndex g, 
										int argc, 
										char **argv, 
										int *pos)
{
	char c;
	Texte *t;
	int sig= GETSIGNPOSOF(g);										/* Number of the sign */
	t= GETTEXTOF(g);
	if (*pos + 2 < argc && ((c= *(argv[*pos])) == '+' || c == '-'))
		{
			int delta;
			delta= atoi(argv[*pos+1]);
			if (strcmp(argv[*pos+2],"quad") == 0)
				{
					int i;
					if (c == '+')
						for(i=0; t->next && i< delta; i++)
							t= t->next;
					else
						for(i=0; t->previous && i< delta; i++)
							t= t->previous;
				}
			else if (strcmp(argv[*pos+2],"line") == 0)
				{
					/* int i;*/
					if (c == '+')
						{
							/* if (t && t->next) t= t->next; */
							t= GETTEXTOF(SESH_moveby_lines(hiero, 
																					 SESH_make_general_index_to(t,0),
																					 delta));
						}
					else 
						{
							/* if t is NL, we want previous NL. If t is not NL, 
								 we want the NL before previous newline. That is
								 if t <> NL, delta = delta + 1 */
							if (t->type != T_NL && t->type != T_NP)
								delta++;
							t= GETTEXTOF(SESH_moveby_lines(hiero, 
																					 SESH_make_general_index_to(t,0),
																					 -delta));
						}
				}
			*pos+= 3;
		}
	else if (*pos < argc)
		{
			if (strcmp(argv[*pos],"linestart") == 0) /* goes to the beginning NL of the line */
				{
					int NOTFOUND = TRUE;
					while (NOTFOUND)
						{
							switch (t->type)
								{
								case T_NL:
								case T_NP:
									NOTFOUND= FALSE;
									break;
								default:
									t= t->previous;
								}
						}
					(*pos)++;
				}
			else if (strcmp(argv[*pos],"lineend") == 0) /* goes to the last sign of the line */
				{
					int NOTFOUND = TRUE;
					Texte *previous;
					if (t && t->next) {
						previous= t;
						t= t->next;
					}
					else {
						previous= t;
						t=t;
					}
					
					while (NOTFOUND)
						{
							if (t)
								switch (t->type)
									{
									case T_NL:
									case T_NP:
										NOTFOUND= FALSE;
										break;
									default:
										previous= t;
										t= t->next;
									}
							else
								NOTFOUND= FALSE;
						}
					t= previous;
					(*pos)++;
				}
			else if (strcmp(argv[*pos],"wordstart") == 0)
				{
					(*pos)++;
				}
			else if (strcmp(argv[*pos],"wordend") == 0)
				{
					(*pos)++;
				}
			else if (sscanf(argv[*pos], ".%d", &sig) == 1)
				{
					(*pos)++;
				}
		}
	return SESH_make_general_index_to(t,sig);
}


/**
 * INDEXES :
 * 
 * Indexes are manipulated through textual description.
 * They are mostly compatible with the indexes of text widgets
 * 
 * An index is the description of a place in the text.
 * The general form of indexes are :
 * 
 * PLACE MODIFIER
 * 
 * where PLACE is one of :
 *
 *   insert => the insertion point
 *   /DESCRIPTION/ => the textual description of a place in the text,
 *                    according to the description system used for it.
 *                    e.g. could be :
 *                    "colonne 1, 3"
 *                    or 
 *                    "page 5, ligne 7"
 *                    or
 *                    "mur nord, colonne 5, 7"
 *                    ...
 *
 *   LINE.QUADNUMBER => absolute coordinates
 *   @X,Y => screen coordinate (mouse position-compatible)
 *   end => end of text 
 *   TAG.first => beginning of TAG // NOT IMPLEMENTED
 *   TAG.end   => end of TAG // NOT IMPLEMENTED
 *
 *   MARK => the place marked with MARK
 * 
 * And MODIFIER is either nothing or :
 * 
 * ('+'|'-') NUMBER ('quad'|'line')
 * ('linestart'|'lineend')
 * ('wordstart'|'wordend')
 * 
 * '.' SIGNNUMBER
 *     (if SIGNNUMBER is 0 or is not set, it stands for "last sign")
 *
 * PUBLIC GeneralIndex
 * SESH_index(Sesh *hiero, int argc, char **argv, int *pos)
 *
 * 
 * Sends back a pointer to a textual item
 * from a textual index description.
 */

PUBLIC GeneralIndex
SESH_index(Sesh *hiero, int argc, char **argv, int *pos)
{
	GeneralIndex g;
	if (argc > *pos)
	{
		g= SESH_index_aux(hiero, argv[*pos]);
		(*pos)++;
	}
	g= SESH_index_modifier(hiero, g, argc, argv, pos);
	return g;
}


/**
 * PUBLIC void
 * SESH_trouve_index(Sesh *hiero, GeneralIndex g, char* index)
 *
 * sends back in char * index the index of the GeneralIndex g
 * in the form LINE.CADRAT
 *
 */

PUBLIC void
SESH_trouve_index(Sesh *hiero, GeneralIndex g, char* index)
{
	Texte * tt, *t;
	int ligne= 0;
	int horizontal= 0;
	t= GETTEXTOF(g);
	tt= hiero->debut_texte;
		
	while(tt != t)
	{
		switch (tt->type)
		{
		case T_NL:
			ligne++;
			horizontal= 0;
			break;
		case T_NP:
			ligne++;
			horizontal= 0;
			break;
		case T_FIXE_COORDONNEES:
		case T_LIGNE_COMMENTAIRE:
		case T_TEXTE:
			horizontal++;
			break;
		}
		tt= tt->next;
	}

	switch (tt->type)
	{
	case T_NL:
		ligne++;
		horizontal= 0;
		break;
	case T_NP:
		ligne++;
		horizontal= 0;
		break;
	case T_FIXE_COORDONNEES:
	case T_LIGNE_COMMENTAIRE:
	case T_TEXTE:
		horizontal++;
		break;
	}
	sprintf(index, "%d.%d", ligne, horizontal);
}



/**
 *
 *
 * SESH_compare_indexes
 *
 * takes two indexes i1 and i2, returns 0
 * if they are equal,
 * a positive number if i1 > i2 (i1 after i2)
 * a negative one if i1 < i2
 *
 */

PUBLIC int
SESH_compare_indexes(Sesh *hiero, GeneralIndex i1, GeneralIndex i2)
{
	Texte *t1, *t2;

	t1= GETTEXTOF(i1);
	t2= GETTEXTOF(i2);

	if (t1 == t2) {
		return i1.s - i2.s;
	} else {
		Texte * f, *b;
		int res= 0;
		/* We start from t1 with two pointers f(orward) and b(backward).
			 If f meets NULL, or b meets start, or either of them meets t2,
			 or the beginning of the text, we stop.  Then the result is
			 chosen accordingly */
		f= b= t1;
		while (f != 0 && b != hiero->debut_texte && 
					 f != t2 && b != t2)
			{
				b= b->previous;
				f= f->next;
			}
		if (f == NULL)
			res= 1;									/* i1 is after i2 */
		if (b == hiero->debut_texte)
			res= -1;
		if (f == t2)
			res= -1;
		if (b == t2)
			res= 1;
		return res;
	}
}



/**
 *
 *
 * The returned value is the absolute textual description of the
 * location 't' according to the coordinate system set for the given
 * text.
 *
 */

PUBLIC char *
SESH_trouve_index_texte(Sesh *hiero, GeneralIndex g)
{
	Texte * tt, *lastt, *t;
	coordsvect coords;
	int i= 0;

	t= GETTEXTOF(g);
	tt= hiero->debut_texte;
	
	for (i=0; i<10; i++)
		coords[i]= 0;

	coords[1]= 0;
	coords[2]= 1;
	do
	{
		switch (tt->type)
		{
		case T_NL:
			coords[1]++;							/* ligne ou colonne */
			break;
		case T_NP:
			coords[1]= 1;
			coords[2]++;							/* "page" */			
			break;
		case T_FIXE_COORDONNEES:
		{
			int i;
			for (i=0; i<10; i++)
				coords[i]= tt->coordonnees.coords[i];
			break;
		}
		case T_LIGNE_COMMENTAIRE:
			break;
		case T_TEXTE:								/* rien */
			break;
		}
		lastt= tt;
		tt= tt->next;
	} while (lastt != t);
	
	return get_text_coordinates(&(hiero->coordinates), coords);
}


/* compares two coordvectors */
PRIVATE int
coordinates_lt(coordsvect v1, coordsvect v2)
{
	int i;
	i = 9;
	while (i >= 0 && v1[i] == v2[i])
		i--;
	if (i == -1)
		return FALSE;
	else
		return (v1[i] < v2[i]);
}


/* le contraire : trouver un Texte * connaissant ses coordonnes */

PRIVATE Texte * 
get_text_from_coordinates(Coordinate_System * sys, char *descr, Texte *depart)
{
	coordsvect coords, current;
	Texte * tt, *lastt;
	int i= 0;

	set_text_coordinate(sys, descr, coords);
	tt= depart;
	
	for (i=0; i<10; i++)
		current[i]= 0;

	current[1]= 0;								/* On commence par un saut de ligne */
	current[2]= 1;

	do
	{
		switch (tt->type)
		{
		case T_NL:
			current[1]++;							/* ligne ou colonne */
			break;
		case T_NP:
			current[1]= 1;
			current[2]++;							/* "page" */			
			break;
		case T_FIXE_COORDONNEES:
		{
			int i;
			for (i=0; i<10; i++)
				current[i]= tt->coordonnees.coords[i];
			break;
		}
		case T_TEXTE:								/* rien */
			break;
		case T_LIGNE_COMMENTAIRE:
			break;
		}
		lastt= tt;
		tt= tt->next;
	} while (tt && coordinates_lt(current, coords));
	
	return lastt;
}


/**
 * PUBLIC void
 * SESH_largeur(Sesh *hiero, double *avant, double *apres)
 *
 * For scrollbars :
 * 
 * sets in avant and apres the percentage of horizontal 
 * text shown
 */

PUBLIC void
SESH_largeur(Sesh *hiero, double *avant, double *apres)
{
	double maxx;
	maxx= SESH_max_x_dim(hiero) * (hiero->echelle);
	if (maxx > 0.0){
		*avant=  hiero->origx/maxx;
		*apres= (hiero->origx + 
						 TKSOC(hiero->affichage)->w * (hiero->echelle))/maxx;
	}
	else
	{
		*avant= 0.0;
		*apres= 1.0;
	}
}


/**
 * PUBLIC void
 * SESH_hauteur(Sesh *hiero, double *avant, double *apres)
 *
 * For scrollbars :
 * 
 * sets in avant and apres the percentage of vertical
 * text shown
 */

PUBLIC void
SESH_hauteur(Sesh *hiero, double *avant, double *apres)
{
	double maxy;
	double first;
	maxy= hiero->max_lignes;
	first= SESH_line_number(hiero, SESH_get_marque(hiero, "ecran"));
	if (maxy > 0.0){
		*avant=  first/maxy;
		*apres= (first + hiero->nombre_de_lignes_affichees)/maxy;
	}
	else
	{
		*avant= 0.0;
		*apres= 1.0;
	}
}

/**
 *
 * Sesh_XUpdate --
 *
 *	Update the Horizontal scrollbar in case the display window X view change
 *  (using xScrollCmd)
 *  
 */

PUBLIC void
SESH_XUpdate(Sesh *SeshPtr)
{
	if (SeshPtr->xScrollCmd)
	{
		double avant, apres;
		PRIVATE char commande[255];
		int result;
		
		SESH_largeur(SeshPtr, &avant, &apres);
		sprintf(commande," %f %f", avant, apres);
		result = Tcl_VarEval(SeshPtr->interp, SeshPtr->xScrollCmd, commande,
												 NULL);
    if (result != TCL_OK) {
			Tcl_AddErrorInfo(SeshPtr->interp,
											 "\n    (horizontal scrolling by Sesh)");
			Tk_BackgroundError(SeshPtr->interp);
		}
	}
}


/**
 *
 * Sesh_YUpdate --
 *
 *	Update the vertical scrollbar in case the display window Y view change
 *  (calls yScrollCmd)
 *
 */

PUBLIC void
SESH_YUpdate(Sesh *SeshPtr)
{
	if (SeshPtr->yScrollCmd)
	{
		double avant, apres;
		PRIVATE char commande[255];
		int result;
		
		SESH_hauteur(SeshPtr, &avant, &apres);
		sprintf(commande," %f %f", avant, apres);
		result = Tcl_VarEval(SeshPtr->interp, SeshPtr->yScrollCmd, commande,
												 NULL);
    if (result != TCL_OK) {
			Tcl_AddErrorInfo(SeshPtr->interp,
											 "\n    (vertical scrolling by Sesh)");
			Tk_BackgroundError(SeshPtr->interp);
		}
	}
}



/**
 * PUBLIC void
 * SESH_get_index_position(Sesh* hiero, Texte *debut, int * x, int * y)
 *
 * Sends back the screen position of a textual item in x and y
 * for use with mouse and the like.
 * 
 */

PUBLIC void
SESH_get_index_position(Sesh* hiero, GeneralIndex gdebut, int * x, int * y)
{
	Texte *t;
	TKSESH_point m;
	XPoint pt;

	t= GETTEXTOF(gdebut);
	while ( !(t->first || t->etat == E_AFFICHE || t->etat == E_PARTIEL))
		t= t->previous;
	if (t->etat == E_AFFICHE || t->etat == E_PARTIEL)
	{
		m.x= (t->x);
		m.y= (t->y);
	}
	else
	{
		m.x= (0.0);
		m.y= (0.0);
	}
	pt= TKSESH_POINT_TO_XPOINT(hiero->affichage, m);
	*x= pt.x;
	*y= pt.y;
}



/*
 * DIMENSIONS related stuff
 */

/*
	 Maximum xdim of texte
*/

/**
 *
 * computes the length of the current line
 * (in pixels)
 *
 */

PUBLIC double
SESH_compute_line_length(Sesh *hiero, Texte **t)
{
	double curr;
	double resol;

	resol= hiero->resolution;
	
	/* find beginning of line. Should always work*/
	while ((*t)->type != T_NL && (*t)->type != T_NP)
		(*t)= (*t)->previous;
	curr= 0.1;

	if ((*t))
		(*t)=(*t)->next;

	while ((*t) && ((*t)->type != T_NL && (*t)->type != T_NP))
		{

			SESH_fix_dims(TKSOC(hiero->affichage), &hiero->geom, (*t));
			
			switch ((*t)->type)
				{
				case T_NL:
				case T_NP:
					break;
				case T_TEXTE:
					{
						if ((*t)->etat == E_NOUVEAU)
							{
								/*
									(*t)->w= largeur_quadra(&(hiero->affichage), *t, resol)+
									hiero->petit_espace;
									(*t)->etat= E_CACHE;
								*/
							}
						// (*t)->x= curr;
						curr+= SESH_texte_width(TKSOC(hiero->affichage), &hiero->geom, (*t));
					}
				break;
				case T_LIGNE_COMMENTAIRE:		/* a faire */
					break;
				default:
					break;
				}
			(*t)= (*t)->next;
		}
	return curr;
}

/**
 * Returns the width and height for a rectangle including the selection.
 *
 * PUBLIC void
 * SESH_compute_sel_dimensions(Sesh *hiero, GeneralIndex deb, GeneralIndex end, double *w, double *h)
 */

#define POSX0 0.1

void
SESH_compute_sel_dimensions(Sesh *hiero, GeneralIndex gdebut, GeneralIndex gfin, double *w, double *h)
{
  Texte * deb, *fin;
  Texte * tt;
  double posx, posy, dy;

  deb= GETTEXTOF(gdebut);
  fin= GETTEXTOF(gfin);
  SESH_ordonner_Texte(&deb, &fin);

	*w= 0.0;
	*h= 0.0;
	
  if (!deb->next)
    return;
	
  tt= deb->next;

  dy= 1.2;
  posx= POSX0;
  posy= -1;
  // posy-= dy;
	
  while (tt != fin->next)
    {
      switch (tt->type) 
				{
				case T_TEXTE:
					posx+= SESH_texte_width(TKSOC(hiero->affichage), &hiero->geom, tt);
					break;
				case T_NL:
				case T_NP:
					if(posx > *w)
						*w= posx;
					posy -= dy;
					posx= POSX0;
					break;
				case T_LIGNE_COMMENTAIRE:
					break;
				case T_FIXE_COORDONNEES:
					break;
				}
      tt= tt->next;
    }
	posy-= 0.1;
	if(posx > *w)
		*w= posx;
	*h= -posy;
}


/**
 * PUBLIC int
 * SESH_max_x_dim(Sesh *hiero)
 *
 * Sends back the maximum x dimension of text.
 */

PUBLIC int
SESH_max_x_dim(Sesh *hiero)
{
	double max =-1;
	double curr= 0;
	int num, maxlignes;
	Texte *t;

	/*t= hiero->debut_texte;*/
	t= SESH_start_of_screen(hiero);

	if (hiero->wrap)
		return TKSOC(hiero->affichage)->w;

	num= -1;
	maxlignes= hiero->nombre_de_lignes_affichees;

	while (t &&  num <= maxlignes)
	{
		curr= SESH_compute_line_length(hiero,&t);
		if (curr > max)
			max= curr;
	}
	if (curr > max)
		max= curr; 
	return max * hiero->htline + 1;
}



/**
 * PUBLIC int
 * SESH_max_y_dim(Sesh *hiero)
 *
 * Sends back the maximum y dimension of text.
 */

PUBLIC int
SESH_max_y_dim(Sesh *hiero)
{
	return (hiero->max_lignes>0?hiero->max_lignes * hiero->htline:hiero->htline);
		/* SESH_line_number(hiero, NULL) * hiero->htline;*/
}



/**
 * 
 * Sends back the line number (from the beginning of the text)
 * of GeneralIndex ghere
 */

PUBLIC int
SESH_line_number(Sesh *hiero, GeneralIndex ghere)
{
	int num = 0;
	Texte *t, *here;
	here= GETTEXTOF(ghere);
	t= hiero->debut_texte;
	while (t && t != here)
	{
		num += SESH_jump_of(hiero, t);
		t= t->next;
	}
	return num;
}



/**
 * int
 * SESH_is_end_of_line(Texte *t)
 *
 * Tells whether a textual item is a line separator or not.
 * Lines ends are LOGICAL lines separators, not graphical.
 * They correspond to -! and -!!
 */

PRIVATE int
SESH_is_end_of_line(Texte *t)
{
	if (t== NULL)
		return TRUE;
	
	return (
					t->type == T_NL
					|| t->type == T_NP);
}



/**
 * int
 * SESH_jump_of(Sesh * hiero, Texte *t)
 *
 * Used to know how many lines are spanned by a textual item.
 */

PUBLIC int
SESH_jump_of(Sesh * hiero, Texte *t)
{
	int num= 0;
	switch (t->type)
	{
		case T_NL:
			num= 1;
			break;
		case T_NP:
				num= 2;
			break;
		case T_FIXE_COORDONNEES:
			num= 1;
			break;
		case T_LIGNE_COMMENTAIRE:
			if (hiero->display_comments)
				num= 1;
			break;
		case T_TEXTE:
			break;
		}
	return num;
}



PUBLIC int
SESH_get_screen_pos(Sesh *hiero, GeneralIndex gt1, GeneralIndex gt2)
{
	Texte *t;
	int num= 0;
	t= GETTEXTOF(gt1);

	/* if the first cadrat is a newpage or newline,
		 we substract 1 from the result. */
	
	if (t->type == T_NP)
		num= 1;
	return num+ SESH_get_number_of_lines(hiero, gt1, gt2);
}

/**
	RMRQ IMPORTANTE : void MDCTexte pour la dfinition des zones.

	SESH_get_number_of_lines is used to know how many lines 
	we have in a text zone. 
	
	gt1..............................gt2
	   [-------------------------------]

	the zone starts with gt1 (excluded) and ends with gt2 (included).

	new (line|page) are part of the line they start.

	An "internal" new line (or page) is worth 1 (resp. 2) lines.
	An "initial" new line (or page) is worth 0 (resp. 1) line ? NO ?
	
	*/


PUBLIC int 
SESH_get_number_of_lines(Sesh *hiero, GeneralIndex gt1, GeneralIndex gt2)
{
	Texte *apres, *t;
	Texte *t1, *t2, *t1a;
	int num= 0;
	t1= t1a= GETTEXTOF(gt1);
	t2= GETTEXTOF(gt2);
	SESH_ordonner_Texte(&t1, &t2);

	if (!t1->next)
		return 0;
	
	if (t2)
 		apres= t2->next;
 	else
 		apres= t2;
 
	t= t1->next;
	/* t=t1; */
	
	/* if the first cadrat is a newpage or newline,
		 we substract 1 from the result. */
	/*
		if (t->type == T_NL 
		|| t->type == T_NP)
		num= -1;
	*/
	
	while (t!= apres)
	{
		num+= SESH_jump_of(hiero, t);
		t= t->next;
	}
	if (t1a == t1)
		return num;
	else
		return -num;
}


/**
 *
 * sends back a pointer to a Texte deltaN lines from "depart"
 *
 */

PUBLIC GeneralIndex
SESH_moveby_lines(Sesh * hiero, GeneralIndex depart, int deltaN)
{
	Texte *t;
	t= GETTEXTOF(depart);
	if (deltaN > 0)
	{
		int i;
		for (i=0; t && t->next && i< deltaN; i++)
			/* i+= */ SESH_goto_next_line(hiero, t, &t);
	}
	else
	{
		int i;
		for (i=0; (! t->first) && i< -deltaN; i++)
			/* i+= */ SESH_goto_previous_line(hiero, t, &t);
	}
	return SESH_make_general_index_to(t,0);
}



/**
 *
 * sends back a pointer to a Texte deltaN lines from "depart",
 * counting lines as displayed on the screen.
 *
 */

PUBLIC GeneralIndex
SESH_moveby_screenlines(Sesh * hiero, GeneralIndex depart, int deltaN)
{
	Texte *t;
	t= GETTEXTOF(depart);
	if (deltaN > 0)
	{
		int i;
		for (i=0; t && t->next && i< deltaN; i++)
			i+= SESH_goto_next_line(hiero, t, &t);
	}
	else
	{
		int i;
		for (i=0; (! t->first) && i< -deltaN; i++)
			i+= SESH_goto_previous_line(hiero, t, &t);
	}
	return SESH_make_general_index_to(t,0);
}



/*
	 danschar return 0 if p is not in t,
	 1 if t is in the first half of t,
	 and 2 if it is in the second half
*/

PRIVATE int danschar(Sesh *hiero, Texte * t, TKSESH_point p)
{
	double x, y;
	x= (p.x);
	y= (p.y);
	if (t->y > y)
		return 0;
	else if (t->x < x)
		if (t->x + SESH_texte_width(TKSOC(hiero->affichage), &hiero->geom, t)/2.0 >= x)
			return 1;
		else if (t->x + SESH_texte_width(TKSOC(hiero->affichage), &hiero->geom, t) >= x)
			return 2;
		else return 0;
	else return 0;
}



/**
 * Texte *
 * SESH_find_screen_position(Texte * depart, TKSESH_point p)
 * 
 * given a point with REAL coordinates (NOT INTEGER),
 * sends back a pointer to the corresponding text.
 *
 */

PUBLIC Texte *
SESH_find_screen_position(Sesh * hiero, Texte * depart, TKSESH_point p)
{
	Texte *t, *tprev, *result;
	int pastrouve;

	t= depart;
	tprev= depart;
	result= depart;
	pastrouve= TRUE;
	while (t && pastrouve)
	{
		int ou;
		if ((ou= danschar(hiero, t, p)))
		{
			pastrouve= FALSE;
			if(ou == 1)
				result= tprev;
			else
				result= t;
		}
		tprev= t;
		t= t->next;
	}
	return result;
}



/**
 * PUBLIC int
 * SESH_goto_previous_line(Sesh *hiero, Texte *here, Texte ** but)
 * 
 * goes to the previous line. 
 * As a line might span more than one screen line, sends back the
 * number of spanned lines
 * 
 */

PUBLIC int
SESH_goto_previous_line(Sesh *hiero, Texte *here, Texte ** but)
{
	Texte *t= here;
	int screen_lines=0;

	if (t->first)
	{
		*but= t;
		return 1;
	}
	else
		t= t->previous;
	
	while (!t->first && !SESH_is_end_of_line(t))
		{
			screen_lines+= SESH_jump_of(hiero, t);
			t= t->previous;
		}

	*but= t;
	return screen_lines+ SESH_jump_of(hiero, t);
}



/**
 * PUBLIC int
 * SESH_goto_next_line(Sesh *hiero, Texte *here, Texte ** but)
 * 
 * goes to the next line. 
 * As a line might span more than one screen line, sends back the
 * number of spanned lines
 * 
 */
	
PUBLIC int
SESH_goto_next_line(Sesh* hiero, Texte *here, Texte **but)
{
	Texte *t= here;
	Texte *av= here;
	int screen_lines=0;
	
	if (t)
		t= t->next;
	
	while (t &&  !SESH_is_end_of_line(t))
	{
		av=t;
		screen_lines+= SESH_jump_of(hiero, t);
		t= t->next;
	}

	if (t)
	{
		*but= t;
		return screen_lines + SESH_jump_of(hiero, t);
	}
	else
	{
		*but= av;
		return screen_lines;
	}		
}

/**
 * Sends a pointer to the beginning of the screen
 */

PUBLIC Texte *
SESH_start_of_screen(Sesh * hiero)
{
	return GETTEXTOF(SESH_get_marque(hiero, "ecran"));
}


/**
 * Increments the counter which is used to know if current cadrats are
 * up to date 
 */

PUBLIC void
SESH_increment_compteur(Sesh *hiero) {
	hiero->precedent_compteur= hiero->compteur;
	if (hiero->compteur < ULONG_MAX - 1)
		hiero->compteur++;
	else
		hiero->compteur=0;
}

/*
 * Local Variables:
 * header: "tksaux.h"
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
