#define NO_CONST 1
#define _POSIX_SOURCE 1

/* This tk module implements a hieroglyphic editor. It uses another
 *	 module, called Bfont, which implements vectorial fonts.

 * The frame of this widget is taken from the Square widget example
 * from TK distribution
 *
 * The sesh widget is (c) 1995,1996,1997,1998 S. Rosmorduc
 *
 */

/*
 * The Square widget is
 * Copyright (c) 1991-1994 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*  1995, 2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: tksesh.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/*
 * $Log: tksesh.c,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 * Revision 1.34  2000/10/31 17:01:17  rosmord
 * *** empty log message ***
 *
 * Revision 1.33  2000/08/23 08:36:29  rosmord
 * fixed problems with WMF file sizes
 *
 * Revision 1.32  2000/08/04 11:53:15  rosmord
 * modified to support multiple devices
 *
 * Revision 1.31  1999/10/04 16:12:59  rosmord
 * first version with complete header
 *
 */


/* #include "tkPort.h"*/

#include <limits.h>
#include "general.h"
#include "seshaux.h"						/* A deplacer dans MDCTexte ??? */
#include "tksesh.h"
#include "tksaux.h"
#include "signs.h"
#include "tksaff.h"
#include "TKSO_tkwin.h"
#include "TKSO_WMF.h"
#include "TKSO_bitmap.h"
#include "tks2dev.h"
#include "ligatur.h"
#include "tksinter.h"

#if defined(__WIN32__)
#include "tkWinPort.h"
#elif defined(MAC_TCL)
#include "tkMacPort.h"
#else
#include "tkUnixPort.h"
#endif

#if defined(__WIN32__)
#include "TKSO_WCB.h"
#endif

/*
 * Information used for argv parsing.
 */

PRIVATE Tk_ConfigSpec configSpecs[] = {
	/* decors */
	{TK_CONFIG_BORDER, "-background", "background", "Background",
	 "#d9d9d9", Tk_Offset(Sesh, bgBorder), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_BORDER, "-background", "background", "Background",
	 "white", Tk_Offset(Sesh, bgBorder), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	 (char *) NULL, 0, 0},
	{TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
	 (char *) NULL, 0, 0},
	{TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	 "2", Tk_Offset(Sesh, borderWidth), 0},
	{TK_CONFIG_SYNONYM, "-fg", "foreground", (char *) NULL,
	 (char *) NULL, 0, 0},
	{TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground",
	 "black", Tk_Offset(Sesh, foreground), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground",
	 "black", Tk_Offset(Sesh, foreground), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_COLOR, "-redcolor", "redcolor", "RedColor",
	 "red", Tk_Offset(Sesh, redcolor), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-redcolor", "redcolor", "RedColor",
	 "black", Tk_Offset(Sesh, redcolor), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_COLOR, "-graycolor", "graycolor", "GrayColor",
	 "gray", Tk_Offset(Sesh, graycolor), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-graycolor", "graycolor", "GrayColor",
	 "white", Tk_Offset(Sesh, redcolor), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_COLOR, "-gramcolor", "gramcolor", "GramColor",
	 "yellow", Tk_Offset(Sesh, gramcolor), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-gramcolor", "gramcolor", "GramColor",
	 "black", Tk_Offset(Sesh, gramcolor), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_COLOR, "-gramendcolor", "gramendcolor", "GramendColor",
	 "darkgreen", Tk_Offset(Sesh, gramendcolor), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-gramendcolor", "gramendcolor", "GramendColor",
	 "black", Tk_Offset(Sesh, gramendcolor), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_COLOR, "-cursorcolor", "cursorcolor", "CursorColor",
	 "blue", Tk_Offset(Sesh, cursorcolor), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_COLOR, "-cursorcolor", "cursorcolor", "CursorColor",
	 "black", Tk_Offset(Sesh, cursorcolor), TK_CONFIG_MONO_ONLY},
	/* dimensions */
	{TK_CONFIG_INT, "-htline", "htline", "Htline",
	 "40", Tk_Offset(Sesh, htline), 0},
	{TK_CONFIG_INT, "-linesep", "linesep", "Linesep",
	 "10", Tk_Offset(Sesh, linesep), 0},
	{TK_CONFIG_DOUBLE, "-origx", "origx", "Origx",
	 "0.0", Tk_Offset(Sesh, origx), 0},
	/* {TK_CONFIG_DOUBLE, "-origy", "origy", "Origy",
		 "0.0", Tk_Offset(Sesh, origy), 0},*/
	{TK_CONFIG_INT, "-width", "width", "Width",
	 "300", Tk_Offset(Sesh, largeur), 0},
	{TK_CONFIG_INT, "-height", "height", "Height",
	 "40", Tk_Offset(Sesh,hauteur), 0},
	{TK_CONFIG_BOOLEAN, "-in_columns", "in_columns", "in_columns",
	 "0", Tk_Offset(Sesh,geom.in_columns), 0},
	{TK_CONFIG_BOOLEAN, "-right_to_left", "right_to_left", "right_to_left",
	 "0", Tk_Offset(Sesh,geom.right_left), 0},
	{TK_CONFIG_BOOLEAN, "-vertical_centering", "vertical_centering", "vertical_centering",
	 "0", Tk_Offset(Sesh,geom.vcenter), 0},
	{TK_CONFIG_BOOLEAN, "-show_words", "show_words", "Show_words",
	 "0", Tk_Offset(Sesh,geom.show_words), 0},
	{TK_CONFIG_BOOLEAN, "-display_comments", "display_comments", "Display_comments",
	 "1", Tk_Offset(Sesh,display_comments), 0},
	{TK_CONFIG_BOOLEAN, "-keep_old", "keep_old", "Keep_old",
	 "1", Tk_Offset(Sesh,keep_old), 0},
	{TK_CONFIG_BOOLEAN, "-draw_cursor", "draw_cursor", "Draw_cursor",
	 "1", Tk_Offset(Sesh,draw_cursor), 0},
	{TK_CONFIG_BOOLEAN, "-takefocus", "take_focus", "Take_Focus",
	 "1", Tk_Offset(Sesh,takefocus), 0},
	{TK_CONFIG_BOOLEAN, "-wrap", "wrap", "Wrap",
	 "0", Tk_Offset(Sesh,wrap), 0},
	{TK_CONFIG_STRING, "-xscrollcommand", "xScrollCommand", "ScrollCommand",
	 "", Tk_Offset(Sesh, xScrollCmd), TK_CONFIG_NULL_OK},
	{TK_CONFIG_STRING, "-yscrollcommand", "yScrollCommand", "ScrollCommand",
	 "", Tk_Offset(Sesh, yScrollCmd), TK_CONFIG_NULL_OK},
	{TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	 (char *) NULL, 0, 0}
};


/*
 * Forward declarations for procedures defined later in this file:
 */

PRIVATE void		SeshCmdDeletedProc _ANSI_ARGS_((
			    ClientData clientData));

PRIVATE int		SeshConfigure _ANSI_ARGS_((Tcl_Interp *interp,
			    Sesh *SeshPtr, int argc, char **argv,
			    int flags));

PRIVATE Tcl_FreeProc SeshDestroy;
/* PRIVATE void		SeshDestroy _ANSI_ARGS_((ClientData clientData)); */

PRIVATE void		SeshDisplay _ANSI_ARGS_((ClientData clientData));

PRIVATE void		SeshEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));

PRIVATE int		SeshWidgetCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *, int argc, char **argv));




PUBLIC void
SeshMyInit(Tcl_Interp *interp)
{
	/*char *libs;*/
	/* char repertoires_fontes[255];*/
	struct tksesh_global *g;

	g= tksesh_get_global();
	initialise_lex();							/* lexical stuff. Char names & co */
	initialise_quadras();					/* memory allocations for quadra */
	SESH_init_textes();						/* idem for texts */
}


/*
 *--------------------------------------------------------------
 *
 * SeshCmd --
 *
 *	This procedure is invoked to process the "Sesh" Tcl
 *	command.  It creates a new "Sesh" widget.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	A new widget is created and configured.
 *
 *--------------------------------------------------------------
 */

PUBLIC int
SeshCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
	Tk_Window mainwindow = (Tk_Window) clientData;
	
	Sesh *SeshPtr;
	Tk_Window tkwin;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args:  should be \"",
										 argv[0], " pathName ?options?\"", (char *) NULL);
		return TCL_ERROR;
	}

	tkwin = Tk_CreateWindowFromPath(interp, mainwindow, argv[1], (char *) NULL);
	if (tkwin == NULL) {
		return TCL_ERROR;
	}
	Tk_SetClass(tkwin, "Sesh");

	/*
	 * Allocate and initialize the widget record.
	 */

	SeshPtr = (Sesh *) ALLOUER(sizeof(Sesh));
	SeshPtr->tkwin = tkwin;
	SeshPtr->display = Tk_Display(tkwin);
	SeshPtr->interp = interp;
	SeshPtr->widgetCmd =
		Tcl_CreateCommand(interp,
											Tk_PathName(SeshPtr->tkwin), SeshWidgetCmd,
											(ClientData) SeshPtr, SeshCmdDeletedProc);

	/*
	 * There be Thyngues special to our widget
	 */

	SeshPtr->borderWidth = 0;
	SeshPtr->bgBorder = NULL;
	SeshPtr->foreground = NULL;
	SeshPtr->cursorcolor = NULL;
	SeshPtr->redcolor = NULL;
	SeshPtr->graycolor = NULL;
	SeshPtr->gramcolor = NULL;
	SeshPtr->gramendcolor = NULL;
	SeshPtr->palette= NULL;
	SeshPtr->relief = TK_RELIEF_FLAT;
	SeshPtr->updatePending = 0;
	SeshPtr->xScrollCmd= NULL;
	SeshPtr->yScrollCmd= NULL;


	/* There be Thyngues concerning the text itself */

	SeshPtr->debut_texte= MDC_FirstTexte();
	SeshPtr->old_debut_ecran= NULL;
	SeshPtr->montexte= SeshPtr->debut_texte;
	SeshPtr->les_marques= NULL;

	SESH_set_marque(SeshPtr, "ecran",
									SESH_make_general_index_to(SeshPtr->debut_texte,0));

	SeshPtr->nombre_de_lignes_affichees= 1;
	SeshPtr->max_lignes= 1;
	SeshPtr->old_max_lignes= 0;
	SeshPtr->htline=  40;
	SeshPtr->linesep= 0;
	SeshPtr->largeur= 300;
	SeshPtr->hauteur= 400;
	SeshPtr->old_w=0;
	SeshPtr->old_h=0;
	SeshPtr->origx= 0.0;
	SeshPtr->old_maxx= 0.0;
	SeshPtr->maxx= 0.0;

	SeshPtr->resolution= 1.0;
	SeshPtr->petit_espace= 0.1;

	SeshPtr->compteur= 0L;
	SeshPtr->precedent_compteur= ULONG_MAX - 1;
	TKSGeom_init(&SeshPtr->geom);
	SeshPtr->display_comments=0;
	SeshPtr->draw_cursor=0;
	SeshPtr->wrap=0;
	SeshPtr->display_needed= TRUE;
	SeshPtr->affichage= TKSO_tkwin_init(SeshPtr->display,
																			0.0, -3.0, 20.0, 0.0, 
																			300, 30, 
																			SeshPtr->tkwin,
																			SeshPtr->interp);
	TKSO_SET_SCALE(SeshPtr->affichage, SeshPtr->htline);

	SeshPtr->old_affichage= TKSO_tkwin_init(SeshPtr->display,
																					0.0, -3.0, 20.0, 0.0, 
																					300, 30, 
																					SeshPtr->tkwin,
																					SeshPtr->interp);
	TKSO_SET_SCALE(SeshPtr->old_affichage, SeshPtr->htline);	
	SeshPtr->past_affichage= 0;
	SeshPtr->past_affichage_is_usable= FALSE;
	SeshPtr->keep_old= FALSE;
	
	init_text_coordinates(&(SeshPtr->coordinates), SeshPtr->interp);

	Tk_CreateEventHandler(SeshPtr->tkwin, ExposureMask|StructureNotifyMask,
												SeshEventProc, (ClientData) SeshPtr);

	if (SeshConfigure(interp, SeshPtr, argc-2, argv+2, 0) != TCL_OK) {
		Tk_DestroyWindow(SeshPtr->tkwin);
		return TCL_ERROR;
	}

	interp->result = Tk_PathName(SeshPtr->tkwin);
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SeshWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *
 * commands :
 * mark (names|set NAME INDEX|unset NAME)
 *       e.g. .s mark set insert insert - 1 quad
 * delete INDEX1 ?INDEX2?
 * get    INDEX1 INDEX2
 * insert INDEX SIGN						 pour l'instant
 * insert INDEX TEXT
 *  grh <INDEX1> <INDEX2>					 ok.
 *  grv <INDEX1> <INDEX2>					 ok.
 *  addline INDEX
 *  xview
 *  yview
 *  explode	                 ok.
 *  explodev                 ok.
 *  help	                   ok.
 *  setfirstline	           ok.
 *  getfirstline	           ok.
 *  dimx	                   ok.
 *  dimy	                   ok.
 *  toggle_word	             (rajouter INDEX)
 *
 *
 *--------------------------------------------------------------
 */

PRIVATE int
Getdebutfin(Sesh *SeshPtr,
						GeneralIndex *debut,
						GeneralIndex *fin,
						int argc,
						char ** argv,
						int *pos)
{
	*debut= SESH_index(SeshPtr, argc, argv, pos);
	if (*pos < argc)
		*fin= SESH_index(SeshPtr, argc, argv, pos);
	else
		*fin= SESH_get_marque(SeshPtr, "end");
	return (GETTEXTOF(*debut) && GETTEXTOF(*fin));
}

PRIVATE int
Getdebut(Sesh *SeshPtr,
				 GeneralIndex *debut,
				 int argc,
				 char ** argv,
				 int *pos)
{
	*debut= SESH_index(SeshPtr, argc, argv, pos);
	return (GETTEXTOF(*debut) != NULL );
}

PRIVATE int
tksmark(ClientData clientData,
				Tcl_Interp  *interp,
				int argc,				/* Number of arguments. */
				char **argv		/* Argument strings. */
				)
{
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " mark", NULL);
		return TCL_ERROR;
	}
	switch (argv[2][0])
	{
	case 'n':
	{
		Marque *m;
		m= SeshPtr->les_marques;
		while (m)
		{
			Tcl_AppendResult(interp, m->nom, NULL);
			m= m->next;
		}
	}
		break;
	case 's':											/* mark set NAME INDEX */
	{
		int pos= 4;
		if (argc < 5)
			return TCL_ERROR;
		if (strcmp("ecran", argv[3]) == 0)
			SeshPtr->display_needed= TRUE;
		SESH_set_marque(SeshPtr, argv[3],
										SESH_index(SeshPtr, argc, argv, &pos));
	}
		break;
	case 'u':
		if (argc < 4)
			return TCL_ERROR;
		SESH_detruire_marque(SeshPtr, argv[3]);
		break;
	default:
		Tcl_AppendResult(interp, "wrong commands: ",
										 argv[0], " mark ", argv[2], NULL);
		return TCL_ERROR;
	}
	return  TCL_OK;
}

PRIVATE int
tksdelete(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about Sesh widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
	int pos;
	GeneralIndex debut, fin;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
		{
			int delta;
			delta= SESH_get_number_of_lines(SeshPtr, debut, fin);
			delta= delta>0?(delta):(-delta);
			SeshPtr->max_lignes -= delta;
			SeshPtr->montexte= detruire_Texte(debut, fin);
			SeshPtr->display_needed= TRUE;
		}
	else
		{
			Tcl_AppendResult(interp, "wrong index ", NULL);
			return TCL_ERROR;
		}
	return TCL_OK;
}


PRIVATE int
tksexplode(ClientData clientData,		/* Information about Sesh widget. */
					 Tcl_Interp *interp,			/* Current interpreter. */
					 int argc,								/* Number of arguments. */
					 char **argv							/* Argument strings. */
					 )
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			char *s;
			s= SESH_breaktext(debut);
			Tcl_AppendResult(interp, s, NULL);
			LIBERE(s);
		}
	else
		{
			Tcl_AppendResult(interp, "wrong index ", NULL);
			return TCL_ERROR;
		}
	return TCL_OK;
}

PRIVATE int
tkslistcad(ClientData clientData,		/* Information about Sesh widget. */
					 Tcl_Interp *interp,			/* Current interpreter. */
					 int argc,								/* Number of arguments. */
					 char **argv							/* Argument strings. */
					 )
{
	int pos;
	GeneralIndex debut, fin;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}

	pos= 2;
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
		{
			char *s;
			s= SESH_breakspan(debut, fin);
			Tcl_AppendResult(interp, s, NULL);
			LIBERE(s);
		}
	else
		{
			Tcl_AppendResult(interp, "wrong index ", NULL);
			return TCL_ERROR;
		}
	return TCL_OK;
}

PRIVATE int
tks2wmf(ClientData clientData,		/* Information about Sesh widget. */
				Tcl_Interp *interp,			/* Current interpreter. */
				int argc,								/* Number of arguments. */
				char **argv							/* Argument strings. */
				)
{
	int pos;
	GeneralIndex debut, fin;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " tks2wmf", NULL);
			return TCL_ERROR;
		}

	pos= 2;
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
		{
			double width, height;
			FILE * f= fopen(argv[pos], "wb");
			TKSO_WMF *mw;
			SESH_compute_sel_dimensions(SeshPtr, debut, fin, &width, &height);
			mw= TKSO_WMF_init(f, 0.0, -height, width, 0.0, 14.0);
			tks2dev(SeshPtr, TKSOC(mw), debut, fin);
			TKSO_WMF_free(mw);
			fclose(f);
		}
	else
		{
			Tcl_AppendResult(interp, "wrong index ", NULL);
			return TCL_ERROR;
		}
	return TCL_OK;
}

#if defined(__WIN32__)
PRIVATE int
tks2wcb(ClientData clientData,		/* Information about Sesh widget. */
				Tcl_Interp *interp,			/* Current interpreter. */
				int argc,								/* Number of arguments. */
				char **argv							/* Argument strings. */
				)
{
	int pos;
	GeneralIndex debut, fin;
	double lineHeight; /* line height (in mm) */

	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 4)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " tks2wcb [-lineHeight H] deb end", 
											 "\n\twhere H is expressed in mm",
											 NULL);
			return TCL_ERROR;
		}

	lineHeight= 5;
	pos= 2;
	
	if (argv[2][0] == '-' && argv[2][1]== 'l') 
		{
			if (sscanf(argv[3], "%lf", &lineHeight) != 1)
				{
					Tcl_AppendResult(interp, "wrong args. Expecting double in ", argv[0],
													 " tks2wcb -lineHeight : ", argv[3], NULL);
				}
			pos+= 2;
		}
	
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
		{
			double width, height;
			TKSO_WCB *mw;
			SESH_compute_sel_dimensions(SeshPtr, debut, fin, &width, &height);
			/* create a window clipboard device */
			/* Currently doesn't use Enhanced Metafiles ! */
			mw= TKSO_WCB_init(0.0, -height, width, 0.0, lineHeight, 0);
			/* draw in this device */
			tks2dev(SeshPtr, TKSOC(mw), debut, fin);
			/* release the device */
			TKSO_WCB_free(mw);
		}
	else
		{
			Tcl_AppendResult(interp, "wrong index ", NULL);
			return TCL_ERROR;
		}
	return TCL_OK;
}
#endif

PRIVATE int
tks2bitmap(ClientData clientData,		/* Information about Sesh widget. */
				Tcl_Interp *interp,			/* Current interpreter. */
				int argc,								/* Number of arguments. */
				char **argv							/* Argument strings. */
				)
{
	int pos;
	GeneralIndex debut, fin;
	Sesh *SeshPtr = (Sesh *) clientData;
	enum BitmapType {PBM, XBM} bitmapType;

	if (argc < 4)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " \n ", 
											 "Usage : ", argv[0], "?-height height", NULL);
			return TCL_ERROR;
		}

	bitmapType= PBM;
	if (argv[1][2] == 'x')
		bitmapType= XBM;

	pos= 2;
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
		{
			double width, height;
			Tcl_Obj *result;
			TKSO_bitmap *mw;
			SESH_compute_sel_dimensions(SeshPtr, debut, fin, &width, &height);
			mw= TKSO_bitmap_init(0.0, -height, width, 0.0, width*SeshPtr->htline, height*SeshPtr->htline);
			tks2dev(SeshPtr, TKSOC(mw), debut, fin);

			if (bitmapType == PBM)
				result= TKSO_bitmap_getPBM(mw);
			else 
				result= TKSO_bitmap_getXBM(mw);
			Tcl_SetObjResult(interp, result);	
			TKSO_bitmap_free(mw);
		}
	else
		{
			Tcl_AppendResult(interp, "wrong index ", NULL);
			return TCL_ERROR;
		}
	return TCL_OK;
}


PRIVATE int
tksget(ClientData clientData,		/* Information about Sesh widget. */
			 Tcl_Interp *interp,			/* Current interpreter. */
			 int argc,								/* Number of arguments. */
			 char **argv							/* Argument strings. */
			 )
{
	int pos;
	GeneralIndex debut, fin;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
		return TCL_ERROR;
	}
	pos= 2;
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
	{
		char *s;
		s= SESH_texte_to_string(debut, fin);
		Tcl_AppendResult(interp, s, NULL);
		LIBERE(s);
	}
	return TCL_OK;
}

PRIVATE int
tksxypos(ClientData clientData,	/* Information about Sesh widget. */
				 Tcl_Interp *interp,			/* Current interpreter. */
				 int argc,								/* Number of arguments. */
				 char **argv							/* Argument strings. */
				 )
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " xypos", NULL);
		return TCL_ERROR;
	}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
	{
		int x, y;
		char s[100];
		SESH_get_index_position(SeshPtr, debut, &x, &y);
		sprintf(s,"%d %d", x, y);
		Tcl_AppendResult(interp,s, NULL);
		return TCL_OK;
	}
	else
	{
		return TCL_ERROR;
	}
}

PRIVATE int
tksinsert(ClientData clientData,		/* Information about Sesh widget. */
							Tcl_Interp *interp,			/* Current interpreter. */
							int argc,								/* Number of arguments. */
							char **argv							/* Argument strings. */
							)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 4)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " insert", NULL);
		return TCL_ERROR;
	}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
	{
		Texte *txt;
		sesh_set_coordinate_system(&(SeshPtr->coordinates));
		txt= (Texte *) sesh_parse_array(argv[pos]);
		if (txt)
		{
			int delta;
			SeshPtr->montexte= SESH_insere_texte(debut,txt);
			delta= SESH_get_number_of_lines(SeshPtr,
																	 debut,
																	 SESH_make_general_index_to(SeshPtr->montexte,0));
			SeshPtr->max_lignes+= (delta>0)?(delta):(-delta);
			SeshPtr->display_needed= TRUE;
			return TCL_OK;
		}
		else {
			extern int erreur_de_syntaxe;
			extern char rapport_erreur_de_syntaxe[];
			if (erreur_de_syntaxe)
			{
				Tcl_AppendResult(interp, rapport_erreur_de_syntaxe, NULL);
				return TCL_ERROR;
			}
			else
				return TCL_OK;
		}
	}
	else
		return  TCL_ERROR;
}

PRIVATE int
tksindex(ClientData clientData,		/* Information about Sesh widget. */
				 Tcl_Interp *interp,			/* Current interpreter. */
				 int argc,								/* Number of arguments. */
				 char **argv							/* Argument strings. */
				 )
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " index", NULL);
		return TCL_ERROR;
	}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
		char resultat[200];
		SESH_trouve_index(SeshPtr, debut, resultat);
		Tcl_AppendResult(interp, resultat, NULL);
		return TCL_OK;
	}
	else
		return  TCL_ERROR;
}

PRIVATE int
tksshow(ClientData clientData,		/* Information about Sesh widget. */
				Tcl_Interp *interp,			/* Current interpreter. */
				int argc,								/* Number of arguments. */
				char **argv							/* Argument strings. */
				)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " show", NULL);
		return TCL_ERROR;
	}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
	{
		double x;
		/* TKSESH_isCadratHorizontallyVisible requires that 
			 TKSESH_redrawWidget has tried to draw the cadrat.
		*/
		
		/*
		 * Other possibility (maybe better. We really need a display to compute such things as 
		 * text length).
		 * 			if (!Tk_IsMapped(SeshPtr->tkwin)) 
		 *  		{ 
		 *  			Tk_MakeWindowExist(SeshPtr->tkwin); 
		 *  			TKSESH_redrawWidget(SeshPtr); 
		 *  		} 
		 */
		if (Tk_IsMapped(SeshPtr->tkwin))
			TKSESH_redrawWidget(SeshPtr);
		
		if (! TKSESH_isCadratHorizontallyVisible(SeshPtr, debut, &x))
			{
					char neworig[100];
					char *myarg[2];
					myarg[0]= "-origx";
					myarg[1]= neworig;
					sprintf(neworig, "%f", x);
					SeshConfigure(interp, SeshPtr, 2, myarg, TK_CONFIG_ARGV_ONLY);
					SeshPtr->display_needed= TRUE;
			}
		
		if (!  TKSESH_isCadratVerticallyVisible(SeshPtr, debut, &debut))
			{
				SESH_set_marque(SeshPtr, "ecran", debut);
				SeshPtr->display_needed= TRUE;
			}
		return TCL_OK;
	}
	else
		return  TCL_ERROR;
}

PRIVATE int
tksxview(ClientData clientData,		/* Information about Sesh widget. */
				 Tcl_Interp *interp,			/* Current interpreter. */
				 int argc,								/* Number of arguments. */
				 char **argv							/* Argument strings. */
				 )
{
	Sesh *SeshPtr = (Sesh *) clientData;
	char *myarg[2];
	char neworig[100];
	myarg[0]= "-origx";
	myarg[1]= neworig;

	if (argc < 2)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " xview", NULL);
		return TCL_ERROR;
	}
	if (argc == 2)
	{
		char avant[100];
		char apres[100];
		double av, ap;
		SESH_largeur(SeshPtr, &av, &ap);
		sprintf(avant, "%f", av);
		sprintf(apres, "%f", ap);
		Tcl_AppendResult(interp, avant ," ", apres, NULL);
	}
	else if (argc == 4 && strcmp(argv[2], "moveto") == 0)
	{
		double frac;
		double maxvu;
		double largeur;
		double maxx;

		sscanf(argv[3], "%lf", &frac);
		largeur= TKSO_GET_REAL_WIDTH(SeshPtr->affichage);
		maxx=  SESH_max_x_dim(SeshPtr) * (SeshPtr->echelle);
		maxvu= maxx * frac;
		if (maxvu > 0.0)
			sprintf(neworig, "%f", maxvu);
		else
			sprintf(neworig, "%s", "0");
		SeshPtr->display_needed= TRUE;
		SeshConfigure(interp, SeshPtr, 2, myarg, TK_CONFIG_ARGV_ONLY);
	}
	else if (argc == 5 && strcmp(argv[2], "scroll") == 0)
	{
		double amount;
		double origine, largeur;

		sscanf(argv[3], "%lf", &amount);
		origine= SeshPtr->origx;
		largeur= TKSO_GET_REAL_WIDTH(SeshPtr->affichage);
		SeshPtr->display_needed= TRUE;
		if (strcmp(argv[4], "units") == 0)
		{

			origine += amount;
		}
		else if (strcmp(argv[4], "pages") == 0)
		{
			origine += amount * largeur;
		}
		else
			return TCL_ERROR;
		if (origine < 0.0)
			origine= 0.0;
		sprintf(neworig, "%f", origine);
		SeshConfigure(interp, SeshPtr, 2, myarg, TK_CONFIG_ARGV_ONLY);
	}
	else
	{
		return TCL_ERROR;
	}
	return TCL_OK;
}

PRIVATE int
tksyview(ClientData clientData,		/* Information about Sesh widget. */
			 Tcl_Interp *interp,			/* Current interpreter. */
			 int argc,								/* Number of arguments. */
			 char **argv							/* Argument strings. */
			 )
{
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 2)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " see", NULL);
		return TCL_ERROR;
	}
	if (argc == 2)
	{
		double maxy;
		char avant[100];
		char apres[100];
		maxy= SeshPtr->max_lignes; /* SESH_line_number(SeshPtr, NULL);*/
		if (maxy > 0.0){
			GeneralIndex debut_ecran;
			double debut;

			debut_ecran= SESH_get_marque(SeshPtr, "ecran");
			debut= SESH_line_number(SeshPtr, debut_ecran);
			sprintf(avant, "%f", debut / maxy);
			sprintf(apres, "%f", (debut + SeshPtr->nombre_de_lignes_affichees)* 1.0/maxy);
			Tcl_AppendResult(interp, avant ," ", apres, NULL);
			return TCL_OK;
		}
		else
		{
			Tcl_AppendResult(interp, "0 1", NULL);
			return TCL_OK;
		}
	}
	else if (argc == 4 && strcmp(argv[2], "moveto") == 0)
	{
		double frac;
		double maxvu;
		sscanf(argv[3], "%lf", &frac);
		maxvu= SeshPtr->max_lignes * frac;
		if (maxvu > 0.0)
		{
			SESH_set_marque(SeshPtr, "ecran",
											SESH_moveby_lines(SeshPtr,
																			SESH_make_general_index_to(SeshPtr->debut_texte,0),
																			maxvu));
		}
		else
		{
			SESH_set_marque(SeshPtr, "ecran",
											SESH_make_general_index_to(SeshPtr->debut_texte,0));
		}
		SeshPtr->display_needed= TRUE;
		return TCL_OK;
	}
	else if (argc == 5 && strcmp(argv[2], "scroll") == 0)
	{
		double amount;
		GeneralIndex debut_ecran;

		debut_ecran= SESH_get_marque(SeshPtr, "ecran");
		sscanf(argv[3], "%lf", &amount);

		SeshPtr->display_needed= TRUE;

		if (strcmp(argv[4], "pages") == 0)
		{
			amount= amount* SeshPtr->nombre_de_lignes_affichees;
			if (1.0 > amount && amount > 0.0)
				amount= 1;
			if (-1.0 < amount && amount < 0.0)
				amount= -1;

		}

		SESH_set_marque(SeshPtr, "ecran",
										SESH_moveby_lines(SeshPtr, debut_ecran, amount));
		return TCL_OK;

	}
	else
	{
		return TCL_ERROR;
	}
}

PRIVATE int
tkssetfirstline(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about Sesh widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc == 3)
	{
		int numero_premiere_ligne= (int)atoi(argv[2]);
		
		SESH_set_marque(SeshPtr, "ecran",
										SESH_moveby_lines(SeshPtr,
																		SESH_make_general_index_to(SeshPtr->debut_texte,0),
																		numero_premiere_ligne));
		SeshPtr->display_needed= TRUE;
		return  TCL_OK;
	}
	else
		return TCL_ERROR;
}

PRIVATE int
tksgetfirstline(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about Sesh widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
	char s[10];
	Sesh *SeshPtr = (Sesh *) clientData;
	sprintf(s,"%d", SESH_line_number(SeshPtr, SESH_get_marque(SeshPtr, "ecran")));
	Tcl_AppendResult(interp,s, NULL);
	return TCL_OK;
}

PRIVATE int
tksdimx(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about Sesh widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
	char s[10];
	Sesh *SeshPtr = (Sesh *) clientData;
	sprintf(s,"%d", SESH_max_x_dim(SeshPtr));
	Tcl_AppendResult(interp,s, NULL);
	return TCL_OK;
}

PRIVATE int
tksdimy(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about Sesh widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
	char s[10];
	Sesh *SeshPtr = (Sesh *) clientData;
	sprintf(s,"%d", SESH_max_y_dim(SeshPtr));
	Tcl_AppendResult(interp,s, NULL);
	return TCL_OK;
}

/*
	 coords :
	 .s coords type <COORDS> NUM
	 .s coords type <COORDS> <VAL1> <VAL2> ...
	 .s coords format {ligne $1 page $2 volume $3}
	 .s coords get <INDEX>
*/

PRIVATE int
tkscoords(ClientData clientData, /* Information about Sesh widget. */
					Tcl_Interp *interp,		/* Current interpreter. */
					int argc,							/* Number of arguments. */
					char **argv						/* Argument strings. */
					)
{
	Sesh *SeshPtr = (Sesh *) clientData;

	if (argc < 3)
	{
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " coords", NULL);
		return TCL_ERROR;
	}
	switch (*argv[2])
	{
	case 't':											/* type */
		if (argc < 5)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " coords", NULL);
			return TCL_ERROR;
		}
		set_type_coordinate(&(SeshPtr->coordinates), atoi(argv[3]), argc - 4, argv + 4);
		break;
	case 'f':											/* format */
		if (argc < 4)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " coords", NULL);
			return TCL_ERROR;
		}
		set_format_coordinate(&(SeshPtr->coordinates), argv[3]);
		break;
	case 'g':											/* get */
	{
		char *res;
		GeneralIndex t;
		int pos= 3;
		if (argc < 4)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " coords", NULL);
			return TCL_ERROR;
		}
		t= SESH_index(SeshPtr, argc, argv, &pos);
		res= SESH_trouve_index_texte(SeshPtr, t);
		Tcl_AppendResult(interp, res, NULL);
		LIBERE(res);
	}
		break;
	default:
		Tcl_AppendResult(interp, "unknown ", argv[0], " coords", argv[2], NULL);
		return TCL_ERROR;
	}
	return TCL_OK;
}

/*
PRIVATE int
tkstogglegram(
    ClientData clientData, 
    Tcl_Interp *interp,			
    int argc,		
    char **argv	
		)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			SESH_toggle_sign_is_grammar(debut);
			SeshPtr->display_needed= TRUE;
		}
	return TCL_OK;
}
*/

/*
PRIVATE int
tkstoggleword(
    ClientData clientData,	
    Tcl_Interp *interp,			
    int argc,				
    char **argv			
		)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			SESH_toggle_sign_is_wordend(debut);
			SeshPtr->display_needed= TRUE;
		}
	return TCL_OK;
}
*/

/**
 *
 * tkssign --
 * sign command :
 * allows one to change some characteristics for a particular sign
 * 
 * .s sign ?-word (O|1|c) ?-gram (0|1|c) ?-rot angle 
 *			   ?-revert (0|1|c) ?-size percent ?-select (O|1|c) 
 *         ?-ignore (0|1|c)
 *         ?-tall (0|1|c|s)
 *         ?-long (0|1|c|s)
 *         coords...
 *
 *
 * -word : word ending
 * -gram : grammatical ending
 * -rot : rotation angle (from 0 to 270, 90 steps only)
 * -revert : is sign reverted ?
 * -size : base size of the sign, as a percentage of its "natural" size
 * -select : is sign selected 
 * -tall : should the sign be streched vertically if needed. possible values : 0, 1, s. If 's' is
 *         chosen, then the choice is up to the system.
 * -long : should the sign be streched horizontal if needed ?
 * -ignore : should the sign be ignored ?
 *
 * coords... : index of the sign.
 *
 * 0 means false, 1 means true, and 'c' means change (toggle) between true and false
 *
 */
PRIVATE int
tkssign(
				ClientData clientData,		/* Information about Sesh widget. */
				Tcl_Interp *interp,			/* Current interpreter. */
				int argc,				/* Number of arguments. */
				char **argv			/* Argument strings. */
				)
{
	int pos;
	int errpos;										/* if there is an error, index of error's position */
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	int grammod=0;								/* do we modify the grammar status ? */
	char gramval;								/* if so, value of the modification */
	int rotmod= 0;								/* rotation */
	int rotval= 0;
	int revmod= 0;								/* reversal */
	char revval= 0;
	int sizemod= 0;								/* size modification */
	int sizeval= 0;
	int selectmod= 0;							/* selection modification */
	char selectval= 0;
	int wordmod=0;								/* same for word */
	char wordval=0;
	int ignoremod= 0;
	int ignoreval= 0;
	int tallmod= 0;
	int tallval= 0;
	int longmod= 0;
	int longval= 0;

	enum {E_no_error, E_number_of_args, E_bad_args, E_bad_angle, E_bad_size, E_bad_reference} myerror= E_no_error;
	
	/* first, let's gather the options */
	pos= 2;
	while (pos < argc && argv[pos][0] == '-' && myerror == E_no_error)
		{
			int lastarg= -1;

			errpos= pos;			
			if (pos + 1 == argc)
				{
					myerror= E_number_of_args;
					break;
				}

			switch (argv[pos][1])
				{
				case 'g':								/* grammar */
					grammod= TRUE;
					lastarg= gramval= argv[pos+1][0];
					break;
				case 'i':								/* ignore */
					ignoremod= TRUE;
					lastarg= ignoreval= argv[pos+1][0];
					break;
				case 'l':								/* long */
					longmod= TRUE;
					lastarg= longval= argv[pos+1][0];
					if (longval != '0' && longval != '1' && longval != 'c' && longval != 'i')
						myerror= E_bad_args;
					break;
				case 'r':
					switch (argv[pos][2]) {
					case 'e':							/* revert */
						revmod= TRUE;
						lastarg= revval= argv[pos+1][0];
						break;
					case 'o':							/* rotation */
						rotmod= TRUE;
						rotval= atoi(argv[pos+1]) % 360;
						if (rotval % 90 != 0)
							myerror= E_bad_angle;
						break;
					}
					break;
				case 's':
					switch (argv[pos][2]) {
					case 'e':							/* select */
						selectmod= TRUE;
						lastarg= selectval= argv[pos+1][0];
						break;
					case 'i':							/* size */
						sizemod= TRUE;
						sizeval= atoi(argv[pos+1]);
						if (sizeval <= 0)
							myerror= E_bad_size;
						break;
					}
					break;
				case 't':								/* tall */
					tallmod= TRUE;
					tallval= argv[pos+1][0];
					if (tallval != '0' && tallval != '1' && tallval != 'c' && tallval != 'i')
						myerror= E_bad_args;
					break;
				case 'w':								/* word */
					wordmod= TRUE;
					lastarg= wordval= argv[pos+1][0];
					break;
				default:
					myerror= E_bad_args;
				}
			
			if (lastarg != -1 && lastarg != '0' && lastarg != '1' && lastarg != 'c')
				myerror= E_bad_args;

			pos+= 2;
		}
	
	if (pos >= argc)
		{
			myerror= E_number_of_args;
		}
	else 
		{
			errpos= pos;
			if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
				{
					/* Here starts the code which actually does something */
					if (grammod)
							SESH_set_sign_property(debut,Pgram, gramval);
					if (revmod)
							SESH_set_sign_property(debut,Prevert, revval);
					if (rotmod)
							SESH_set_sign_property(debut,Protation, rotval);
					if (selectmod)
							SESH_set_sign_property(debut,Pselect, selectval);
					if (sizemod)
							SESH_set_sign_property(debut,Psize, sizeval);
					if (wordmod)
							SESH_set_sign_property(debut,Pword, wordval);
					if (ignoremod)
						SESH_set_sign_property(debut,Pignore, ignoreval);
					if (tallmod)
						SESH_set_sign_property(debut,Ptall, tallval);
					if (longmod)
						SESH_set_sign_property(debut,Plong, longval);
					SeshPtr->display_needed= TRUE;
				}
			else {
				myerror= E_bad_reference;
			}
		}
	
	switch (myerror) {
	case E_no_error:
		return TCL_OK;
	case E_number_of_args:
		Tcl_AppendResult(interp, "wrong # args: ", argv[0], " sign", NULL);
		return TCL_ERROR;
	case E_bad_reference:
		Tcl_AppendResult(interp, "bad reference in ", argv[0], " sign :", argv[errpos], NULL);
		return TCL_ERROR;
	case E_bad_args:
		Tcl_AppendResult(interp, "wrong arg in ", argv[0], " sign", argv[errpos], argv[errpos+1], NULL);
		return TCL_ERROR;
	case E_bad_size:
		Tcl_AppendResult(interp, "wrong size in ", argv[0], " sign", argv[errpos], argv[errpos+1], NULL);
		return TCL_ERROR;
	case E_bad_angle:
		Tcl_AppendResult(interp, "wrong angle in ", argv[0], " sign", argv[errpos], argv[errpos+1], NULL);
		return TCL_ERROR;
	}
	return TCL_OK;
}

PRIVATE int
tksnbrsignes(
    ClientData clientData,		/* Information about Sesh widget. */
    Tcl_Interp *interp,			/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv			/* Argument strings. */
		)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			int d;
			static char nbr[6];
			d= SESH_nombres_de_signes_dans_Texte(GETTEXTOF(debut));
			sprintf(nbr, "%d", d);
			Tcl_AppendResult(interp, nbr, NULL);
		}
	return TCL_OK;
}

PRIVATE int
tkslistedessignes(
    ClientData clientData,		/* Information about Sesh widget. */
    Tcl_Interp *interp,			/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv			/* Argument strings. */
		)
{
	int pos;
	int opts;
	GeneralIndex debut, fin;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 4)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " liste_signes ?OPTS DEBUT FIN", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	opts= SIGNS_ONLY;
	if (strcmp(argv[pos], "-word") == 0) {
		pos++;
		opts |= SIGNS_EOW;
	}
	if (strcmp(argv[pos], "-gram") == 0) {
		pos++;
		opts |= SIGNS_GRAM;
	}
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
		{
			char *s;
			s= SESH_listesignes(debut, fin,opts);
			Tcl_AppendResult(interp, s, NULL);
			LIBERE(s);
		}
	return TCL_OK;
}

typedef enum {
	bad, gt, ge, lt, le, eq, ne
} compareop;

compareop
tksgetcompareop(int argc, char **argv, int*pos)
{
	compareop res;
	if (*pos >= argc)
		return bad;

	res= bad;
	if (strcmp(argv[*pos], "==") == 0)
		res= eq;
	else if (strcmp(argv[*pos], "!=") == 0)
		res= ne;
	else if (strcmp(argv[*pos], ">") == 0)
		res= gt;
	else if (strcmp(argv[*pos], "<") == 0)
		res= lt;
	else if (strcmp(argv[*pos], "<=") == 0)
		res= le;
	else if (strcmp(argv[*pos], ">=") == 0)
		res= ge;
	(*pos)++;
	return res;
}

PRIVATE int
tkscompare(
    ClientData clientData,		/* Information about Sesh widget. */
    Tcl_Interp *interp,			/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv			/* Argument strings. */
		)
{
	int pos;
	GeneralIndex debut, suite;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 5)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " compare", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			compareop op;
			int oldpos= pos;

			op= tksgetcompareop(argc, argv, &pos);
			
			if (op != bad)
				{
					if (Getdebut(SeshPtr, &suite, argc, argv, &pos))
						{
							int comp;
							comp= SESH_compare_indexes(SeshPtr, debut, suite);
							switch (op) {
							case bad:					/* impossible */
								break;
							case eq:
								comp= (comp == 0);
								break;
							case ne:
								comp= (comp != 0);
								break;
							case gt:
								comp= (comp > 0);
								break;
							case lt:
								comp= (comp < 0);
								break;
							case ge:
								comp= (comp >= 0);
								break;
							case le:
								comp= (comp <= 0);
								break;
							}
							if (comp == 0)
								Tcl_AppendResult(interp, "0", NULL);
							else 
								Tcl_AppendResult(interp, "1", NULL);
						}
					else
						{
							Tcl_AppendResult(interp, "wrong args: ", argv[0], " compare", argv[1], NULL);
							return TCL_ERROR;
						}
				}
			else 
				{
					Tcl_AppendResult(interp, "wrong operator in ", argv[0], " compare :", argv[oldpos], NULL);
					return TCL_ERROR;
				}
		}
	else {
		Tcl_AppendResult(interp, "wrong args: ", argv[0], " compare", argv[1], NULL);
		return TCL_ERROR;
	}
	return TCL_OK;
}


PRIVATE int
tksisgrammar(
    ClientData clientData,		/* Information about Sesh widget. */
    Tcl_Interp *interp,			/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv			/* Argument strings. */
		)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			if (SESH_sign_get_p(debut, Pgram))
				Tcl_AppendResult(interp, "1", NULL);
			else
				Tcl_AppendResult(interp, "0", NULL);
		}
	return TCL_OK;
}


PRIVATE int
tksisend(
    ClientData clientData,		/* Information about Sesh widget. */
    Tcl_Interp *interp,			/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv			/* Argument strings. */
		)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			if (SESH_sign_get_p(debut, Pword))
				Tcl_AppendResult(interp, "1", NULL);
			else
				Tcl_AppendResult(interp, "0", NULL);
		}
	return TCL_OK;
}

PRIVATE int
tksisvertical(
							ClientData clientData,		/* Information about Sesh widget. */
							Tcl_Interp *interp,			/* Current interpreter. */
							int argc,				/* Number of arguments. */
							char **argv			/* Argument strings. */
							)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			if (SESH_is_vertical(debut))
				Tcl_AppendResult(interp, "1", NULL);
			else
				Tcl_AppendResult(interp, "0", NULL);
		}
	return TCL_OK;
}


PRIVATE int
tksconfigure(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about Sesh widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
	Sesh *SeshPtr = (Sesh *) clientData;
	int result;
	if (argc == 2) {
		result = Tk_ConfigureInfo(interp, SeshPtr->tkwin, configSpecs,
															(char *) SeshPtr, (char *) NULL, 0);
	} else if (argc == 3) {
		result = Tk_ConfigureInfo(interp, SeshPtr->tkwin, configSpecs,
															(char *) SeshPtr, argv[2], 0);
	} else {
		result = SeshConfigure(interp, SeshPtr, argc-2, argv+2,
													 TK_CONFIG_ARGV_ONLY);
	}
	SeshPtr->display_needed= TRUE;
	/* Invalidate all visible text : */
	SESH_increment_compteur(SeshPtr);
	SESH_increment_compteur(SeshPtr);
	return  result;
}

PRIVATE int
tkscget(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about Sesh widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
	Sesh *SeshPtr = (Sesh *) clientData;

	if (argc != 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
											argv[0], " cget option\"",
											(char *) NULL);
		return TCL_ERROR;
	}
	else
		return Tk_ConfigureValue(interp, SeshPtr->tkwin, configSpecs,
															 (char *) SeshPtr, argv[2], 0);
}


PRIVATE int
tksselect(clientData, interp, argc, argv)
		 ClientData clientData;		/* Information about Sesh widget. */
		 Tcl_Interp *interp;			/* Current interpreter. */
		 int argc;				/* Number of arguments. */
		 char **argv;			/* Argument strings. */
{
	int pos;
	GeneralIndex debut, fin;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " undocumented0 index1 ?index2", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
		{
			SESH_set_selected(debut, fin, TRUE);
			SeshPtr->display_needed= TRUE;
		}
	else
		{
			Tcl_AppendResult(interp, "wrong index ", NULL);
			return TCL_ERROR;
		}
	return TCL_OK;
}

PRIVATE int
tksunselect(clientData, interp, argc, argv)
		 ClientData clientData;		/* Information about Sesh widget. */
		 Tcl_Interp *interp;			/* Current interpreter. */
		 int argc;				/* Number of arguments. */
		 char **argv;			/* Argument strings. */
{
	int pos;
	GeneralIndex debut, fin;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " undocumented1 index1 ?index2", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebutfin(SeshPtr, &debut, &fin, argc, argv, &pos))
		{
			SESH_set_selected(debut, fin, FALSE);
			SeshPtr->display_needed= TRUE;
		}
	else
		{
			Tcl_AppendResult(interp, "wrong index ", NULL);
			return TCL_ERROR;
		}
	return TCL_OK;
}


/*
PRIVATE int
tksselectsign(
    ClientData clientData,		
    Tcl_Interp *interp,			
    int argc,				
    char **argv			
		)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 4)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " undocumented2 INDEX VALUE", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			int value;
			if (pos < argc)
				{
					value= (int)atoi(argv[pos]);
					SESH_set_sign_selected(debut, value);
					SeshPtr->display_needed= TRUE;
				}
			else
				{
					Tcl_AppendResult(interp, "need a value in ", argv[0], " undocumented2 INDEX VALUE", NULL);
					return TCL_ERROR;
				}
		}
	return TCL_OK;
}
*/


PRIVATE int
tksgetselectsign(
    ClientData clientData,		/* Information about Sesh widget. */
    Tcl_Interp *interp,			/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv			/* Argument strings. */
		)
{
	int pos;
	GeneralIndex debut;
	Sesh *SeshPtr = (Sesh *) clientData;
	if (argc < 3)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], " delete", NULL);
			return TCL_ERROR;
		}
	pos= 2;
	if (Getdebut(SeshPtr, &debut, argc, argv, &pos))
		{
			if (SESH_sign_get_p(debut, Pselect))
				Tcl_AppendResult(interp, "1", NULL);
			else
				Tcl_AppendResult(interp, "0", NULL);
		}
	return TCL_OK;
}


/*
	Auxiliary function used by tksimg.
	returns a dump of the window in 
	xpm format. 

	New version : use the TKSO_bitmap library.
*/

PRIVATE void
TKSESH_output_dump_pbm(Tcl_Interp *interp, TKSO_tkwin *o)
{

	Tcl_Obj *result;

	// result= TKSO_bitmap_getPBM(out);
	Tcl_SetObjResult(interp, result);	
}

PRIVATE void
TKSESH_output_dump_xbm(Tcl_Interp *interp, TKSO_tkwin *o)
{
	unsigned int x, y;
	char width[100], height[100];
	XImage *imagePtr;
	Tcl_Obj *result;

	/* black and white code : */
	
	result= Tcl_NewStringObj(0,0);
	/* sprintf(s, "%d %d\n", 8 * ((o->w + 7) / 8), o->h);*/
	sprintf(width, "%d", TKSOC(o)->w);
	sprintf(height, "%d", TKSOC(o)->h);
	
	Tcl_AppendStringsToObj(result,
												 "#define Toto_width ", width, "\n",
												 "#define Toto_height ", height, "\n",
												 "static char Toto_bits[] = {", NULL);
	
	imagePtr = TKSO_tkwin_getImage(o);
	
	if (1) {
		unsigned char c=255;
		char s[10];
		int pos= 7;
		int col=0;
		int notfirst= FALSE;
		for (y=0; y < TKSOC(o)->h; y++)
			{
				for (x=0; x < TKSOC(o)->w; x++)
					{
						unsigned long pix;
						pix= XGetPixel(imagePtr, x, y);
						if (pix != 0L) 
							c &= ~(1 << (7 - pos));
						pos--;
						if (pos == -1)
							{
								pos= 7;
								if (notfirst)
									Tcl_AppendStringsToObj(result,", ", NULL);
								if (col++ % 10 == 0)
									Tcl_AppendStringsToObj(result,"\n", NULL);
								sprintf(s, "0x%02x", c);
								Tcl_AppendStringsToObj(result, s, NULL);

								notfirst= TRUE;
								c=255;
							}
					}
				if (pos != 7)
					{
						pos= 7;
						if (notfirst)
							Tcl_AppendStringsToObj(result,", ", NULL);
						if (col++ % 10 == 0)
							Tcl_AppendStringsToObj(result,"\n", NULL);
						sprintf(s, "0x%02x", c);
						Tcl_AppendStringsToObj(result, s, NULL);
						c=255;
					}
			}
		Tcl_AppendStringsToObj(result, "}\n", NULL);
	}
	
	XDestroyImage(imagePtr);
	Tcl_SetObjResult(interp, result);	
}

PRIVATE int
tksimg(
    ClientData clientData,		/* Information about Sesh widget. */
    Tcl_Interp *interp,			/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv			/* Argument strings. */
		)
{
	Sesh *SeshPtr = (Sesh *) clientData;
	int as_xbm= FALSE;

	if (argc > 3 || argc < 2)
		{
			Tcl_AppendResult(interp, "wrong # args: ", argv[0], NULL);
			return TCL_ERROR;
		}
	if (argc== 3 && strcmp(argv[2], "-xbm") == 0)
		{
			as_xbm= TRUE;
		}
	if (!Tk_IsMapped(SeshPtr->tkwin))
		{
			Tk_MakeWindowExist(SeshPtr->tkwin);
			TKSESH_redrawWidget(SeshPtr);
		}
	if (as_xbm)
		TKSESH_output_dump_xbm(interp, SeshPtr->old_affichage);
	else
		TKSESH_output_dump_pbm(interp, SeshPtr->old_affichage);
	return TCL_OK;
}



PRIVATE int
tkshelp();

PRIVATE struct {
	char* nomcom;
	int (*f)();
	char * help;
} tkcommandes[]=
{
{"mark", tksmark, "(names|set NAME INDEX|unset)"},
{"delete", tksdelete, "INDEX1 ?INDEX2?"},
{"get", tksget, "INDEX1 ?INDEX2?"},
{"insert", tksinsert, "INDEX TEXT"},
{"index", tksindex, "INDEX"},
{"see", tksshow, "INDEX"},
{"xypos", tksxypos, "INDEX"},
{"xview", tksxview, ""},
{"yview", tksyview, ""},
{"explode", tksexplode, ""},	/* A RENOMMER  */
{"listecadrats", tkslistcad, "INDEX1 INDEX2: return-separated list of quadras"},	/* A RENOMMER  */
{"coords", tkscoords, ":manipulate logical coordinates."},
{"help", tkshelp, ""},
{"setfirstline", tkssetfirstline, "LINE_NUMBER"},
{"getfirstline", tksgetfirstline, ""},
{"dimx", tksdimx, ""},
{"dimy", tksdimy, ""},
/*{"toggle_word", tkstoggleword, ""},*/
/*{"toggle_gramm", tkstogglegram, ""},*/
{"nbr_signes", tksnbrsignes, "INDEX : number of signs in a cadrat."},
{"liste_signes", tkslistedessignes, "INDEX: list of signs in a cadrat."},
{"isgram", tksisgrammar, "INDEX"},
{"iswordend", tksisend, "INDEX"},
{"isvertical", tksisvertical, "INDEX"},
{"compare", tkscompare, "INDEX op INDEX, compare two indexes. op : <, >, ==, !=, <=, or >="},
	/* {"redisplay", tksisvertical, "-- redraws a widget"},*/
{"undocumented0", tksselect, "INDEX1 INDEX2"},
{"undocumented1", tksunselect, "INDEX1 INDEX2"},
/*{"undocumented2", tksselectsign, "INDEX VALUE (0|1): set sign selection"},*/
{"undocumented3", tksgetselectsign, "INDEX : get sign selection value"},
	/*{"undocumented4", tksimg, "get a ppm representation of the glyphs."},*/
{"towmf", tks2wmf, "saves a wmf representation of the glyphs."},
#if defined(__WIN32__)
{"towinclipboard", tks2wcb, "saves a wmf representation of the glyphs in the window clipboard."},	
#endif
{"topbm", tks2bitmap, "get a pbm representation of the glyphs."},
{"toxbm", tks2bitmap, "get a xbm representation of the glyphs."},
{"sign", tkssign, "allows various manipulations on a particular sign in a cadrat"},
/* Generic functions. Is there any way of NOT re-writting them ? */
{"configure", tksconfigure, ""},
{"cget", tkscget, ""},
{0,0, ""}
};


PRIVATE int
tkshelp(ClientData clientData,	/* Information about Sesh widget. */
				Tcl_Interp *interp,			/* Current interpreter. */
				int argc,								/* Number of arguments. */
				char **argv							/* Argument strings. */
				)
{
	int i;
	for (i= 0; tkcommandes[i].nomcom; i++)
		Tcl_AppendResult(interp,
										 tkcommandes[i].nomcom, " ",
										 tkcommandes[i].help,
										 "\n", NULL);
	return  TCL_OK;
}



PRIVATE int
SeshWidgetCmd(ClientData clientData,		/* Information about Sesh widget. */
							Tcl_Interp *interp,			/* Current interpreter. */
							int argc,				/* Number of arguments. */
							char **argv			/* Argument strings. */
							)
{
	Sesh *SeshPtr = (Sesh *) clientData;
	int result = TCL_OK;
	int trouve= FALSE;
	int i;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
										 argv[0], " option ?arg arg ...?\"", (char *) NULL);
		return TCL_ERROR;
	}

	Tk_Preserve((ClientData) SeshPtr);

	for (i=0, trouve= FALSE; tkcommandes[i].nomcom && (!trouve); i++)
		trouve= (strcmp(tkcommandes[i].nomcom, argv[1]) == 0);

	if (trouve)
		result= (tkcommandes[i-1].f)(clientData, interp, argc, argv);
	else {
		Tcl_AppendResult(interp, "wrong command : ", argv[1], (char *) NULL);
		result= TCL_ERROR;
	}

	if (result == TCL_OK && !SeshPtr->updatePending) {
		Tk_DoWhenIdle(SeshDisplay, (ClientData) SeshPtr);
		SeshPtr->updatePending = 1;
	}

	Tk_Release((ClientData) SeshPtr);
	return result;
}

/*
 *----------------------------------------------------------------------
 *
 * SeshConfigure --
 *
 *	This procedure is called to process an argv/argc list in
 *	conjunction with the Tk option database to configure (or
 *	reconfigure) a Sesh widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for SeshPtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

PRIVATE void
Sesh_set_size(Sesh *SeshPtr, int w, int h)
{
	TKSO_SET_WINDOW(SeshPtr->affichage,
									TKSO_init_rectangle((SeshPtr->origx),
																			(0.0 - h* (SeshPtr->echelle)),
																			((SeshPtr->origx) +
																			 (w* (SeshPtr->echelle))),
																			0.0),
									w, h);
	/*
	 * If we are drawing in columns, nombre_de_lignes_affichees
	 * is in fact the number of columns !!!!!
	 */
	if (TKSGeom_isInColumnsP(&SeshPtr->geom))
		SeshPtr->nombre_de_lignes_affichees=
			(w + SeshPtr->linesep +1.0e-6) / (SeshPtr->htline*COLUMNRATIO);
		else
			SeshPtr->nombre_de_lignes_affichees=
				(h + SeshPtr->linesep +1.0e-6) / (SeshPtr->htline);
	/* (h % (SeshPtr->htline+ SeshPtr->linesep)) / SeshPtr->htline ;*/
	if (w != Tk_Width(SeshPtr->tkwin) || h != Tk_Height(SeshPtr->tkwin))
		Tk_GeometryRequest(SeshPtr->tkwin, w, h);
	Tk_SetInternalBorder(SeshPtr->tkwin, 0);
	/* a change could have been made */
	SESH_XUpdate(SeshPtr);
	SESH_YUpdate(SeshPtr);
	SeshPtr->display_needed= TRUE;
}

PRIVATE int
SeshConfigure(interp, SeshPtr, argc, argv, flags)
    Tcl_Interp *interp;			/* Used for error reporting. */
    Sesh *SeshPtr;			/* Information about widget. */
    int argc;				/* Number of valid entries in argv. */
    char **argv;			/* Arguments. */
    int flags;				/* Flags to pass to
					 * Tk_ConfigureWidget. */
{

	if (Tk_ConfigureWidget(interp, SeshPtr->tkwin, configSpecs,
												 argc, argv, (char *) SeshPtr, flags) != TCL_OK) {
		return TCL_ERROR;
	}

	/* the values of htline, linesep, origx, and (origy ?) have been set.
	 * Let's calculate the rest of 'hem
	 */

	if (! TKSO_tkwin_hasPalette(SeshPtr->affichage))
		{
			XColor *colors[TKSESHNBCOLORS];
			colors[0]= Tk_3DBorderColor(SeshPtr->bgBorder);
			colors[1]= SeshPtr->foreground;
			colors[2]= SeshPtr->gramcolor;
			colors[3]= SeshPtr->cursorcolor;
			colors[4]= SeshPtr->gramendcolor;
			colors[5]= SeshPtr->cursorcolor;
			colors[6]= SeshPtr->redcolor;
			colors[7]= SeshPtr->graycolor;
			SeshPtr->palette= TKSO_tkwin_new_palette(TKSESHNBCOLORS, colors);
			TKSO_tkwin_set_palette(SeshPtr->affichage, SeshPtr->palette);
			TKSO_tkwin_set_palette(SeshPtr->old_affichage, SeshPtr->palette);
			// A REMPLACER PAR QUELQUE CHOSE DE PROPRE !!!!
			// SeshPtr->old_affichage.p= SeshPtr->affichage.p;
			// The palette should in fact be kept in some external part.
		}
	
	TKSO_SET_SCALE(SeshPtr->affichage, (SeshPtr->htline));
	SeshPtr->echelle= 1.0/ (SeshPtr->htline);
	SeshPtr->virtuallinesep= SeshPtr->linesep* SeshPtr->echelle;
	Sesh_set_size(SeshPtr, SeshPtr->largeur, SeshPtr->hauteur);
	SeshPtr->resolution= 1.0* (SeshPtr->htline - SeshPtr->linesep)/ (SeshPtr->htline);

	/* Triple buffering on ??? */
	if (SeshPtr->keep_old && SeshPtr->past_affichage == 0)
		SeshPtr->past_affichage= TKSO_tkwin_init(SeshPtr->display,
																						 0.0, -3.0, 20.0, 0.0, 
																						 300, 30, 
																						 SeshPtr->tkwin,
																						 SeshPtr->interp);
	if (! SeshPtr->keep_old && SeshPtr->past_affichage != 0)
		{
			TKSO_tkwin_free_draw(SeshPtr->past_affichage);
			TKSO_tkwin_free(SeshPtr->past_affichage);
			SeshPtr->past_affichage= 0;
		}
	/*
	 * Set the background for the window and create a graphics context
	 * for use during redisplay.
	 */

	Tk_SetWindowBackground(SeshPtr->tkwin,
												 Tk_3DBorderColor(SeshPtr->bgBorder)->pixel);

	/*
	 * Register the desired geometry for the window.  Then arrange for
	 * the window to be redisplayed.
	 TKSESH_output_set_drawable(&(SeshPtr->affichage),
	 Tk_WindowId(SeshPtr->tkwin));
	 */
	
	Tk_SetInternalBorder(SeshPtr->tkwin, SeshPtr->borderWidth);
	if (!SeshPtr->updatePending) {
		Tk_DoWhenIdle(SeshDisplay, (ClientData) SeshPtr);
		SeshPtr->updatePending = 1;
	}
	/* a change could have been made */
	SESH_XUpdate(SeshPtr);
	SESH_YUpdate(SeshPtr);
	SeshPtr->display_needed= TRUE;
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SeshEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on Seshs.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

/* this is an auxiliary procedure in case something has to be done */
PRIVATE void
traite_reconfigure(ClientData clientData);

PRIVATE void
SeshEventProc(clientData, eventPtr)
	ClientData clientData;	/* Information about window. */
	XEvent *eventPtr;		/* Information about event. */
{
	Sesh *SeshPtr = (Sesh *) clientData;

	if (eventPtr->type == Expose) {
		SeshPtr->display_needed= TRUE;
		if (!SeshPtr->updatePending) {
			Tk_DoWhenIdle(SeshDisplay, (ClientData) SeshPtr);
	    SeshPtr->updatePending = 1;
		}
	} else if (eventPtr->type == ConfigureNotify) {
		traite_reconfigure(clientData);
		SeshPtr->display_needed= TRUE;
		if (!SeshPtr->updatePending) {
	    Tk_DoWhenIdle(SeshDisplay, (ClientData) SeshPtr);
	    SeshPtr->updatePending = 1;
		}
	} else if (eventPtr->type == DestroyNotify) {
		if (SeshPtr->tkwin != NULL) {
	    SeshPtr->tkwin = NULL;
	    Tcl_DeleteCommand(SeshPtr->interp,
												Tcl_GetCommandName(SeshPtr->interp,
																					 SeshPtr->widgetCmd));
		}
		if (SeshPtr->updatePending) {
	    Tk_CancelIdleCall(SeshDisplay, (ClientData) SeshPtr);
		}
		Tk_EventuallyFree((ClientData) SeshPtr, SeshDestroy);
	}
}


void
traite_reconfigure(ClientData clientData)
{
	Sesh *SeshPtr = (Sesh *) clientData;

	if (Tk_Width(SeshPtr->tkwin) <= 1 ||
			Tk_Height(SeshPtr->tkwin) <= 1)
		return;

	if ((SeshPtr->largeur  != Tk_Width(SeshPtr->tkwin))
		|| (SeshPtr->hauteur != Tk_Height(SeshPtr->tkwin)))
	{
		Sesh_set_size(SeshPtr,
									Tk_Width(SeshPtr->tkwin),
									Tk_Height(SeshPtr->tkwin));
		SeshPtr->largeur= Tk_Width(SeshPtr->tkwin);
		SeshPtr->hauteur= Tk_Height(SeshPtr->tkwin);
	}
}

/*
 *----------------------------------------------------------------------
 *
 * SeshCmdDeletedProc --
 *
 *	This procedure is invoked when a widget command is deleted.  If
 *	the widget isn't already in the process of being destroyed,
 *	this command destroys it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The widget is destroyed.
 *
 *----------------------------------------------------------------------
 */

PRIVATE void
SeshCmdDeletedProc(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
	Sesh *SeshPtr = (Sesh *) clientData;
	Tk_Window tkwin = SeshPtr->tkwin;

	/*
	 * This procedure could be invoked either because the window was
	 * destroyed and the command was then deleted (in which case tkwin
	 * is NULL) or because the command was deleted, and then this procedure
	 * destroys the widget.
	 */

	if (tkwin != NULL) {
		SeshPtr->tkwin = NULL;
		Tk_DestroyWindow(tkwin);
	}
}

/*
 *--------------------------------------------------------------
 *
 * SeshDisplay --
 *
 *	This procedure redraws the contents of a Sesh window.
 *	It is invoked as a do-when-idle handler, so it only runs
 *	when there's nothing else for the application to do.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information appears on the screen.
 *
 *--------------------------------------------------------------
 */

PRIVATE void
SeshDisplay(clientData)
    ClientData clientData;	/* Information about window. */
{
	Sesh *SeshPtr = (Sesh *) clientData;
	Tk_Window tkwin = SeshPtr->tkwin;
	Texte * debut_ecran;

	SeshPtr->updatePending = 0;
	if (!Tk_IsMapped(tkwin)) {
		return;
	}

	/*
	 * Redraw the widget's background and border.
	 */
	/* SeshPtr->affichage.d= Tk_WindowId(tkwin);*/

	/*
	 * Display the Sesh.
	 */

	TKSESH_redrawWidget(SeshPtr);

	if (SeshPtr->old_maxx != SeshPtr->maxx)
	{
		SeshPtr->old_maxx= SeshPtr->maxx;
		SESH_XUpdate(SeshPtr);
	}

	debut_ecran= GETTEXTOF(SESH_get_marque(SeshPtr, "ecran"));
	if (SeshPtr->old_max_lignes != SeshPtr->max_lignes
			|| debut_ecran != SeshPtr->old_debut_ecran)
	{
		SeshPtr->old_max_lignes= SeshPtr->max_lignes;
		SeshPtr->old_debut_ecran= debut_ecran;
		SESH_YUpdate(SeshPtr);
	}
}


/*
 *----------------------------------------------------------------------
 *
 * SeshDestroy --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a Sesh at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the Sesh is freed up.
 *
 *----------------------------------------------------------------------
 */

PRIVATE void
SeshDestroy(char *blockPtr)
{
	Sesh *SeshPtr = (Sesh *) blockPtr;
	
	Tk_FreeOptions(configSpecs, (char *) SeshPtr, SeshPtr->display, 0);
	TKSO_tkwin_free_draw(SeshPtr->affichage);
	TKSO_tkwin_free_draw(SeshPtr->old_affichage);
	if (SeshPtr->past_affichage_is_usable)
		TKSO_tkwin_free_draw(SeshPtr->past_affichage);
	TKSO_tkwin_free_palette(SeshPtr->display, SeshPtr->palette);
	SESH_detruire_marques(SeshPtr);
	FreeTexte(SeshPtr->debut_texte);
	LIBERE((char *) SeshPtr);
}


/*
 * Local Variables:
 * tab-width: 2
 * End:
 */
