# Dictionary Browser. 

# Current organization :

# windows for hieroglyphs, translitteration, and translation

package provide seshDico 0.0
package require myStoopWidgets
package require seshWidget 

class DicoBrowser {
    
    proc DicoBrowser {this name} STopLevel {$name} {
	set ($this,wf) [new DicoBrowserWF $this]
    }
    
    proc ~DicoBrowser {this} {
	delete $($this,wf)
	unset (singleton)
    }

    proc getDicoBrowser {} {
	if {! [info exists (singleton)]} {
	    set (singleton) [Component::buildNewComponent DicoBrowser dicoBrowse]
	}
	return $(singleton)
    }
    
    virtual proc completorHook {this} {
	# translitteration 
	set ($this,translitteration) \
	    [Component::buildNewComponent SLabelEntry $this -auto -label "Translitteration" -font mdctranslitlc]
	# hieroglyphs
	set ($this,hierobox) \
	    [Component::buildNewComponent SLabelFrame $this -auto -text "Hieroglyphs"]
	set inner [FrameContainer::getFrame $($this,hierobox)]
	set ($this,hiero) \
	    [Component::buildNewComponent SeshW $inner -auto]
	Component::grid  $($this,hiero) -sticky news
	# translation
	set ($this,translation) \
	    [Component::buildNewComponent SLabelEntry $this -auto -label "Translation"]
	# options ?
	set ($this,optionsBox) \
	    [Component::buildNewComponent SLabelFrame $this -auto -text "options"]
	set inner [FrameContainer::getFrame $($this,optionsBox)]
	set ($this,subString) \
	    [Component::buildNewComponent SCheckButton $inner -auto -anchor n \
		 -text "Search substring" -variable DicoBrowser::($this,subSearch)]
	set ($this,and) \
	    [Component::buildNewComponent SCheckButton $inner -auto -anchor n \
		 -text "and-search" -variable DicoBrowser::($this,andSearch)]
	Component::grid [list $($this,subString) $($this,and)] -sticky w
	# clear/find/new
	set ($this,operations) \
	    [Component::buildNewComponent SFrame $this -auto]
	set box $($this,operations)
	set ($this,find) \
	    [Component::buildNewComponent SButton $box -auto -text "find"\
		 -command [list DicoBrowser::findAction $this]]
	set ($this,clear) \
	    [Component::buildNewComponent SButton $box -auto  -text "clear"\
		 -command [list DicoBrowser::clearAction $this]]
	set ($this,new) \
	    [Component::buildNewComponent SButton $box -auto  -text "new"\
		 -command [list DicoBrowser::newAction $this]]

	# button box :
	Component::grid [list $($this,find) $($this,clear) $($this,new)] -sticky news
	
	# browser
	set ($this,browser) [Component::buildNewComponent BDBBrowser $this -auto \
				 [DicoBrowserWF::getDataSource $($this,wf)]\
				 {value}]
	BDBBrowser::setFont $($this,browser) mdctranslitlc
	BDBBrowser::setCommand $($this,browser) [list DicoBrowser::loadCommand $this]
	# Layout for the whole widget :
	Component::grid $($this,translitteration) -sticky ew
	Component::grid $($this,hierobox) -sticky ew
	Component::grid $($this,translation) -sticky ew
	Component::grid $($this,operations) -sticky ew
	Component::grid $($this,optionsBox) -sticky ew
	Component::grid $($this,browser) -sticky news
	Component::rowConfigure $this 5 -weight 1
	Component::columnConfigure $this 0 -weight 1
    }

    proc reset {this} {
	setHieroglyphs $this ""
	setTranslation $this ""
	setTranslitteration $this ""
    }

    # Create a new dictionnary window using the
    # data in translitteration and translation
    proc newAction {this} {
	PRIVSetWFValues $this
	set d [buildNewDicoEntry $this]
	DicoTk::setText $d [DicoBrowserWF::buildSimpleEntry $($this,wf)]
    }

    proc clearAction {this} {
	reset $this
	PRIVSetWFValues $this
    }

    proc findAction {this} {
	PRIVSetWFValues $this
	set ids [DicoBrowserWF::find $($this,wf)]
	foreach id $ids {
	    buildNewDicoEntryForID $this $id
	}
    }
    
    proc loadCommand {this idx} {
	set id [DicoBrowserWF::getOIDForID $($this,wf) $idx]
	buildNewDicoEntryForID $this $id
    }

    # Fix the values in the workflow according to the 
    # content of the widgets
    proc PRIVSetWFValues {this} {
	set hiero [SeshW::getText $($this,hiero) 0.0 end]
	set translation [Component::cget $($this,translation) text]
	set translitteration [Component::cget $($this,translitteration) text]
	DicoBrowserWF::setHieroglyphs $($this,wf) $hiero
	DicoBrowserWF::setTranslation $($this,wf) $translation
	DicoBrowserWF::setTranslitteration $($this,wf) $translitteration
	DicoBrowserWF::setSubSearch $($this,wf) $($this,subSearch)
	DicoBrowserWF::setAndSearch $($this,wf) $($this,andSearch)
    }

    proc buildNewDicoEntry {this} {
	set topName [DicoBrowserWF::getNewTopName $($this,wf)]
	set t [STopLevel::buildNewTopLevel STopLevel $topName]
	set d [Component::buildNewComponent DicoTk $t -auto GRP]
	Component::grid $d -sticky news
	Component::columnConfigure $t 0 -weight 1
	Component::rowConfigure $t 0 -weight 1
	return $d
    }
    
    proc buildNewDicoEntryForID {this id} {
	set d [buildNewDicoEntry $this]
	DicoTk::load $d $id
    }

    proc setHieroglyphs {this txt} {
	SeshW::setText $($this,hiero) $txt
    }

    proc setTranslation {this txt} {
	Component::cset $($this,translation) text $txt
    }

    proc setTranslitteration {this txt} {
	Component::cset $($this,translitteration) text $txt
    }
}


