package provide seshDico 0.0
package require legendes
package require tkseshUtils
package require seshnormalize
# This class is a utility class for dealing with tags.
#
# it's currently not OO at all, being an adaptation from the old dico widget.
#
#
# It deals with two kinds of tags, 
# hereafter called "tags" and "ntags".
#
# "tags" are tags at the model level. A "TITLE" tag would be an example for that.
#
# "ntags" come directly from the text widget's internals. They are a combination of a "tag" 
# and of a number, which correspond to the tag's depth.
# 

class TagHandler {

    proc TagHandler {this} {
    }

    proc ~TagHandler {this} {
    }

    # Data associated to tag
    variable THE_TAS
    array set THE_TAGS {}
    variable ATOMIC_TAGS {};		# regular expression which recognise "atomic tags"
    variable LEGALTAGS 

    # getTagName returns the name part of a tag.
    proc getTagName {ntag} {
	regexp {^[^_]*} $ntag result 
	return $result
    }

    proc getTagDepth {ntag} {
	regexp {[0-9]+$} $ntag result
	return $result
    }
    
    proc buildTagBegin {name num} {
	return "begin_${name}_${num}"
    }

    proc buildTagBody {name num} {
	return "${name}_${num}"
    }

    proc buildTagEnd {name num} {
	return "end_${name}_${num}"
    }

    # Compare two tags. The order depends on the tag number, 
    # and in fact 
    proc compareTags {ntag1 ntag2} {
	set t1 [getTagDepth $ntag1]
	set t2 [getTagDepth $ntag2]
	return [expr $t1 - $t2]
    }

    
    # declare the various options for the tag.
    proc declareTag {TAG args} {
	variable THE_TAGS
	variable LEGALTAGS
	TkseshUtils::args opts {
	    -begin "\n" -middle "" -end "\n" 
	    -conf "" -bconf "" -econf "" -save "" -load ""
	    -index 0
	    -bindings {}
	    -keybind {}
	    -insert_at_end 0
	    -normalize identity
	    -description "tag non dcrit"
	    -htmlsave ""
	    -htmlbegin ""
	    -htmlend ""
	    -latexsave ""
	    -latexbegin ""
	    -latexend ""
	} $args
	set THE_TAGS($TAG,begin) $opts(-begin)
	set THE_TAGS($TAG,middle) $opts(-middle)
	set THE_TAGS($TAG,end) $opts(-end)
	set THE_TAGS($TAG,conf) $opts(-conf)
	set THE_TAGS($TAG,bconf) $opts(-bconf)
	set THE_TAGS($TAG,econf) $opts(-econf)
	set THE_TAGS($TAG,save) $opts(-save)
	set THE_TAGS($TAG,load) $opts(-load)
	set THE_TAGS($TAG,index) $opts(-index)
	set THE_TAGS($TAG,bindings) $opts(-bindings)
	set THE_TAGS($TAG,keybind) $opts(-keybind)
	set THE_TAGS($TAG,normalize) $opts(-normalize)
	set THE_TAGS($TAG,insert_at_end) $opts(-insert_at_end)
	set THE_TAGS($TAG,description) $opts(-description)
	set THE_TAGS($TAG,htmlsave) $opts(-htmlsave)
	set THE_TAGS($TAG,htmlbegin) $opts(-htmlbegin)
	set THE_TAGS($TAG,htmlend) $opts(-htmlend)
	set THE_TAGS($TAG,latexsave) $opts(-latexsave)
	set THE_TAGS($TAG,latexbegin) $opts(-latexbegin)
	set THE_TAGS($TAG,latexend) $opts(-latexend)

        foreach {b com} $THE_TAGS($TAG,keybind) {
	    set THE_TAGS($TAG,bind,$b) $com
	}
    }


    proc declareAllTags {} {
	variable LEGALTAGS
	variable tagDeclared
	if {[info exists tagDeclared]} {
	    return
	}
	set tagDeclared 1
	# Declarations of TAGS.
	# To declare a tag, one must :
	# give it a name. A TAG name 
	# should be uppercase letters only.
	# No numerals !
	#
	# a) State its grammar (below)
	# b) describe it with declareTag
	# d) insert creation menu where needed
	#    (for example :
	#	$w.menus.insert.m add command -label "usage" \
				#	    -command [list DicoTk::insert_parenthesis $w USE])
	#
	# In the case of tags which create a new window, like hieroglyphs
	# One needs to 
	# 1) write a creation function. 
	#    The function will insert the window at the insertion point
	#    Its arguments should be {w {txt ""}}
	#    The name of the window should be $w.'s'TAG$SUBWIN($w). 
	#       e.g. $w.sWRI1
	#    SUBWIN should then be augmented  
	# 2) write the save code. The save code is a function that 
	#          takes $w and the base window as arguments, and
	#          should return a text. Note : to obtain the text window
	#          from the base window, you should add a '.t'
	#
	# 3) The save function and the load function should be passed 
	#    as arguments to declareTag with the options -save and -load
	
	array set LEGALTAGS {
	    GRP {GRP TRAD WRI MTRI COM DET REF AREF USE GRA AUT EXE EMP
		VAR}
	    WRI ""
	    DET ""
	    HIE ""
	    COM {HIE MTRI TRI REF AREF GRP PCDATA USE EXE GRA COP}
	    TRAD {PCDATA MOT QQUN QQCHOSE}
	    TRI {PCDATA}
	    COP {PCDATA}
	    MTRI {PCDATA}
	    REF {PCDATA}
	    AREF {PCDATA}
	    USE {HIE PCDATA MOT QQUN QQCHOSE MTRI}
	    EXE {HIE PCDATA MOT QQUN QQCHOSE COP}
	    MOT {}
	    QQUN {}
	    QQCHOSE {}
	    GRA {}
	    LATTER {}
	    OK {}
	    NK {}
	    LE {}
	    VAR {}
	    AUT {PCDATA}
	    EMP {OK NK LE LATTER ABRV}
	    NOTE {TITLE SECTION}
	    TITLE {PCDATA HIE TRI COP}
	    SUBTITLE {PCDATA HIE TRI COP}
	    SECTION {SUBTITLE PAR SECTION}
	    PAR { PCDATA TRI HIE COP REF AREF}
	    fin {}
	    debut {}
	}

	declareTag WRI -save getHieroAsText -load DicoTk::insertWriting -index 1 \
	    -htmlsave getHieroAsHTML\
	    -latexsave getHieroAsLatex\
	    -normalize DicoTk::normalizeHieroglyphs\
	    -description [legendes::trans "written form of a word"]
	
	declareTag DET -save getHieroAsText -load DicoTk::insertDeterminative -index 1 \
	    -htmlsave getHieroAsHTML\
	    -latexsave getHieroAsLatex\
	    -normalize DicoTk::normalizeHieroglyphs\
	    -description [legendes::trans "Possible determinative for a word"]
	
	declareTag HIE -save getHieroAsText -load DicoTk::insertHiero \
	    -htmlsave getHieroAsHTML\
	    -latexsave getHieroAsLatex\
	    -normalize DicoTk::normalizeHieroglyphs\
	    -description [legendes::trans "Hieroglyphic Text"]
	
	declareTag TRAD  \
		-begin [legendes::trans "\ntranslation : "]\
		-end ".\n"\
		-bconf "-foreground red"\
		-htmlbegin [concat "\n<dl>\n<dt>" [legendes::trans "\ntranslation : "] "<dd>"]\
		-htmlend "</dd></dl>\n"\
		-conf {-font {Times 14 italic}}\
		-index 1\
		-description [legendes::trans "translation"]
    
	declareTag TRI \
		-begin " "\
		-end " "\
		-conf {-foreground blue -font {mdctranslitlc 18}} \
		-htmlbegin {<FONT face="Transliteration">}\
		-htmlend {</FONT>}\
		-normalize seshnormalize::norm_translit\
		-description [legendes::trans "Text in transliteration"]
    
	declareTag MTRI \
		-begin " "\
		-end " "\
		-conf {-foreground blue -font {mdctranslitlc 18}} \
		-index 1\
		-htmlbegin {<FONT face="Transliteration">}\
		-htmlend {</FONT>}\
		-normalize seshnormalize::norm_translit\
		-description [legendes::trans "Transliteration of a word"]
	
	
	declareTag COP \
		-begin " "\
		-end " "\
		-conf {-font {Coptic 18}}\
		-htmlbegin {<FONT face="Coptic">}\
		-htmlend {</FONT>}\
		-keybind {
	                 $ 
	                  }\
		-description [legendes::trans "Coptic text"]
	    
    
        declareTag COM \
		-begin [legendes::trans "\ncomment : "]\
		-end ".\n"\
		-htmlbegin "<P>"\
		-htmlend "</P>\n\n"\
		-bconf "-foreground blue"\
		-conf {-font {Times 14} -lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
		-description [legendes::trans "Free text comment"]
    

        declareTag REF \
		-begin {[} -end {]} \
		-conf {-foreground darkblue -font {helvetica 13}}\
		-bindings {
	    {<Double-Button-1> {DicoTk::callWindowCommand DicoTk::seeRef %W}}
	    {<Enter> {DicoTk::callWindowCommand DicoTk::setBlue %W}}
	    {<Leave> {DicoTk::callWindowCommand DicoTk::unsetBlue %W}}
	}\
	        -description [legendes::trans "Text reference"]

    declareTag AREF \
	-begin {[} -end {]} \
	-conf {-foreground darkblue -font {helvetica 13}}\
        -description [legendes::trans "External reference"]
    
    declareTag GRA \
	-begin "(" \
	-end ")" \
	-conf "-foreground blue"\
        -description [legendes::trans "grammar"]

    declareTag  USE\
        -begin [legendes::trans "expr : "] \
	-bconf "-font {helvetica 13}"\
	-end "." \
	-htmlbegin {<P><FONT face="Transliteration">}\
	-htmlend "</FONT></P>\n\n"\
	-conf "-foreground blue -font {mdctranslitlc 18}"\
        -description [legendes::trans "Expression"]
    
    declareTag  EXE\
        -begin [legendes::trans "example : "] \
	-bconf "-font {helvetica 13}"\
	-end "." \
	-htmlbegin {<BLOCKQUOTE><FONT face="Transliteration">}\
	-htmlend "</Font></BLOCKQUOTE></P>\n\n"\
	-conf "-foreground blue -font {mdctranslitlc 18}"\
       -description [legendes::trans "Example"]

    declareTag MOT \
	-begin [legendes::trans "<word" ] \
	-end ">" \
	-insert_at_end 1\
	-htmlbegin "~~"\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<WORD>"]
	    
    declareTag EMP \
	-begin {[} -end {]} \
	-htmlbegin {[} -htmlend {]} \
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "Domain and type of uses"]
	    
    declareTag ABRV \
	-begin [legendes::trans ".abbrev"] -end {.} \
        -insert_at_end 1\
	-htmlbegin { <Small>ABR.</Small> }\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<abbreviation>"]

    declareTag LATTER \
	-insert_at_end 1\
	-htmlbegin { <Small>LATTER</Small> }\
	-begin [legendes::trans ".later"] -end {.} \
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Later value>"]

    declareTag VAR \
	-insert_at_end 1\
	-htmlbegin { <Small>VAR.</Small> }\
	-begin [legendes::trans ".var"] -end {.} \
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Variant>"]
    
    declareTag OK \
	-insert_at_end 1\
	-begin [legendes::trans ".Old Kingdom"] -end {.} \
	-htmlbegin [concat " <Small>" [legendes::trans ".Old Kingdom"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Old Kingdom>"]

    declareTag NK \
	-insert_at_end 1\
	-begin [legendes::trans ".New Kingdom"] -end {.} \
	-htmlbegin [concat " <Small>" [legendes::trans ".New Kingdom"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<New Kingdom>"]

    declareTag LE \
	-insert_at_end 1\
	-begin [legendes::trans ".Late Egyptian"] -end {.} \
	-htmlbegin [concat " <Small>" [legendes::trans ".Late Egyptian"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Late Egyptian>"]

    declareTag QQCHOSE \
	-insert_at_end 1\
	-begin [legendes::trans "<Something"] \
	-end ">" \
	-htmlbegin [concat " <Small>" [legendes::trans "Something"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Something>"]

    declareTag QQUN \
	-insert_at_end 1\
	-begin [legendes::trans "<Someone"] \
	-end ">" \
	-htmlbegin [concat " <Small>" [legendes::trans "Someone"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Someone>"]

    declareTag AUT \
	-begin [legendes::trans "(Author :"] \
	-end ")" \
	-bconf "-font {helvetica}"\
	-conf "-foreground blue"\
	-description [legendes::trans "Signature"]

    declareTag GRP \
	-begin "\n" \
	-end "\n" \
	-bconf "-foreground red -spacing1 10"\
	-econf "-foreground red"\
	-conf {-lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
	 -description [legendes::trans "Group"]
	
    declareTag NOTE \
	-begin "\n" \
	-end "\n" \
	-bconf "-foreground red -spacing1 10"\
	-econf "-foreground red"\
	-conf {-lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
	-description [legendes::trans "Free text note"]

    declareTag TITLE \
	    -index 1\
	    -begin "-" \
	    -end "-\n" \
	    -htmlbegin {<H1>}\
	    -htmlend "</H1>\n"\
	    -latexbegin {\begin{center}\huge } \
	    -latexend {\end{center}} \
	    -conf {-justify center -foreground blue -font {helvetica 18 bold}}\
	    -description [legendes::trans "note title"]

    declareTag SUBTITLE \
	    -begin " " \
	    -end "\n" \
	    -htmlbegin {<H2>}\
	    -htmlend "</H2>\n"\
	    -latexbegin "\\section{" \
	    -latexend "}\n" \
	    -conf {-justify left -foreground blue -font {helvetica 16 bold}}\
	    -description [legendes::trans "section title"]


    declareTag SECTION \
	    -begin "\n" \
	    -end "\n" \
	    -bconf "-foreground red -spacing1 10"\
	    -econf "-foreground red"\
	    -conf {-lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
	    -description [legendes::trans "Section of text"]
	
    declareTag PAR \
	    -begin " " \
	    -end "\n" \
	    -bconf "-foreground red -spacing1 10"\
	    -econf "-foreground red"\
	    -htmlbegin "<P>"\
	    -htmlend "</P>\n\n"\
	    -latexend "\n\n"\
	    -conf {-lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
	    -description [legendes::trans "Paragraph"]
	
	variable ATOMIC_TAGS
	set ATOMIC_TAGS {TRAD_}
    }

    # Return a couple tag + value for indexing the base.
    # The value is normalized.

    proc tag_index {tag value} {
	set normValue [getNormalizedValue $tag $value]
	return [list $tag $normValue $value]
    }

    proc getNormalizedValue {tag value} {
	variable THE_TAGS
	return [$THE_TAGS($tag,normalize) $value]
    }

    proc identity {x} {
	return $x
    }

    # TagHandler::listIndexedTags --
    #          returns the tags to be indexed from a dictionary entry
    # Arguments:
    #   txt : text description of the entry
    #
    # Results:
    #        A list of elements {attributes value}, 
    #        describing the data to be saved

    proc listIndexedTags {txt} {
	set tmp [listTagsAux $txt]
	return [uniq $tmp]
    }
    
    # TagHandler::listTagsAux --
    #          returns the tags to be indexed from a dictionary entry
    # Arguments:
    #   txt : text description of the entry
    #
    # Results:
    #        A list of elements {attributes value}, 
    #        describing the data to be saved

    proc listTagsAux {txt} {
	variable THE_TAGS
	set result ""
	if {[isTag $txt]} {
	    set tag_name [lindex $txt 0]
	    set tag_data [lindex $txt 1]

	    if {$THE_TAGS($tag_name,index)} {
		lappend result [tag_index $tag_name $tag_data]
	    } else {
		foreach i $tag_data {
		    set result [concat $result [listTagsAux $i]]
		}
	    }
	}
	return $result
    }

    # TagHandler::buildTaggedText --
    #          build a correct tag
    #
    # Arguments:
    #   tag : name of the tag
    #   val : content
    #
    # Results:
    #   a correct tag.
    #   Note that if the tag value is one element long,
    #   there will be accolades anyway.
    #   In another life, I'll make PCDATA a first class TAG

    proc buildTaggedText {TAG val} {
	if { $val == "" } {
	    return [list $TAG {}]
	} else {
	    if {[llength $val] > 1} {
		return [list $TAG $val]
	    } else {
 		return "$TAG {$val}"
	    }
	}
    }
    
    proc isInAtomicTag {ltags} {
	variable ATOMIC_TAGS
	return [regexp -- $ATOMIC_TAGS $ltags toto]
    }

    # Test if a given tag can be inserted in a given environment.
    # 	tag : the tag to be inserted
    #	tags : the environment (a list of actual window tags, with a number).
    
    proc isLegalTagInTags {tag ntags} {
	# First, it is illegal to write anything in a 
	# begin or end tag :
	# (the general case works for these !)
	#
	# get the innermost tag :
	set ltags [lsort -command compareTags $ntags]
	set innertag [getTagName [lindex $ltags end]]
	# Test if current tag is compatible with innertag :
	if {[lsearch -exact [getLegalTags $innertag] $tag] != -1} {
	    return 1
	} else {
	    return 0
	}
    }

    proc isTag {txt} {
	return [expr [llength $txt] != 1]
    }

    proc getLoadFunction {tagName} {
	variable THE_TAGS
	return $THE_TAGS($tagName,load)
    }

    proc hasLoadFunction {tagName} {
	return [expr {[getLoadFunction $tagName] ne ""}]
    }
 
    proc hasKeyBinding {tag key} {
	variable THE_TAGS
	return [info exists THE_TAGS($tag,bind,$key)]
    }

    proc getKeyBinding {tag key} {
	variable THE_TAGS
	return $THE_TAGS($tag,bind,$key)
    }

    proc getLegalTags {containerTag} {
	variable LEGALTAGS
	return $LEGALTAGS($containerTag)
    }

    proc getTagDescription {tag} {
	variable THE_TAGS
	return $THE_TAGS($tag,description)
    }

    proc getTagConfiguration {tag part} {
	variable THE_TAGS
	switch -- $part {
	    begin {
		return $THE_TAGS($tag,bconf)
	    }
	    end {
		return $THE_TAGS($tag,econf)
	    }
	    body {
		return $THE_TAGS($tag,conf)
	    }
	}
    }

    proc getTagBindings {tag} {
	variable THE_TAGS
	return $THE_TAGS($tag,bindings)
    }

    proc getBeginText {tag} {
	variable THE_TAGS
	return $THE_TAGS($tag,begin)
    }

    proc getEndText {tag} {
	variable THE_TAGS
	return $THE_TAGS($tag,end)
    }

    proc getMiddleText {tag} {
	variable THE_TAGS
	return $THE_TAGS($tag,middle)
    }

    proc isTagInsertAtEnd {tag} {
	variable THE_TAGS
	return $THE_TAGS($tag,insert_at_end)
    }

    proc getTagSaveProc {tag} {
	variable THE_TAGS
	return $THE_TAGS($tag,save)
    }
}