
# This class represent an affine transformation.

class AffineTransformation {

    # PI Constant :

    set (PI) [expr {acos(-1)}]
    # descr is a list of six values :
    # ax bx dx ay by cy, and the image of a point m (x,y) by the transformation shall be :
    # xt= ax*x + bx*y + dx ; yt= ay*x+ by*y + dy

    proc AffineTransformation {this descr} {
	set ($this,transfo) $descr
    }

    proc  ~AffineTransformation {this} {
    }

    # AffineTransformation::applyToCoords --
    # 	apply the transformation to a point described by x,y coordinates
    # Parameters:
    #	x,y : the points coordinates
    # Results:
    #	a list of two coordinates corresponding to the image of x,y
    # Side Effects:
    #	None
    # Use:
    # Called by:

    proc applyToCoords {this x y} {
	set res [list]
	foreach {a b d} $($this,transfo) {
	    lappend res [expr {$a*$x+$b*$y+$d}]
	}
	return $res
    }

    # AffineTransformation::apply --
    # 	apply the transformation to a point described by x,y coordinates
    # Parameters:
    #	m : a point
    # Results:
    #	none
    # Side Effects:
    #	The point's coordinates are modified.
    # Use:
    # Called by:

    proc apply {this m} {
	foreach {x y} [applyToCoords $this [Point::x $m] [Point::y $m]] {
	    Point::setcoords $m $x $y
	}
    }


    # AffineTransformation::makeRotation --
    # 	build a rotation
    # Parameters:
    #	center : point
    #   angle : angle in radian.
    # Results:
    #	a point, being m's image.
    # Side Effects:
    #	None
    # Use:
    # Called by:

    proc makeRotation {center angle} {
	set c [expr {cos($angle)}]
	set s [expr {sin($angle)}]
	return [makeRotationByCosAndSin $center $c $s]
	
    }

    # AffineTransformation::makeRotation --
    # 	build a rotation, giving the center and 
    #	the cosinus and sinus of the rotation.
    #	it may seems strange to use such a function,
    #	but it's usually easier from a graphic interface
    #	to designate a rotation by example (which gives a cosinus and a sinus)
    #	than by angle.
    # Parameters:
    #	center : point
    #   c, s : cosinus and sinus of the rotation angle.
    # Results:
    #	a point, being m's image.
    # Side Effects:
    #	None
    # Use:
    # Called by:
    proc makeRotationByCosAndSin {center c s} {
	set xc [Point::x $center]
	set yc [Point::y $center]
	
	# Image of 0 :
	set x0 [expr {$xc * (1 - $c) + $s * $yc}]
	set y0 [expr {-$s*$xc + (1-$c)*$yc}]
	
	set tr [list $c [expr {-$s}] $x0 $s $c $y0]
	return [new AffineTransformation $tr]
    }
    
    
    proc makeLineReflection {p v} {
	Point::normalize $v
	set px [Point::x $p]
	set py [Point::y $p]
	set vx [Point::x $v]
	set vy [Point::y $v]

	set ax [expr {2*$vx*$vx -1}]
	set bx [expr {2*$vx*$vy}]
	set dx [expr {2*((1-$vx*$vx)*$px - $py*$vx*$vy)}]

	set ay [expr {2*$vx*$vy}]
	set by [expr {2*$vy*$vy -1}]
	set dy [expr {2*((1-$vy*$vy)*$py - $px*$vx*$vy)}]

	set tr [list $ax $bx $dx $ay $by $dy]
	return [new AffineTransformation $tr]
    }

    
}