#
# Widget to display and manage the tksesh font database.
#
package require seshw
package require tkDB

class CharDBNavigator {
    
    proc CharDBNavigator {this parent appli name wf args} SScrolledWindow {$parent $name $args} Observer {} {
	set ($this,appli) $appli
	#set ($this,wf) [new CharDBWorkflow [TkseshDB::getTkseshDB]]
	set ($this,wf) $wf
	Observable::add $($this,wf) $this
    }

    proc ~CharDBNavigator {this} {
	delete $($this,scrollArea)
    }

    virtual proc creatorHook {this} {
	SScrolledWindow::_creatorHook $this
    }
    
    proc getImageName {this} {
	return Image.$this
    }

    virtual proc completorHook {this} {
	# Create the inner scrollable frame
	set ($this,scrollArea) \
	    [Component::buildNewComponent SScrollableFrame $this -auto -areawidth 0]
	# The frame we can add things into :
	set ($this,frame) [FrameContainer::getFrame $($this,scrollArea)]
	set p [Component::getPath $($this,frame)]
	set img [image create bitmap [getImageName $this]]
	frame $p.glyphDraw -width 100 -height 100
	
	LabelEntry $p.filter -label "Filter" -side top \
		-command [list CharDBNavigator::setFilter $this]\
		-textvariable CharDBNavigator::($this,filter)

	# Code widget :
	set wf $($this,wf)
	set codeDS [CharDBWorkflow::getCodeDS $wf]
	set glyphDS [CharDBWorkflow::getGlyphDS $wf]
	
	
	set ($this,code) \
	    [Component::buildNewComponent BDBField $($this,frame) code $codeDS "code"]
	
	set ($this,codeNavigator) \
	    [Component::buildNewComponent BDBNavigator $($this,frame) codeNav $codeDS]

	LabelEntry $p.userCode -label "User Glyph Code" -side top -editable 0\
		-textvariable CharDBNavigator::($this,glyphCode)
	LabelEntry $p.phoncode -label "phonetic code" -side top -state disabled
	
	set ($this,glyphNavigator) \
	    [Component::buildNewComponent BDBNavigator $($this,frame) glyphNav $glyphDS]
	
	Button $p.open -text "open char" -command [list CharDBNavigator::openChar $this]
	Button $p.new -text "new char" -command [list CharDBNavigator::newChar $this]
	
	grid $p.glyphDraw - -sticky news
	grid $p.filter - -sticky ew
	grid $p.code $p.codeNav -sticky ew
	grid $p.userCode - -sticky ew
	grid $p.phoncode - -sticky ew
	grid $p.glyphNav -sticky ew
	grid $p.open - -sticky ew
	grid $p.new - -sticky ew
	SScrolledWindow::setWidget $this $($this,scrollArea)
	update $this $($this,wf) INIT
    }

    virtual proc update {this observable msg} {
	set ($this,filter) [CharDBWorkflow::getFilter $($this,wf)]
	if {[CharDBWorkflow::hasGlyph $($this,wf)]} {
	    # set ($this,code) [CharDBWorkflow::getCode $($this,wf)]
	    set ($this,glyphCode) [CharDBWorkflow::getCurrentGlyphCode $($this,wf)]
	    drawCurrentGlyph $this
	} else {
	    set ($this,code) ""
	}
    }

    proc setFilter {this} {
	CharDBWorkflow::setFilter $($this,wf) $($this,filter)
    }
    
    proc openChar {this} {
	# Open the current char
	# CharEditAppli::editDBChar $app [list CharDBNavigator::setChar $this $code] $charcode $this
	# The validator function is called with the char drawing as last argument.
	set app $($this,appli)
	# editChar arguments : appli drawing validator label
	set txt [CharDBWorkflow::getCurrentSignDrawing $($this,wf)]
	set code [CharDBWorkflow::getCurrentGlyphCode $($this,wf)]
	CharEditAppli::editDBChar $app $txt $code
    }
    
    proc newChar {this} {
	set app $($this,appli)
	# editChar arguments : appli drawing validator label
	CharEditAppli::editDBChar $app [list 0 0 0 0 0 0] [CharDBWorkflow::getNewGlyphLabel $($this,wf)]
    }
    
    # Fixes the value for a given char, and makes this char the current one.
    proc setChar {this code drawing} {
	CharDBWorkflow::setChar $($this,wf) $code $drawing
    }
    
    proc getCurrentGlyphCode {this} {
	return [CharDBWorkflow::getCurrentGlyphCode $($this,wf)]
    }
    # PRIVATE
    proc drawCurrentGlyph {this} {
	set d [DummyDrawer::getDrawer]
	set drawing [CharDBWorkflow::getCurrentSignDrawing $($this,wf)]
	#puts "drawing is $drawing"
	tksesh_defchar tmp $drawing
	seshw::set_text $d tmp
	seshw::autosize $d
	set data  [$d toxbm 1.0 end]
	[getImageName $this] configure -data $data
	set p [Component::getPath $($this,frame)]
	
	catch {destroy $p.glyphDraw.label}
	label $p.glyphDraw.label -image [getImageName $this]
	pack $p.glyphDraw.label
	#$p.glyphDraw configure -text "HELLO"
	#$p.glyphDraw configure -image [getImageName $this]
    }

}

namespace eval DummyDrawer {
    variable isLoaded
    set isLoaded 0
    proc getDrawer {} {
	variable isLoaded
	if {! $isLoaded} {
	    set isLoaded 1
	    seshw::create_passive .seshBitmapDrawer -htline 120
	}
	return .seshBitmapDrawer
    }
}
