
# Cleaner and nicer version of the sign editor

# * Using the nice BWidgets
# * interfaced with the database
# * multiple windows
# * undo/redo facilities

# Export facilities for bzrfonts
# dynamic cut and paste.

# BWidget Implementation

# Note on font edition :
# Load Font : create a new font window in the navigator.
# New Font 
# Close Current Font
# 
# Interaction between the font windows and the char windows :
# Double-click on a char in the font windows : create an editor for this
# 	char. Font name and char number appear in the title. 
#
#
# drag and drop ????
#
#

package require seshuser
seshuser::loaduser

package require myStoopWidgets
package require BWidget 1.3
package require BWidgetBindImage
package require legendes
package require svg
#TitleFrame .tf


# Is both a mediator and the top level GUI. Should be cut in two

class CharEditAppli {


    proc CharEditAppli {this parent name args} SMainFrame {$parent $name $args} {
	set ($this,numEdit) 0
	set ($this,numFont) 0

	#setEditorObject $this -1 -1
	#setEditorObject $this 0
    }

    proc ~CharEditAppli {this} {
    }

    
    # enable or disable the menus
    # which make sense only for a font menu
    proc enableFontMenu {this value} {
	SMainFrame::setMenuState $this fontMenu  $value
    }

    # enable or disable the menus
    # which make sense only for a char editor

    proc enableCharEditMenu {this value} {
	SMainFrame::setMenuState $this charEditMenu $value
    }
    
    virtual proc creatorHook {this} {
	# Menu is a read-only option. We must fix it
	# before widget creation
	set descmenu [buildMenu $this]
	Component::addArgs $this -menu $descmenu

	# Call parent procedure to create the widget
	SMainFrame::_creatorHook $this

	
	Component::configure $this -textvariable CharEditAppli::($this,status)
	#Component::pack  $this -fill both -expand yes
	SMainFrame::addIndicator $this -text "The TKsesh Font Editor"
	SMainFrame::addIndicator $this -text "Version [getVersion $this] ([getDate $this])"

	enableCharEditMenu $this 0
	enableFontMenu $this 0

	set ($this,status) ready

	set paned [Component::buildNewComponent SPanedWindow [FrameContainer::getFrame $this] -auto -side bottom]
	
	set p1 [SPanedWindow::add $paned -minsize 300]
	set p2 [SPanedWindow::add $paned -minsize 100]

	set ($this,container) $paned
	set ($this,chars) $p1
	set ($this,directories) $p2
	
	# The toolbar is not really a widget.
	
	set ($this,toolbar) [new CharEditToolBar $this]
	
	# charBook and noteBook creation
	set ($this,charBook) [Component::buildNewComponent SNoteBook $p1 -auto -width 200 -height 300]
	set ($this,fontBook) [Component::buildNewComponent SNoteBook $p2 -auto -width 200 -height 300]
	
	#$win.book compute_size
	
	grid [Component::getPath $paned] -sticky news
	set frame [Component::getPath [FrameContainer::getFrame $this]]
	grid columnconfigure $frame 0 -weight 1 
	grid rowconfigure $frame 0 -weight 1 -minsize 300
	# Component::grid $paned -fill both -expand 
	Component::pack [getCharBook $this] -fill both -expand yes
	Component::pack [getFontBook $this] -fill both -expand yes
	# Page id will be added to the next command.
	SNoteBook::bindimages [getCharBook $this] <ButtonPress-1> \
	    "CharEditAppli::deleteSignWindow $this"
	
	# Create a CharDBNavigator and add it to the fontBook :
	createCharDB $this
    }

    proc createCharDB {this} {
	# fix a problem with scrollarea on macs.
	set isMac 0
	if {$::tcl_platform(platform) eq "macintosh"} {
	    set isMac 1
	}
	
	# First, create the char database interface
	set page [SNoteBook::insert [getFontBook $this] end]
	SNoteBookPage::itemConfigure $page -raisecmd [list CharEditAppli::handleRaisedFont $this]
	SNoteBookPage::itemConfigure $page -text "Sign Database"

	set ($this,charDBWorkFlow) [new CharDBWorkflow [TkseshDB::getTkseshDB]]

	if {$isMac} {
	    set fd [Component::buildNewComponent \
			CharDBNavigator [FrameContainer::getFrame $page] $this -auto $($this,charDBWorkFlow) -auto none]
	} else {
	    set fd [Component::buildNewComponent \
			CharDBNavigator [FrameContainer::getFrame $page] $this -auto $($this,charDBWorkFlow)]
	}
	
	set ($this,charDBNavigatorPage) $page
	set ($this,charDBNavigator) $fd
	
	Component::pack $fd -fill both -expand yes
	SNoteBookPage::raise $page
	SNoteBookPage::setObject $page $fd

	# Now, create a page to edit char values :
	set pageVal [SNoteBook::insert [getFontBook $this] end]
	SNoteBookPage::itemConfigure $pageVal -raisecmd [list CharEditAppli::handleRaisedFont $this]
	SNoteBookPage::itemConfigure $pageVal -text "Sign Values"

	if {$isMac} {
	    set valobj [Component::buildNewComponent \
			    CharDBValueEditor [FrameContainer::getFrame $pageVal] $this -auto $($this,charDBWorkFlow) -auto none]
	} else {
	    set valobj [Component::buildNewComponent \
			    CharDBValueEditor [FrameContainer::getFrame $pageVal] $this -auto $($this,charDBWorkFlow)]
	}
	set ($this,charDBValuePage) $pageVal
	set ($this,charDBValueEditor) $valobj

	Component::pack $valobj -fill both -expand yes
	SNoteBookPage::raise $pageVal
	SNoteBookPage::setObject $pageVal $valobj

	SNoteBookPage::raise $page
	


	# return $page
    }
    
    virtual proc completorHook {this} {
    }

    proc getVersion {this} {return 2}

    proc getDate {this} {return {Mon Dec 24 17:13:59 CET 2001}}
    
    proc buildMenu {this} {
	# TAGS for menus :
	# 	charEditMenu : this menu works on a char editor
	#	fontMenu : this menu works on a font editor
	#	
	set res {
	    "&File" all file 0 {
		{
		    command "&New Font" {} "Create a new BZR Font" {}
		    -command {CharEditAppli::newFont WINDOW}
		}
		{
		    command "&Load Font" {} "Load a font from a bzr file" {}
		    -command {CharEditAppli::loadFont WINDOW}
		}
		{
		    command "&Save Font" {fontMenu} "save a font in a bzr file" {}
		    -command {CharEditAppli::saveFont WINDOW}
		}
		{
		    command "&Save Font As" {fontMenu} "save a font in a bzr file" {}
		    -command {CharEditAppli::saveFontAs WINDOW}
		}
		{
		    command "&Export Font to MF" {fontMenu} "save a font in a metafont file for use with HieroTeX" {}
		    -command {CharEditAppli::exportFontMF WINDOW}
		}
		{
		    command "&Export Font to Type 1" {fontMenu} "save a font as a postscript type 1 file." {}
		    -command {CharEditAppli::exportFontType1 WINDOW}
		}
		{
		    command "Delete Font window" {fontMenu} "Delete current font window" {}
		    -command {CharEditAppli::deleteUpperFontWindow WINDOW}
		}
		separator
		{
		    command "&Delete sign window" {charEditMenu} "Delete current sign window" {}
		    -command {CharEditAppli::deleteUpperSignWindow WINDOW}
		}
		{
		    command "Save char to file" {charEditMenu} \
			"Save the sign displayed in the database in a file." {}
		    -command {CharEditAppli::saveChar WINDOW}
		}
		{
		    command "Save all personnal glyphs to file" {charEditMenu} \
			"Save all the user's glyphs in a file." {}
		    -command {CharEditAppli::saveAllGlyphs WINDOW}
		}
		{
		    command "Load char from file" {charEditMenu} \
			"load a sign into the database" {}
		    -command {CharEditAppli::loadChar WINDOW}
		}
		{
		    command "Import SVG drawing" {charEditMenu} \
			"import an svg drawing file" {}
		    -command {CharEditAppli::svgImport WINDOW}
		}
		{
		    command "&Validate sign window" {charEditMenu} "Save the current sign where it belongs" {}
		    -command {CharEditAppli::validateUpperSign WINDOW}
		}
		separator
		{
		    command "Load Image" {charEditMenu} "Load Background image" {} -command {CharEditAppli::loadImage WINDOW}
		}
		{
		    command "Clear Image" {charEditMenu} "Clear Background image" {} -command {CharEditAppli::clearImage WINDOW}
		}
		separator
		{
		    command "E&xit" {} "Exit BWidget demo" {} -command exit
		}
	    }

	    "&Edit" all edit 0 {
		{
		    command "&Undo" {charEditMenu} "Undo last command" {}
		    -command {CharEditAppli::undo WINDOW}
		}
		separator
		{
		    command "Cut" {charEditMenu} "Cut selected font part" {}
		    -command {CharEditAppli::cut WINDOW}
		}
		{
		    command "Copy" {charEditMenu} "Copy selected font part " {}
		    -command {CharEditAppli::copy WINDOW}
		}
		{
		    command "Paste" {charEditMenu} "Paste  selected font part" {}
		    -command {CharEditAppli::paste WINDOW}
		}
		{
		    command "SelectAll" {charEditMenu} "Select all" {Ctrl a}
		    -command {CharEditAppli::selectAll WINDOW}
		}
		{
		    command "DeselectAll" {charEditMenu} "Deselect all" {}
		    -command {CharEditAppli::deselectAll WINDOW}
		}
	    }
	    "&Options" all options 0 {
	    }
	}
	set res [string map [list WINDOW $this] $res]
	return $res
    }
    
    proc getCharBook {this} {
	return $($this,charBook)
    }
    
    proc getFontBook {this} {
	return $($this,fontBook)
    }
    
    # Create a new sign editing window.
    # returns the created page.

    proc createSignWindow {this} {
  	set page [SNoteBook::insert [getCharBook $this] end -image ::icon::fileclose]

	SNoteBookPage::itemConfigure $page -raisecmd [list CharEditAppli::handleRaisedChar $this]
	
	# as the canvas hasn't got any scaling facilities,
	# using the canvas scroll commands would require us to use *THREE* coordinate systems.
	# So, we do it by hand.

	set frame [FrameContainer::getFrame $page]
	set editorWidget  [Component::buildNewComponent CharEdit $frame -auto]
	set e [Component::getPath $editorWidget]
	set framePath [Component::getPath $frame]

	scrollbar $framePath.xs -command "CharEdit::xview $editorWidget" -orient horizontal
	scrollbar $framePath.ys -command "CharEdit::yview $editorWidget" -orient vertical

	CharEdit::setscale $editorWidget 10 20 300
	CharEdit::repere $editorWidget
	CharEdit::setXScrollBar $editorWidget $framePath.xs
	CharEdit::setYScrollBar $editorWidget $framePath.ys
	CharEdit::notifyScrollBars $editorWidget

	grid $e $framePath.ys -sticky news
	grid $framePath.xs - -sticky ew
	grid columnconfigure $framePath 0 -weight 1
	grid rowconfigure $framePath 0 -weight 1

	#setEditorObject $this $page $editorWidget
	
	# Wiiiiilld addition to SNoteBookPage :
	# (instead of deriving a proper new class)
	#set SNoteBookPage::($page,editor) $editorWidget
	
	SNoteBookPage::setObject $page $editorWidget
	
	pack $framePath -fill both -expand yes
	
	SNoteBookPage::raise $page

	return $page
    }
    
    # undo --
    # 	undo the last command in the current font editor
    #

    proc undo {this} {	
	CharEdit::undo [getCurrentEditorObject $this]
    }
    
    proc getCurrentEditorPage {this} {
	set book [getCharBook $this]
	return [SNoteBook::getUpper $book]
    }

    proc getCurrentFontPage {this} {
	set win [getFontBook $this]
	return [SNoteBook::getUpper $win]
    }

    proc getCurrentEditorObject {this} {
	return [SNoteBookPage::getObject [getCurrentEditorPage $this]]
    }

    proc getCurrentFontDisplay {this} {
	return [SNoteBookPage::getObject [getCurrentFontPage $this]]
    }

    proc deleteUpperFontWindow {this} {
	set win [getFontBook $this]
	set page [getCurrentFontPage $this]
	if {$page != 0 && $page != $($this,charDBNavigatorPage)} {
	    set fwin [getCurrentFontDisplay $this]
	    set allClear 1
	    foreach s [getCharEditsForFont $this $fwin] {
		set allClear [expr $allClear && [deleteSignWindow $this $s]]
	    }
	    if {! $allClear} {
		MessageDlg .couldNotDo -icon error -type ok \
			-message "There are still edited signs in this font.\n I could not delete the font window."
		return
	    }
	    if {[FontDisplay::getChanged $fwin]} {
		set answer [MessageDlg .validateDestroy -icon warning -type okcancel \
			-message "This font hasn't been saved.\nYou will lose your edits.\nProceed ?"]
		if {$answer == 1} {
		    return
		}
	    }
	    
	    SNoteBook::deletePage $win $page
	    handleRaisedFont $this
	}
    }

    
    proc deleteUpperSignWindow {this} {
	set result 0
	set win [getCharBook $this]
	# catch to protect us if there is no page to destroy.
	set page [getCurrentEditorPage $this]
	if {$page != 0} {
	    set result [deleteSignWindow $this $page]
	}
	return $result
    }

    # CharEditAppli::deleteSignWindow --
    # 	deletes a sign displaying window
    # Parameters:
    #	w : the chareditappli the window belongs to
    #	p : the page to delete
    # Options:
    # Results:
    # 	1 if the window was deleted else 0 
    # Side Effects:
    #	the page is deleted
    # Use:
    #	PUBLIC
    # Called by:
    
    proc deleteSignWindow {this p} {
	set charbook [getCharBook $this]
	set edit [SNoteBookPage::getObject $p]
	if {[CharEdit::getChanged $edit]} {
	    set label [SNoteBookPage::itemCGet $p text]
	    set answer [MessageDlg .validateDestroy -icon warning -type okcancel -default 1\
		    -message "char $label hasn't been saved.\nYou will lose your edits.\nProceed ?"]
	    if {$answer == 1} {
		return 0
	    }
	}
	unRegisterCharEditForFont $this $p
	SNoteBook::deletePage $charbook $p
	handleRaisedChar $this
	return 1
    }

    proc validateUpperSign {this} {
	set page [getCurrentEditorPage $this]
	if {$page != 0} {
	    set edit [getCurrentEditorObject $this]
	    if {[CharEdit::getChanged $edit]} {
		set txt [CharEdit::asText $edit]
		CharEdit::setChanged $edit 0
		eval [getCharWValidator $this $page] [list $page] [list $txt]
	    }
	}
    }
    
    # loadFont --
    # 	Loads a BZR font
    #	Create a corresponding font window.

    proc loadFont {this} {
	set patterns {{{BZR file} {.bzr}} {Any *}}
	set name [tk_getOpenFile \
		-defaultextension .bzr -filetypes $patterns]
	
	if {$name eq ""} return
	set b [new BZRFont]
	BZRFont::load $b $name
	
	createFontWindow $this $b
	# temporary :
	#  	createSignWindow $this
	#  	set e [getCurrentEditor $this]
	#  	set txt [BZRFont::getChar $b 32]
	#  	CharEdit::setText $e $txt
	# SUPRESS THE NEXT LINE : 
	# set VAR($this,font) $b
    }

    proc newFont {this} {
	set b [new BZRFont]
	createFontWindow $this $b
    }

    # saveFont --
    # 	save a BZR font

    proc saveFont {this} {
	if {[getCurrentFontPage $this] != 0} {
	    set fwin [getCurrentFontDisplay $this]
	    set font [FontDisplay::getFont $fwin]
	    BZRFont::saveAs $font [BZRFont::getFileName $font]
	    FontDisplay::setChanged $fwin 0
	}
    }

    proc saveFontAs {this} {
	if {[getCurrentFontPage $this] != 0} {
	    set patterns {{{BZR file} {.bzr}} {Any *}}
	    set name [tk_getSaveFile \
		    -defaultextension .bzr -filetypes $patterns]
	    set fwin [getCurrentFontDisplay $this]
	    set font [FontDisplay::getFont $fwin]
	    BZRFont::setFileName $font $name
	    BZRFont::saveAs $font $name
	    SNoteBookPage::itemConfigure [getCurrentFontPage $this] -text $name
	    FontDisplay::setChanged $fwin 0
	}
    }
    
    proc exportFontMF {this} {
	if {[getCurrentFontPage $this] != 0} {
	    set patterns {{{Metafont file} {.mf}} {Any *}}
	    set name [tk_getSaveFile \
		    -defaultextension .mf -filetypes $patterns]
	    set fwin [getCurrentFontDisplay $this]
	    set font [FontDisplay::getFont $fwin]
	    BZRFont::exportAsMF $font $name
	}
    }
    
    proc exportFontType1 {this} {
	if {[getCurrentFontPage $this] != 0} {	
	    set patterns {{{postscript type 1} {.pfa}} {Any *}}
	    set name [tk_getSaveFile \
		    -defaultextension .pfa -filetypes $patterns]
	    set fwin [getCurrentFontDisplay $this]
	    set font [FontDisplay::getFont $fwin]
	    BZRFont::exportAsType1 $font $name
	}
    }

    # displayChar --
    # 	displays the num th char from font in the uppermost window.
    #proc displayChar {this font num} {}

    # setCharFromEditor --
    #	sets the numth char in font from the editor window ed
    #proc setCharFromEditor {this font num edi} {}

    # clearCharEdit --
    #	clears a char editor.
    #	if the char has been edited and not saved,
    #	display a warning message.
    #
    #	returns:	1 if we should proceed, 0 if this operation was cancelled.

    proc clearCharEdit {this ced} {
	if {[CharEdit::getChanged $ced]} {
	    MessageDialog .reallyClear -title "Really clear?" -icon question -type okcancel\
		    -message "Warning : char has been modified, but not saved.\n Really clear it ?"
	    set res [Message::create .reallyClear]
	    if {$res != 0} {
		return 0
	    }
	}
	CharEdit::clear $ced
	CharEdit::setChanged 0
	return 1
    }

    # createFontWindow --
    # 	creates a new font window.
    #	w the base window application
    #	b the font
    #
    #	returns : an id for the the window.
    
    proc createFontWindow {this b} {
	set page [SNoteBook::insert [getFontBook $this] end]

	SNoteBookPage::itemConfigure $page -raisecmd [list CharEditAppli::handleRaisedFont $this]

	SNoteBookPage::itemConfigure $page -text [BZRFont::getFileName $b]
	
	#set fd [FontDisplay::create [getFontWindow $this $pageID] $this]
	set fd [Component::buildNewComponent FontDisplay [FrameContainer::getFrame $page] $this -auto]
	FontDisplay::setFont $fd $b

	Component::pack $fd -fill both -expand yes
	
	SNoteBookPage::raise $page
	SNoteBookPage::setObject $page $fd
	initCharEditsForFont $this $fd
	return $page
    }


    # This proc is called when the raised font window change.
    # it fixes menus and the like accordingly

    proc handleRaisedFont {this} {
	set page [SNoteBook::getUpper [getFontBook $this]]
	if {$page == 0 || $page == $($this,charDBNavigator)} {
	    enableFontMenu $this 0
	} else {
	    enableFontMenu $this 1
	}
    }

    # This proc is called when the raised char window change.
    # it fixes menus and the like accordingly

    proc handleRaisedChar {this} {
	set page [SNoteBook::getUpper [getCharBook $this]]
	if {$page == 0} {
	    enableCharEditMenu $this 0
	} else {
	    CharEditToolBar::setEditor $($this,toolbar) [SNoteBookPage::getObject $page]
	    enableCharEditMenu $this 1
	}
    }

    # editChar --
    #	open an editing window for a char
    #	
    #	w : the application window
    #	txt : the char described in the usual font format (see CharEdit.tcl)
    #	validator : a function call which shall be eval-uated when the char is validated.
    #			it should either save the char in its font (for bzrfont chars)
    #			or save the char in the database.
    #			The last argument of the validator will be the text.
    #	label : label to use for the editchar window
    # Results:
    #	sends back the charEdit object which was created.
    
    
    proc editChar {this txt validator label} {
	set i [createSignWindow $this]
	# Set the name for the sign window.
	SNoteBookPage::itemConfigure $i -text $label
	set e [SNoteBookPage::getObject $i]
	CharEdit::setText $e $txt
	setCharWValidator $this $i $validator
	return $i
    }


    # truc::editFontChar --
    # 	edit a char from a font editor
    # Parameters:
    #	f : the font editor window
    #   num : the code for the char
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    #proc editFontChar {this txt validator label f} {}
    proc editFontChar {this fontDisplay num} {
	set font [FontDisplay::getFont $fontDisplay]
	set txt [BZRFont::getChar $font $num]
	set validator [list CharEditAppli::validateFontChar $this $fontDisplay $num]
	set label "[file tail [BZRFont::getFileName $font]]($num)"
	set page [editChar $this $txt $validator $label]
	registerCharEditForFont $this $fontDisplay $page
    }

    proc validateFontChar {this fontDisplay num page txt} {
	FontDisplay::setChar $fontDisplay $num $txt
    }
    
    # Edit a char from the database.
    proc editDBChar {this txt code} {
	# validator should be an object, not a callback.
	set validator "CharEditAppli::validateDBChar $this"
	set i [editChar $this $txt $validator $code]
    }
    
    # Validator for new Chars in the database.
    # calls the right WorkFlow functions,
    # AND ensures the label and names for the char are updated.
    proc validateDBChar {this page txt} {
	# puts "from $page with $txt"
	# get the code
	set code [SNoteBookPage::itemCGet $page text]
	# validate the char in the database
	CharDBNavigator::setChar $($this,charDBNavigator) $code $txt
	# retrieve the code from the charDBNavigator
	set code [CharDBNavigator::getCurrentGlyphCode $($this,charDBNavigator)]
	# fix the code in the SNoteBookPage
	SNoteBookPage::itemConfigure $page -text $code
	
    }

    # registerCharEditForFont --
    # 	remembers that a given char editor correspond to a given FontDisplay
    # Parameters:
    #	w :
    #	f : the font editor window
    #	e : the char editor *page*
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc registerCharEditForFont {this f e} {
	set ($this,isEditFor,$e) $f
	lappend ($this,fontwin,$f,charEdits) $e
    }

    # unRegisterCharEditForFont --
    #	un-register a charEdit. To be used when closing the said char editor.
    
    proc unRegisterCharEditForFont {this e} {
	# Will fail for database chars.
	catch {
	    set f [getFontForChar $this $e]
	    unset ($this,isEditFor,$e)
	    listreplace ($this,fontwin,$f,charEdits) $e
	}
    }
    
    proc getFontForChar {this e} {
	return $($this,isEditFor,$e)
    }

    # truc::fonction --
    # 	description...
    # Parameters:
    #	w :
    #	f : the font window
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc initCharEditsForFont {this f} {
	set ($this,fontwin,$f,charEdits) [list]
    }

    # truc::fonction --
    # 	description...
    # Parameters:
    #	w :
    #	f : the font *window*
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc getCharEditsForFont {this f} {
	return $($this,fontwin,$f,charEdits)
    }

    proc setCharWValidator {this i v} {
	set ($this,$i,validator) $v
    }
    
    proc getCharWValidator {this i} {
	return $($this,$i,validator)
    }

    proc copy {this} {
	setCharClipboard $this [CharEdit::getSelection [getCurrentEditorObject $this]]
    }

    proc paste {this} {	
	set txt [getCharClipboard $this]
	CharEdit::pasteAction [getCurrentEditorObject $this] $txt
    }

    proc selectAll {this} {
	CharEdit::selectAll [getCurrentEditorObject $this]
    }

    proc deselectAll {this} {
	CharEdit::deselectAll [getCurrentEditorObject $this]
    }

    proc cut {this} {
    }

    proc loadImage {this} {
	set patterns {{{gif file} {.gif}} {Any *}}
	set name [tk_getOpenFile \
		      -defaultextension .gif -filetypes $patterns]
	set e [getCurrentEditorObject $this]
	CharEdit::loadimg $e $name
    }

    proc clearImage {this} {
	set e [getCurrentEditorObject $this]
	CharEdit::delimg $e
    }

    proc setCharClipboard {this txt} {
	set ($this,clipboard) $txt
    }

    proc getCharClipboard {this} {
	return $($this,clipboard)
    }
    
    proc loadChar {this} {
	set patterns {{{Tksesh markup language} {.tml}} {Any *}}
	set wf $($this,charDBWorkFlow)
	set fname [tk_getOpenFile -defaultextension .tml -filetypes $patterns]
	if {$fname eq ""} return
	set f [open $fname]
	set t [read $f]
	close $f

	set code [CharDBWorkflow::importChar $wf $t]
	#puts "code is $code"
	if {$code ne ""} {
	    CharDBWorkflow::setFilter $wf $code
	}
    }

    proc saveChar {this} {
	set patterns {{{Tksesh markup language} {.tml}} {Any *}}
	set wf $($this,charDBWorkFlow)
	set t [CharDBWorkflow::exportCurrentChar $wf]
	set fname [tk_getSaveFile -defaultextension .tml -filetypes $patterns]
	set f [open $fname w]
	puts $f $t
	close $f
    }

    proc saveAllGlyphs {this} {
	set patterns {{{Tksesh markup language} {.tml}} {Any *}}
	set wf $($this,charDBWorkFlow)
	set t [CharDBWorkflow::exportUserChars $wf]
	set fname [tk_getSaveFile -defaultextension .tml -filetypes $patterns]
	set f [open $fname w]
	puts $f $t
	close $f
    }

    proc svgImport {this} {
	set patterns {{"svg file" {.svg}} {"xml file" {.xml}} {Any *}}
	set name [tk_getOpenFile \
		      -defaultextension .svg -filetypes $patterns]
	if {$name eq ""} return
	set f [open $name]
	set txt [read $f]
	close $f
	
	set l [SVGReader::svg2lst $txt]
	CharEdit::addtext [getCurrentEditorObject $this] $l
    }
}

set top [new STopLevel ""]
set appli [Component::buildNewComponent CharEditAppli $top -auto]
pack [Component::getPath $appli] -fill both -expand yes
update idletask
