# Everything related to glyphs encoding is stored in this package 

package provide glyphCode 1.0

namespace eval glyphCode {

    variable unknown 
    
    # Ordering function for glyphs
    proc compareCodes {s1 s2} {
	array set rgltr {A 1 B 2 C 3 D 4 E 5 F 6 G 7 H 8 I 9 J 10 K 11 L 12 M 13 N 14 \
			     O 15 P 16 Q 17 R 18 S 19 T 20 U 21 V 22 W 23 X 24 Y 25 Z 26 Aa 27}
	regexp {([A-Z]a?)([0-9]+)} $s1 dummy LTR1 COD1
	regexp {([A-Z]a?)([0-9]+)} $s2 dummy LTR2 COD2
	set ltrcmp [expr $rgltr($LTR1) - $rgltr($LTR2) ]
	if [expr $ltrcmp != 0] {
	    return $ltrcmp
	} else {
	    return [expr $COD1 - $COD2]
	}
    }

    # Build a code representation for a user glyph,
    # 	uid : the user id
    #	mid : the machine id
    #	id : the glyph id
    #
    # See also tables.sql
    #
    proc buildCode {uid mid id} {
	return "UG${uid}M${mid}N${id}"
    }

    # Build a pseudo-code if no id is known.
    proc buildNewGlyphLabel {uid mid} {
	return "UG${uid}M${mid}N??"
    }

    proc getUserGlyphDataFromCode {code Ruid Rmid Rid} {
	upvar $Ruid uid $Rmid mid $Rid id
	regexp {UG(.*)M(.*)N(.*)} $code dummy uid mid id
    }

    # isGardiner is true if code is either a User glyph code or a Gardiner code
    #
    # MODIFY THIS STUFF, to use getCodeType.
    proc isGardiner {code} {
	switch -- $code {
	    nn -
	    nTrw {return 1}
	    default {return [regexp -- {^(([A-Z]|Aa)[0-9]+)|[oO]|UG.*} $code]}
	}
	#return [regexp {^[A-Za-z]+[0-9]+[A-Za-z]*} $code]
	#return [regexp -- {^(([A-Z]|Aa)[0-9]+)|[oO]|2|[4-9]+|UG.*} $code]	
    }
    
    # Low level norm associates a canonical code to a sign.
    # signs as stored in "Manuel de codage" format can have multiple codes :
    #	* phonetic code
    #	* "Gardiner" code
    #   * User Glyph code
    #	* numeric code for numbers
    #
    # Some sign have only one or two of these codes.
    # 
    
    # find which kind of code this is
    # key :
    #	G : gardiner code
    #	N : numeric (1,2...)
    #	U : user glyph code (form : UGxxMxxNxx)
    #	P : phonetic code
    #	S : special codes from manuel de codage definition. (not yet used).
    #	    these codes are mostly considered as Gardiner codes.
    #	O : other
    #
    # Note that Phonetic codes are not phonetic values.
    # for instance, "m" is the phonetic code for sign G17
    # "M" is the phonetic code for sign Aa13
    # and both G17 and Aa13 have phonetic value "m".
    # 
    # Also, Aa13 (for instance) does correspond to a certain glyph code
    # but "M" is not a phonetic code for this glyph code.

    proc getCodeType {code} {
	switch -regexp -- $code {
	    {nn} {return G}
	    {nTrw} {return G}
	    {^`$} {return P}
	    {^([A-Z]|Aa)[0-9]+} {return G}
	    {^UG[0-9]+M[0-9]+N[0-9]+$}  {return U}
	    {^[oO]$} {return O}
	    {^[A-Za-z]+$} {return P}
	    {[0-9]+}  {return N}
	    default {return O}
	}
    }

    proc resetUnknown {} {
	variable unknown
	array unset unknown
    }

    proc addToUnknown {code} {
	variable unknown
	set unknown($code) 1
    }

    proc getUnknows {} {
	variable unknown
	return [array names unknown]
    }

    proc suppressFromUnknown {code} {
	variable unknown
	array unset unknown($code)
    }

}