package provide ligature 1.0
package require TkseshDB
package require glyphDAL
package require glyphCode

namespace eval ligature {
    
    # saves a ligature in the tksesh database.
    # The routine will associate a code to the ligature according to the
    # following preferences :
    # for a given sign, phonetic codes are preferred over Gardiner codes 
    # which are prefered over UserGlyphCodes
    
    proc saveLigature {command} {
	set code [PRIVGenerateCodeForLigature $command]
	glyphDAL::saveLigature $code $command
    }
    
    # The routine which generates a code for the ligature
    # See the rule above.

    proc PRIVGenerateCodeForLigature {command} {
	set signs {}
	foreach {glyph x y s} [lrange $command 1 end] {
	    set type [glyphCode::getCodeType $glyph]
	    switch -- $type {
		N -
		P {
		    lappend signs $glyph
		}
		U -
		G -
		O {
		    set equivs [glyphDAL::getEquivalentCodes $glyph]
		    set bestCode $glyph
		    # test each code (normally in fact only two)
		    foreach c $equivs {
			# If we find a phonetic code, we choose it.  
			set phon [glyphDAL::getPhoneticCodeForSignCode $c]
			if {[string length $phon] != 0} {
			    set bestCode $c
			    break
			}
			# if the code is a "gardiner code", it's a best working
			# hypothesis (given that a phonetic code will always win)
			if {[glyphCode::getCodeType $c] eq "G"} {
			    set bestCode $c
			}
		    }
		    lappend signs $bestCode
		}
	    }
	}
	
	# There are no space nor special character in codes, so 
	# the following is safe :
	return [string map {" " "&"} $signs] 
    }

    # Load (or reload) ligatures from the database and defines them.
    proc loadLigatures {} {
	glyphDAL::forEachLigature ligature::defineLigature
    }

    proc exists {code} {
	return [glyphDAL::ligatureExists $code]
    }

    #
    proc getLigature {code} {
	return [glyphDAL::getLigature $code]
    }

    # tells the tksesh widget about a ligature :
    proc defineLigature {code command} {
	#puts "defining $code as $command"
	eval tksesh_setlig $command
    }
    
}