# tradedit.tcl --
#
#        basic bindings for editing translation and translitteration
#
# Copyright (c) 1998 Serge Rosmorduc
#

# THIS FILE CONTENT IS CURRENTLY DEPRECATED.
# 

# Basic editing for translation files

# System architecture :
#
# The database contains various hieroglyphic texts, identified by their names.
# A project is the union of (at that time)
# One hieroglyphic text, one transliteration, one translation.
# n-to-n relationships between these types of elements is a project.

# Text format : TAG tabulation text
# TAGS :
# a) text coordinates
# b) NAME : name of hieroglyphic text
# c) TRANSLIT : name of translitteration

package provide tradedit 0.0
package require legendes
package require seshw
#package require dbfilesel

 
namespace eval tradedit {
    if {0} {
    namespace export create createmenu
    variable VARS

    proc createmenu {m w} {
	menu $m
	menu $m.files 
	$m add cascade -menu $m.files -label [legendes::trans "File"]
	$m.files add command -label [legendes::trans "New"]  -command [list tradedit::new_trad $w]
	$m.files add command -label [legendes::trans "Load"]  -command [list tradedit::load_file $w]
	$m.files add command -label [legendes::trans "Save"]  -command [list tradedit::save $w]
	$m.files add command -label [legendes::trans "Quit"]  -command [list destroy $w]

	
	menu $m.divers
	$m add cascade -menu $m.divers -label [legendes::trans "Misc"]
	$m.divers add command -label [legendes::trans "Open dictionary"] -command [list findWordInDictionary $w.s]\
	-accelerator "ESC-d"	
    }
    
    
    proc init {w} {
	variable VARS
	set VARS($w,hiero,file) "";	# name of the hieroglyphic text in 
					# the database
	set VARS($w,trans,file) "";		# translitteration file for the text
	set VARS($w,trad,file) "";		# translation file for the text
	set VARS($w,trans,write) 0;		# is translitteration file writable ?
	set VARS($w,trad,write) 0;		# is  translation file writable ?
    }

    proc clear {w} {
	init $w
	seshw::clear_text [seshw::gethwindow $w.s]
	$w.trad delete 0.0 end
	$w.trans delete 0.0 end
    }
    
    proc new_trad {w} {
	variable VARS
	clear $w
	catch {destroy .tmp}
	toplevel .tmp
	dbfilesel::create .tmp.hiero \
	    -ext hie
	dbfilesel::create .tmp.trans \
	    -ext trl
	dbfilesel::create .tmp.trad \
	    -ext txt
	label .tmp.hierol -text [legendes::trans "Hieroglyphic text"]
	label .tmp.transl -text [legendes::trans "Transliteration"]
	label .tmp.tradl -text [legendes::trans "Translation"]

	button .tmp.ok -text [legendes::trans "Ok"] -command [list tradedit::validate_file_choice $w]
	button .tmp.cancel -text [legendes::trans "Cancel"] -command {destroy .tmp}
	
	grid .tmp.hierol .tmp.hiero
	grid .tmp.transl .tmp.trans
	grid .tmp.tradl .tmp.trad
	grid .tmp.ok .tmp.cancel
    }

    proc load_file {w} {
	variable VARS
	catch {
	    clear $w
	    set name [dbfilesel::getname -ext txt]
	    
	    # First, get the names from the dbfilesel
	    set VARS($w,hiero,file) ""
	    set VARS($w,trans,file) ""
	    set VARS($w,trad,file) $name
	    
	    load_subfiles $w
	}
    }

    proc known {f} {
	return [expr [string length $f] != 0]
    }

    proc known_and_exists {w what} {
	variable VARS
	set name $VARS($w,$what,file)
	return [expr [known $name] \
		    && [seshtext::exists $name [suffix $what]]]
    }
    
    proc validate_file_choice {w} {
	variable VARS
	#
	# try to find the files to load
	
	# nofound is true if unsufficient info is provided.
	# The information is sufficient if all names can be guessed.
	
	set notfound 1

	# First, get the names from the dbfilesel

	set VARS($w,hiero,file) [dbfilesel::gettext .tmp.hiero]
	set VARS($w,trans,file) [dbfilesel::gettext .tmp.trans]
	set VARS($w,trad,file) [dbfilesel::gettext .tmp.trad]

	# If trad is known and the file exists => good
	# else if trans is known and file exists => good 
	# else if hiero is known => good
	# else bad.

	if  [known_and_exists $w trad] {
	    set notfound 0
	} else {
	    if  [known_and_exists $w trans] {
		set notfound 0
	    } else {
		if  [known $VARS($w,hiero,file)] {
		    set notfound 0
		}
	    }
	}
	
	if {! [known $VARS($w,trad,file)]} {
	    if [known $VARS($w,trans,file)] {
		set VARS($w,trad,file) $VARS($w,trans,file)
	    } else {
		set VARS($w,trad,file) $VARS($w,hiero,file)
	    }
	}
	
	if  {! [known $VARS($w,trans,file)]} {
	    set VARS($w,trans,file) $VARS($w,hiero,file)
	}

	if {$notfound} {
	    error [legendes::trans "Not enough information"]
	}
	load_subfiles $w
	destroy .tmp
    }

    
    # Loads the files whose name was selected
    # if a file does not exist AND the directory is not writable by the user,
    # generates an error
    
    # Otherwise, sets the readonly variable for the file if needed
    # what = one of {trad,trans}
    

    proc test_file {w what} {
	variable VARS
	set f $VARS($w,$what,file) 
	set VARS($w,$what,write) [seshtext::private $f] 
    }

    # load_subfiles : load all the files if they exist, starting
    # with the translation.
    # if a file already exit, the system will load the corresponding parent
    # files.

    proc load_subfiles {w} {
	variable VARS

	# loading translation...
	test_file $w trad
	if [seshtext::exists $VARS($w,trad,file) txt] {
	    from_text $w trad [seshtext::getfrombase $VARS($w,trad,file) txt]
	}
	
	# loading transliteration
	test_file $w trans
	if [seshtext::exists $VARS($w,trans,file) trl] {
	    from_text $w trans [seshtext::getfrombase $VARS($w,trans,file) trl]
	}

	# loading hieroglyhs
	if {! [seshtext::exists $VARS($w,hiero,file)] } {
	    error [legendes::trans "text %s does not exist"  $VARS($w,hiero,file)]
	}
	seshw::fromtext $w.s [seshtext::getfrombase $VARS($w,hiero,file)]
    }


#     proc select_and_load {w} {
# 	variable VARS
# 	set patterns {{{MDC file} {.hie}} {{tksesh project} {.tks}}  {Any *}}
# 	set nom [tk_getOpenFile \
# 		     -defaultextension .hie -filetypes $patterns]
# 	regsub {^(.*)(\.[^.]*)$} $nom {\1} nom
# 	set VARS($w,fic) $nom
# 	### A CHANGER ....
# 	# seshw::load $w.s "$nom.hie"
# 	set fh [open "$nom.hie"]
# 	set hieros [read $fh]
# 	close $fh
# 	#puts $w
# 	seshw::fromtext $w.s $hieros
# 	catch {load_text $w trad "$nom.txt"}
# 	catch {load_text $w trans  "$nom.trs"}
# 	#prepare $w $w.trad
# 	$w.s.s mark set insert 0.0
#     }

    proc save {w} {
	variable VARS
	if $VARS($w,trans,write) {
	    seshtext::savetobase $VARS($w,trans,file) [to_text $w trans] trl
	}

	if $VARS($w,trad,write) {
	    seshtext::savetobase $VARS($w,trad,file) [to_text $w trad] txt
	}
    }


    # prepare is at the time being, NOT USED.  the idea behind prepare
    # is to have explicit markers in the text.  This should allow
    # parallels to be edited.  Yet, for single texts, this will be an
    # unnecessary burden. Hence, I decided to ommit it now.
    # One may remark that the current editing format support 
    # multiple texts.
    
    proc prepare {w subwin} {
	set h [seshw::gethwindow $w.s]
	$h mark set pos 1.0
	set i 0
	$subwin tag configure coord -background green
	while {[$h compare pos < end]} {
	    incr i
	    set lab $subwin.lab$i
	    label $lab -text "[$h coords get pos lineend]"
	    $subwin window create insert -window $lab
	    $subwin tag add coord$i "insert - 1 chars"
	    $subwin insert insert "\n"
	    $h mark set pos pos + 1 line
	}
    }
    
    proc create {w} {
	variable VARS
	#seshw::create_editor $w.s -height 173 -wrap 1 -htline 30
	seshw::create_editor $w.s -height 100 -wrap 1 -htline 30
	text $w.trans -height 10 \
	    -yscrollcommand "$w.sb2y set" \
	    -xscrollcommand "$w.sb2x set"\
	    -font {Transliteration 18}
	scrollbar $w.sb2x -command "$w.trans xview" -orient horizontal
	scrollbar $w.sb2y -command "$w.trans yview" -orient vertical
	
	text $w.trad -height 10 \
	    -yscrollcommand "$w.sb1y set" \
	    -xscrollcommand "$w.sb1x set"\
	    -font {times 14}
	scrollbar $w.sb1x -command "$w.trad xview" -orient horizontal
	scrollbar $w.sb1y -command "$w.trad yview" -orient vertical

	set VARS($w.trad,setpos) s1
	set VARS($w.trans,setpos) s1
	set VARS($w.s,setpos) s2
	set VARS($w.trad,parent) $w
	set VARS($w.trans,parent) $w
	set VARS($w.s,parent) $w
	set VARS($w.trad,getpos) getpos1
	set VARS($w.trans,getpos) getpos1
	set VARS($w.s,getpos) getpos2
	set VARS($w.trad,others) [list $w.trans $w.s]
	set VARS($w.trans,others) [list $w.trad $w.s]
	set VARS($w.s,others)  [list $w.trad $w.trans]


	set VARS($w,insync) 0

	seshw::bind_onmove $w.s tradedit::sesh_callback [list $w.trad $w.trans]

	grid $w.s - -sticky nsew
	grid $w.trans $w.sb2y -sticky nsew
	grid $w.sb2x -sticky ew
	grid $w.trad $w.sb1y -sticky nsew
	grid $w.sb1x -sticky ew
	
	grid rowconfigure $w 0 -weight 0
	grid rowconfigure $w 1 -weight 2
	grid rowconfigure $w 3 -weight 3
	grid rowconfigure $w 2 -weight 0
	grid rowconfigure $w 4 -weight 0
	grid columnconfigure $w 0 -weight 1
	grid columnconfigure $w 1 -weight 0

	bind $w.trans <F2> [list tradedit::insert_translit $w]
	
	foreach sub [list $w.trans $w.trad] {
	    set b [bindtags $sub]
	    lappend b edittrad
	    bindtags $sub $b
	}
	
	bind edittrad <Key> {tradedit::syncr %W}
	bind edittrad <Button> {tradedit::syncr %W}
	bind edittrad <Motion> {tradedit::syncr %W}

	bind $w.s <Key-Escape><Key-d> "dicoload::fromsesh $w.s; break"
	bind $w.trad <Tab> "tradedit::setfocus $w.trans ; break"
	bind $w.trans <Tab> "tradedit::setfocus $w.trad ; break"

	if 0 {
	    bind edit <Down>  "tradedit::down %W ; break"
	    bind $w.trans <Return>  "tradedit::down %W ; break"
	    bind $w.trad <Down> "tradedit::down %W; break"
	    bind $w.trad <Return> "tradedit::down %W; break"
	    bind $w.trans <Up>  "tradedit::up %W; break"
	    bind $w.trad <Up> "tradedit::up %W; break"
	}
    }
    
    proc insert_translit {w} {
	set txt [$w.s.s liste_signes insert linestart insert lineend]
	set tr [translit $txt]
	$w.trans insert insert $tr
    }
    
    proc setfocus {win} {
	focus $win
    }

    proc down {w} {
	variable VARS
	go_next_line $w
	syncr $w
    }

    proc up {w} {
	variable VARS
	go_previous_line $w
	syncr $w
    }

    proc s1 {w pos} {
	# If pos is farer than end,
	# we should add lines
	if [$w compare end == $pos.0 ] {
	    scan [$w index end] "%d." end
	    for {set line $end} {$line <= $pos} {incr line} {
		$w insert end "\n"
	    }
	}
	$w mark set insert "$pos.0"
	$w see insert
    }

    
    proc s2 {w pos} {
	set win [seshw::gethwindow $w]
	seshw::sesh_depl $win "$pos.0"
	$win mark set ecran insert	
    }

    proc getpos1 {w} {
	set pos [$w index insert]
	scan $pos "%d." pos
	return $pos
    }

    proc getpos2 {w} {
	set win [seshw::gethwindow $w]
	set pos [$win index insert]
	scan $pos "%d." pos
	return $pos
    }

    proc sesh_callback {hiero data} {
	set w1 [lindex $data 0]
	set w2 [lindex $data 1]
	syncr $hiero
    }

    # synchronize other windows with w1
    proc syncr {w1} {
	variable VARS
	# Avoid recursive calls :
	set w $VARS($w1,parent)
	if {! $VARS($w,insync)} {
	    set VARS($w,insync) 1
	    set pos [$VARS($w1,getpos) $w1]
	    foreach other $VARS($w1,others) {
		$VARS($other,setpos) $other $pos
	    }
	    set VARS($w,insync) 0
	}
    }

    proc go_next_line {win} {
	$win mark set pos "insert lineend"
	$win mark set insert "insert + 1 line lineend"
	if [$win compare insert == pos] {
	    $win insert insert "\n"
	}
	$win see insert
    }

    proc go_previous_line {win} {
	$win mark set insert "insert - 1 line"
	$win see insert
    }

    proc parent {win} {
	regsub {\.[a-zA-Z]*$} $win {} res
	return $res
    }

    proc suffix {what} {
	switch -exact -- $what {
	    hiero {set res hie}
	    trans {set res trl}
	    trad  {set res txt}
	    default {error "BAD type for file : $what"}
	}
	return $res
    }
    
#     proc load_text {w what} {
# 	variable VARS
# 	set f $VARS($w,$what,file)
# 	set win $w.$what
# 	$win delete 0.0 end	
# 	set fic [open $f]
# 	set txt [read $fic]
# 	from_text $win $txt
# 	close $fic
#     }
    
    proc from_text {w what txt} {
	variable VARS 
	set win $w.$what
	$win delete 0.0 end
	set txt [split $txt "\n\r"]
	
	regsub "TKSSRC\t" [lindex $txt 0] "" parent

	switch -exact -- $what {
	    trans {set VARS($w,hiero,file) $parent}
	    trad  {set VARS($w,trans,file) $parent}
	    default {error "BAD type for file : $what"}
	}
	
	set txt [lrange $txt 1 end]
	foreach l $txt {
	    regsub {^([^	])*	} $l {} l
	    $win insert insert $l
	    $win insert insert "\n"
	}
	$win mark set insert 0.0	
    }

    
    proc heading {w what} {
	variable VARS
	switch -exact -- $what {
	    trans {set src $VARS($w,hiero,file)}
	    trad  {set src $VARS($w,trans,file)}
	    default {error "BAD type for file : $what"}
	}
	return [format "TKSSRC\t%s\n" $src]
    }

    proc to_text {w what} {
	set win $w.$what
	set hiero  $w.s
	set hierowin [seshw::gethwindow $hiero]
	set txt [heading $w $what]

	$win mark set posi 1.0
	$hierowin mark set posi 1.0
	while {[$win compare posi < end]} {
	    append txt "[$hierowin coords get posi lineend]\t"
	    append txt [$win get posi "posi lineend"]
	    append txt "\n"
	    $win mark set posi "posi + 1 lines"
	    $hierowin mark set posi posi + 1 line
	}
	return $txt
    }

#proc getline {win idx} {
#regexp {([0-9]*)\..*} [$win index $idx] toto result
#return $result
#}
    
#     proc save_text {win name} {
# 	set txt [to_text $win]
# 	set f [open $name w]
# 	puts $f $txt
# 	close $f
#     }
}
}

